import 'package:sendbird_sdk/features/poll/poll.dart';
import 'package:sendbird_sdk/features/poll/poll_data.dart';
import 'package:sendbird_sdk/features/poll/poll_option.dart';
import 'package:sendbird_sdk/params/poll_params.dart';
import 'package:sendbird_sdk/params/poll_retrieval_params.dart';
import 'package:sendbird_sdk/params/poll_update_params.dart';
import 'package:sendbird_sdk/query/poll_list/poll_list_query.dart';
import 'package:sendbird_sdk/query/poll_list/poll_voter_list_query.dart';
import 'package:sendbird_sdk/request/messages/message_get_request.dart';
import 'package:sendbird_sdk/request/messages/user_message_update_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../common/callback.dart';
import '../../common/common.dart';
import '../../common/helpers/private_api_client.dart';
import '../../common/sdksetup.dart';
import '../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late User user;
  late List<User> users;
  late ChannelEventListener event;
  late GroupChannel channel;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    users = await testApiCreateUsers(3);
    user = await testApiConnect(users.first.userId);
    event = ChannelEventListener();
    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = [user.userId];
    channel = await createGroupChannel(params);
  });

  tearDown(() async {
    sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('poll list query ', () {
    test('should succeed with list of all polls', () async {
      // Create Poll
      final params = PollParams(title: 'test', options: ['1']);
      final params2 = PollParams(title: 'test 2', options: ['1']);
      final params3 = PollParams(title: 'test 3', options: ['1']);

      final poll1 = await Poll.create(params: params);
      final poll2 = await Poll.create(params: params2);
      final poll3 = await Poll.create(params: params3);

      final messageParam1 = UserMessageParams(
          message: 'poll test', pollId: poll1.id, includePollDetails: true);
      final messageParam2 = UserMessageParams(
          message: 'poll test 2', pollId: poll2.id, includePollDetails: true);
      final messageParam3 = UserMessageParams(
          message: 'poll test 3', pollId: poll3.id, includePollDetails: true);

      await sendUserMessageViaAPI(
          channel: channel, params: messageParam1, senderId: user.userId);
      await sendUserMessageViaAPI(
          channel: channel, params: messageParam2, senderId: user.userId);
      await sendUserMessageViaAPI(
          channel: channel, params: messageParam3, senderId: user.userId);

      final result = await PollListQuery(
        channelUrl: channel.channelUrl,
        channelType: ChannelType.group,
      ).loadNext();

      expect(result.length, equals(3));
    });

    test('should succeed with list of voters for a poll', () async {
      // Create Poll
      final pollParams = PollParams(title: 'test', options: ['1']);
      final poll1 = await Poll.create(params: pollParams);

      // Send message with poll
      final messageParam1 = UserMessageParams(
          message: 'poll test', pollId: poll1.id, includePollDetails: true);
      await sendUserMessageViaAPI(
        channel: channel,
        params: messageParam1,
        senderId: user.userId,
      );

      // Vote Poll
      await poll1.vote(
        channelType: ChannelType.group,
        channelUrl: channel.channelUrl,
        optionIds: [poll1.details!.options[0].id],
      );

      final pollRetrievalParams = PollRetrievalParams(
        channelUrl: channel.channelUrl,
        pollId: poll1.id,
        showPartialVoterList: true,
        channelType: ChannelType.group,
      );

      //Read Poll
      final result = await Poll.get(params: pollRetrievalParams);

      expect(result.details!.options[0].partialVoters.length, equals(1));
    });

    test('should succeed with list of voters for a poll option', () async {
      // Create Poll
      final pollParams = PollParams(title: 'test 1', options: ['1']);
      final poll1 = await Poll.create(params: pollParams);

      // Send message with poll
      final messageParam1 = UserMessageParams(
          message: 'poll test', pollId: poll1.id, includePollDetails: true);
      await sendUserMessageViaAPI(
        channel: channel,
        params: messageParam1,
        senderId: user.userId,
      );

      // Vote Poll
      await poll1.vote(
        channelType: ChannelType.group,
        channelUrl: channel.channelUrl,
        optionIds: [poll1.details!.options[0].id],
      );

      // Get list of voters
      final result = await PollVoterListQuery(
              channelType: ChannelType.group,
              channelUrl: channel.channelUrl,
              pollId: poll1.id,
              optionId: poll1.details!.options[0].id)
          .loadNext();
      expect(result.length, equals(1));
    });
  });

  group('poll create', () {
    test('should succeed with valid parameters', () async {
      final params = PollParams(title: 'title', options: ['1', '2', '3'])
        ..data = PollData(text: 'title')
        ..isAnonymous = true
        ..allowUserSuggestion = true
        ..allowMutipleVotes = true;

      final result = await Poll.create(params: params);

      expect(result.title, equals('title'));
      expect(result.details!.isAnonymous, isTrue);
      expect(result.details!.allowUserSuggestion, isTrue);
    });

    test('should throw an exception with more than 20 options', () async {
      final options = List.filled(21, 'test');

      expect(() {
        PollParams(title: 'title', options: options);
      }, throwsA(isA<AssertionError>()));
    });

    test('should throw an exception when exceeding max length of text',
        () async {
      var text = '';
      for (var i = 0; i < 2001; i++) {
        text += 'a';
      }

      expect(() {
        PollParams(title: text, options: ['1']);
      }, throwsA(isA<AssertionError>()));
    });
  });

  group('messages with poll', () {
    test('should succeed in sending message with linked to a poll', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);

      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = result.id;

      // Send UserMessage with Poll linked
      var result2 = await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      expect(result2.poll, isNotNull);
      expect(result2.poll!.title, equals('test'));
    });

    test('should succeed in updating message with linked to a poll', () async {
      final params = PollParams(title: 'poll test', options: ['1', '2', '3']);
      final poll1 = await Poll.create(params: params);

      final params2 =
          PollParams(title: 'poll test update', options: ['1', '2']);
      final poll2 = await Poll.create(params: params2);

      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = poll1.id;
      final mParams2 =
          UserMessageParams(message: 'test update', includePollDetails: true)
            ..pollId = poll2.id;

      // Send UserMessage with Poll linked
      var mResult = await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      expect(mResult.poll, isNotNull);
      expect(mResult.poll!.title, equals('poll test'));

      //Update Message with linked poll
      await sdk.getInternal().api.send(
            ChannelUserMessageUpdateApiRequest(
                channelType: ChannelType.group,
                channelUrl: channel.channelUrl,
                messageId: mResult.messageId,
                params: mParams2),
          );

      //Retrieve Message with linked poll
      var mResult2 = await PrivateApiClient().send(ChannelMessageGetRequest(
        channelType: ChannelType.group,
        channelUrl: channel.channelUrl,
        messageId: mResult.messageId,
        params: MessageRetrievalParams(
            channelType: ChannelType.group,
            channelUrl: channel.channelUrl,
            messageId: mResult.messageId)
          ..includePollDetails = true,
      ));

      expect(mResult2.message, equals('test update'));
      expect(mResult2.poll, isNotNull);
      expect(mResult2.poll.title, equals('poll test update'));
    });

    test('should succeed in viewing message with includePollDetails', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);
      expect(result == result, isTrue);

      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = result.id;

      //Send Message with linked poll
      var result2 = await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      expect(result2.poll, isNotNull);
      expect(result2.poll!.title, equals('test'));

      var params2 = MessageRetrievalParams(
          channelType: ChannelType.group,
          channelUrl: channel.channelUrl,
          messageId: result2.messageId)
        ..includePollDetails = true;

      //Retrieve Message with poll
      var result3 = await PrivateApiClient().send(ChannelMessageGetRequest(
          params: params2,
          channelType: ChannelType.group,
          channelUrl: channel.channelUrl,
          messageId: result2.messageId));

      expect(result3.poll, isNotNull);
      expect(result3.poll.title, equals('test'));
    });
  });

  group('poll equality', () {
    test('should return true when comparing two polls with same reference',
        () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);
      expect(result == result, isTrue);
    });

    test(
        'should return true when comparing two polls with same property values',
        () async {
      var result = Poll(id: 1, title: 'test');
      var result2 = Poll(id: 1, title: 'test');
      expect(result == result2, isTrue);
    });

    test(
        'should return false when comparing two polls with different property values',
        () async {
      final params = PollParams(title: 'test', options: ['1']);
      final params2 = PollParams(title: 'test 2', options: ['1']);
      final result = await Poll.create(params: params);
      final result2 = await Poll.create(params: params2);
      expect(result == result2, isFalse);
    });

    test('should return false when comparing two different object', () async {
      var result2 = 1;

      final params = PollParams(title: 'test', options: ['1']);
      final result = await Poll.create(params: params);
      // ignore: unrelated_type_equality_checks
      expect(result == result2, isFalse);
    });
  });

  group('poll option equality', () {
    test(
        'should return true when comparing two options with the same reference',
        () async {
      var result = PollOption(
        pollId: 1,
        createdAt: 0,
        id: 1,
        text: '',
        updatedAt: 0,
        voteCount: 0,
        partialVoters: [],
      );
      expect(result == result, isTrue);
    });

    test(
        'should return true when comparing two options with same property values',
        () async {
      var result = PollOption(
        pollId: 1,
        createdAt: 0,
        id: 1,
        text: '',
        updatedAt: 0,
        voteCount: 0,
        partialVoters: [],
      );
      var result2 = PollOption(
        pollId: 1,
        createdAt: 0,
        id: 1,
        text: '',
        updatedAt: 0,
        voteCount: 0,
        partialVoters: [],
      );
      expect(result == result2, isTrue);
    });

    test('should return false when comparing two options with disparities',
        () async {
      var result = PollOption(
        pollId: 1,
        createdAt: 0,
        id: 1,
        text: '',
        updatedAt: 0,
        voteCount: 0,
        partialVoters: [],
      );
      var result2 = PollOption(
        pollId: 2,
        createdAt: 0,
        id: 2,
        text: '',
        updatedAt: 0,
        voteCount: 0,
        partialVoters: [],
      );
      expect(result == result2, isFalse);
    });

    test('should return false when comparing two different object', () async {
      var result = PollOption(
        pollId: 1,
        createdAt: 0,
        id: 1,
        text: '',
        updatedAt: 0,
        voteCount: 0,
        partialVoters: [],
      );
      var result2 = 1;
      // ignore: unrelated_type_equality_checks
      expect(result == result2, isFalse);
    });
  });

  group('poll get', () {
    test('should succeed with valid poll id', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);
      expect(result.id, isNotNull);

      // send user message with poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = result.id;

      // Send UserMessage with Poll linked
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      final pollRetrievalParams = PollRetrievalParams(
        channelUrl: channel.channelUrl,
        pollId: result.id,
        channelType: ChannelType.group,
      );

      // Get Poll info
      final result2 = await Poll.get(params: pollRetrievalParams);

      expect(result2.title, equals('test'));
      expect(result2.details!.options.length, equals(3));
    });

    test('should throw an exception with invalid poll id', () async {
      final pollRetrievalParams = PollRetrievalParams(
        channelUrl: channel.channelUrl,
        pollId: -1,
        channelType: ChannelType.group,
      );

      expect(Poll.get(params: pollRetrievalParams),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('poll update', () {
    test('should succeed with valid parameter', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);
      expect(result.id, isNotNull);

      //Update Poll
      //? Options are not updated in PollUpdateRequest
      final pollData = PollData(text: 'data');
      final updatedParams = PollUpdateParams(title: 'new test', data: pollData);
      final result2 = await result.update(params: updatedParams);
      expect(result2.title, equals('new test'));
      expect(result2.details!.options.length, equals(3));
      expect(result.id, equals(result2.id));
      expect(result.details!.status, equals(result2.details!.status));
      expect(result.details!.createdAt, equals(result2.details!.createdAt));
      expect(result.details!.closeAt, equals(result2.details!.closeAt));
      expect(result.details!.allowMultipleVotes,
          equals(result2.details!.allowMultipleVotes));
      expect(result.details!.allowUserSuggestion,
          equals(result2.details!.allowUserSuggestion));
      expect(result.details!.isAnonymous, equals(result2.details!.isAnonymous));
      expect(result2.details!.data!.text, equals('data'));
    });

    test('should succeed and apply update event', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      final mesg = await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      // Update Poll
      final pollParams2 = PollUpdateParams(title: 'test update');
      pollResult.update(params: pollParams2);

      // PEDI Command Recieved
      final result =
          await event.waitFor([EventFuncName.onPollUpdated], timeout: 30);

      expect(result[0].data[0].poll.title, equals('test update'));
      expect(mesg.applyPollUpdateEvent(result[0].data[0]), isTrue);
      expect(mesg.poll!.title, equals('test update'));
    });

    test('should throw an exception when title exceeds 2000 letters', () async {
      var text = '';
      for (var i = 0; i < 2001; i++) {
        text += 'a';
      }
      expect(
          () => PollUpdateParams(title: text), throwsA(isA<AssertionError>()));
    });

    test('should throw an exception with invalid poll', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);

      //Delete poll
      await result.delete();

      //Update Poll
      //? Options are not updated in PollUpdateRequest
      final updatedParams = PollUpdateParams(title: 'new test');
      expect(result.update(params: updatedParams),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('poll close', () {
    test('should succeed with valid poll', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);
      expect(result.id, isNotNull);

      // Closes Poll
      final result2 = await result.close();
      expect(result2.title, equals('test'));
    });

    test('should throw an exception with already closed poll', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);
      expect(result.id, isNotNull);

      // Closes Poll
      await result.close();

      //Vote Poll on Closed Poll
      expect(
        result.vote(
            channelType: ChannelType.group,
            channelUrl: channel.channelUrl,
            optionIds: [result.details!.options[0].id]),
        throwsA(isA<BadRequestError>()),
      );
    });
  });

  group('poll delete', () {
    test('should succeed', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);
      expect(result.id, isNotNull);

      //Delete Poll
      await result.delete();
    });

    test('should succeed and apply update event stating removed', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      final mesg = await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      //Delete Poll
      pollResult.delete();

      // PEDI Command Recieved
      final result =
          await event.waitFor([EventFuncName.onPollUpdated], timeout: 30);

      expect(result[0].data[0].status, equals('removed'));
      expect(mesg.applyPollUpdateEvent(result[0].data[0]), isTrue);
      expect(mesg.poll!.details!.status, equals('removed'));
    });

    test('should throw an exception with invalid poll', () async {
      final params = PollParams(title: 'test', options: ['1', '2', '3']);
      final result = await Poll.create(params: params);
      expect(result.id, isNotNull);

      //Delete Poll
      await result.delete();

      //Delete Invalid Poll
      expect(result.delete(), throwsA(isA<BadRequestError>()));
    });
  });

  group('poll add option', () {
    test('should succeed with valid text', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);
      expect(pollResult.id, isNotNull);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      // Add option to poll
      final result2 = await pollResult.addOption(
          channelType: ChannelType.group,
          channelUrl: channel.channelUrl,
          text: 'option 1');

      expect(result2.details!.options.length, equals(4));
      expect(result2.details!.options[0].text, equals('1'));
      expect(result2.details!.options[1].text, equals('2'));
      expect(result2.details!.options[2].text, equals('3'));
      expect(result2.details!.options[3].text, equals('option 1'));
    });

    test('should throw an exception when text exceed 2000 letters', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);
      expect(pollResult.id, isNotNull);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      var text = '';
      for (var i = 0; i < 2001; i++) {
        text += 'a';
      }
      // Add option to poll
      expect(
          () => pollResult.addOption(
                channelUrl: channel.channelUrl,
                text: text,
                channelType: ChannelType.group,
              ),
          throwsA(isA<AssertionError>()));
    });

    test('should throw an exception when exceed 20 options', () async {
      var optionList = <String>[];
      for (var i = 0; i < 20; i++) {
        optionList.add('1');
      }
      final pollParams = PollParams(title: 'test', options: optionList);
      final pollResult = await Poll.create(params: pollParams);
      expect(pollResult.id, isNotNull);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      // Add option to poll
      expect(
          () => pollResult.addOption(
              channelType: ChannelType.group,
              channelUrl: channel.channelUrl,
              text: 'text'),
          throwsA(isA<BadRequestError>()));
    });

    test('should throw an exception with closed poll', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);
      expect(pollResult.id, isNotNull);

      // Closes Poll
      await pollResult.close();
      expect(
          pollResult.addOption(
            channelUrl: channel.channelUrl,
            text: 'text',
            channelType: ChannelType.group,
          ),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('poll read option', () {
    test('should succeed with valid parameters', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      //Get Poll Option
      final pollResult2 = await PollOption.get(
        channelType: ChannelType.group,
        channelUrl: channel.channelUrl,
        optionId: pollResult.details!.options[0].id,
        pollId: pollResult.id,
      );

      expect(pollResult2.text, equals('1'));
    });
  });

  group('poll update option', () {
    test('should succeed with valid params', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      // Update Poll Option
      final result =
          await pollResult.details!.options[0].update(text: 'change 1');

      expect(result.details!.options[0].text, equals('change 1'));
    });
  });

  group('poll delete option', () {
    test('should succeed with valid params', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      //Delete Poll Option
      await pollResult.details!.options[0].delete();

      final pollRetrievalParams = PollRetrievalParams(
        channelType: ChannelType.group,
        channelUrl: channel.channelUrl,
        pollId: pollResult.id,
      );

      //Get Poll
      final result = await Poll.get(params: pollRetrievalParams);
      expect(result.details!.options.length, equals(2));
    });
  });

  group('poll vote', () {
    test('should succeed to cast vote on valid option', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      //Vote poll
      final pollResult2 = await pollResult.vote(
          channelType: ChannelType.group,
          channelUrl: channel.channelUrl,
          optionIds: [pollResult.details!.options[0].id]);

      expect(pollResult2.title, equals('test'));
      expect(pollResult2.details!.voterCount, equals(1));
    });

    test('should succeed to retract vote on valid option', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      //Vote poll
      final pollResult2 = await pollResult.vote(
          channelType: ChannelType.group,
          channelUrl: channel.channelUrl,
          optionIds: [pollResult.details!.options[0].id]);
      expect(pollResult2.details!.voterCount, equals(1));

      //Cancel Vote poll
      final pollResult3 = await pollResult.vote(
        channelUrl: channel.channelUrl,
        optionIds: [],
        channelType: ChannelType.group,
      );
      expect(pollResult3.details!.voterCount, equals(0));
    });

    test('should throw an exception to vote/retract on invalid option',
        () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      // Vote Poll
      expect(
          pollResult.vote(
            channelUrl: channel.channelUrl,
            optionIds: [-1],
            channelType: ChannelType.group,
          ),
          throwsA(isA<BadRequestError>()));
    });

    test('should throw an exception to vote/retract on closed poll', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      //  Closes Poll
      final result2 = await pollResult.close();

      // Vote Poll
      expect(
          result2.vote(
              channelType: ChannelType.group,
              channelUrl: channel.channelUrl,
              optionIds: [pollResult.details!.options[0].id]),
          throwsA(isA<InternalServerError>()));
    });

    test('should throw an exception to vote/retract on invalid channel',
        () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      // Vote Poll
      expect(
          pollResult.vote(
            channelType: ChannelType.group,
            channelUrl: '',
            optionIds: [pollResult.details!.options[0].id],
          ),
          throwsA(isA<BadRequestError>()));
    });

    test('should throw an exception to vote/retract on not linked poll',
        () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      // Vote Poll
      expect(
          pollResult.vote(
              channelType: ChannelType.group,
              channelUrl: channel.channelUrl,
              optionIds: [pollResult.details!.options[0].id]),
          throwsA(isA<BadRequestError>()));
    });

    test('should succeed to applyVoteEvent with valid params', () async {
      final pollParams = PollParams(title: 'test', options: ['1', '2', '3']);
      final pollResult = await Poll.create(params: pollParams);

      //Send UserMessage Linked with Poll
      final mParams =
          UserMessageParams(message: 'test', includePollDetails: true)
            ..pollId = pollResult.id;
      final mesg = await sendUserMessageViaAPI(
          channel: channel, params: mParams, senderId: user.userId);

      // Vote Poll
      pollResult.vote(
        channelType: ChannelType.group,
        channelUrl: channel.channelUrl,
        optionIds: [pollResult.details!.options[0].id],
      );

      final result =
          await event.waitFor([EventFuncName.onPollVoted], timeout: 30);

      expect(mesg.poll!.details!.options[0].voteCount, equals(0));
      expect(mesg.applyPollVoteEvent(result[0].data[0]), isTrue);
      expect(mesg.poll!.details!.options[0].voteCount, equals(1));

      final pollRetrievalParams = PollRetrievalParams(
        channelType: ChannelType.group,
        channelUrl: channel.channelUrl,
        pollId: pollResult.id,
      );

      final mesg2 = await Poll.get(params: pollRetrievalParams);
      expect(mesg2.details!.voterCount, equals(1));
      expect(mesg2.details!.votedOptionIds.length, equals(1));
    });
  });
}
