import 'package:sendbird_sdk/request/reaction/channel_reaction_add_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../common/callback.dart';
import '../../common/common.dart';
import '../../common/helpers/private_api_client.dart';
import '../../common/sdksetup.dart';
import '../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late User user;
  late List<User> users;
  late ChannelEventListener event;
  late GroupChannel channel, channel2;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    users = await testApiCreateUsers(2);
    user = await testApiConnect(users.first.userId);
    event = ChannelEventListener();
    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = users.map((e) => e.userId).toList();
    channel = await createGroupChannel(params);
    channel2 = await createGroupChannel(params..userIds = [users.first.userId]);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('groupchannel adding reaction', () {
    late UserMessage message;
    setUp(() async {
      final params = UserMessageParams(message: 'message');
      message = await sendUserMessageViaAPI(
        channel: channel,
        params: params,
        senderId: users.last.userId,
      );
    });

    test('should succeed with a user message', () async {
      final res = await channel.addReaction(message, 'happy');
      expect(res, isNotNull);
      expect(res.messageId, equals(message.messageId));
      expect(res.key, equals('happy'));
      expect(res.operation, equals(ReactionEventAction.add));
      expect(res.userId, equals(user.userId));
    });

    test('should throw an exception when channels are different', () async {
      expect(() => channel2.addReaction(message, 'happy'),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with invalid params', () async {
      expect(() => channel.addReaction(message, ''),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel deleting reaction', () {
    late UserMessage message;
    setUp(() async {
      final params = UserMessageParams(message: 'message');
      message = await sendUserMessageViaAPI(
        channel: channel,
        params: params,
        senderId: users.last.userId,
      );
    });

    test('should succeed with a user message', () async {
      final res = await channel.addReaction(message, 'happy');
      expect(res, isNotNull);

      final res2 = await channel.deleteReaction(message, 'happy');
      expect(res2, isNotNull);
      expect(res2.messageId, equals(message.messageId));
      expect(res2.key, equals('happy'));
      expect(res2.operation, equals(ReactionEventAction.delete));
      expect(res2.userId, equals(user.userId));
    });

    test('should throw an exception when channels are different', () async {
      expect(() => channel2.deleteReaction(message, 'happy'),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with invalid params', () async {
      expect(() => channel.deleteReaction(message, ''),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel applying reaction event', () {
    late UserMessage message;
    final params = UserMessageParams(message: 'message');

    setUp(() async {
      message = await sendUserMessageViaAPI(
        channel: channel,
        params: params,
        senderId: users.last.userId,
      );
    });

    test('should return true on message', () async {
      final res = await channel.addReaction(message, 'happy');
      await event.waitFor([EventFuncName.onReactionUpdated]);

      final succeed = message.applyReactionEvent(res);
      expect(succeed, isTrue);
      expect(message.reactions?.length, equals(1));
    });

    test('should return ture and merge multiple reaction operations', () async {
      final res = await channel.addReaction(message, 'happy');
      await event.waitFor([EventFuncName.onReactionUpdated]);

      var succeed = message.applyReactionEvent(res);
      expect(succeed, isTrue);
      expect(message.reactions?.length, equals(1));

      PrivateApiClient().send(ChannelReactionAddRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: message.messageId,
        key: 'happy',
        userId: users.last.userId,
      ));

      final result = await event.waitFor([EventFuncName.onReactionUpdated]);
      final reactionEvent = result.first.data.last as ReactionEvent;
      succeed = message.applyReactionEvent(reactionEvent);
      expect(succeed, isTrue);
      expect(message.reactions?.length, equals(1));
      expect(message.reactions?.first.userIds.length, equals(2));
    });

    test('should return false on wrong message', () async {
      final res = await channel.addReaction(message, 'happy');
      await event.waitFor([EventFuncName.onReactionUpdated]);

      final message2 = await sendUserMessageViaAPI(
        channel: channel,
        params: params,
        senderId: users.last.userId,
      );

      expect(message2.applyReactionEvent(res), isFalse);
    });
  });
}
