import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/utils/extensions.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/helpers/private_api_client.dart';
import '../../../common/request.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late User user;
  late List<User> users;
  late ChannelEventListener event;
  late GroupChannel channel;

  setUpAll(
    () async {
      testInfo = await testCreateApp();

      await PrivateApiClient().send(
        AppAttributeUpdateRequest(
          appId: testInfo.appId,
          attributeName: 'enable_message_threading',
          value: true,
        ),
      );
    },
  );

  setUp(
    () async {
      sdk = await testApiSdkSetup(testInfo: testInfo);
      users = await testApiCreateUsers(4);
      user = await testApiConnect(users.first.userId);
      event = ChannelEventListener();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      channel = await createGroupChannel(params);
    },
  );

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group(
    'groupchannel send message as reply',
    () {
      test(
        'should success with replyToChannel and parentMessage',
        () async {
          final parentMessage = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'message'),
            senderId: user.userId,
          );

          final replyToChannelParamsTrue =
              UserMessageParams(message: 'this is reply')
                ..parentMessageId = parentMessage.messageId
                ..data = 'data'
                ..replyToChannel = true;

          final msg = await sendUserMessageViaAPI(
            channel: channel,
            params: replyToChannelParamsTrue,
            senderId: user.userId,
          );

          expect(msg.parentMessage, isNotNull);
          expect(msg.replyToChannel, isTrue);
          expect(msg.message, equals('this is reply'));
          expect(msg.data, equals('data'));
          expect(msg.parentMessageId, equals(parentMessage.messageId));
        },
      );

      test(
        'should success with user message on root user message',
        () async {
          final parentMessage = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'message'),
            senderId: user.userId,
          );

          final params = UserMessageParams(message: 'this is reply')
            ..parentMessageId = parentMessage.messageId
            ..data = 'data';

          final msg = await sendUserMessageViaAPI(
            channel: channel,
            params: params,
            senderId: user.userId,
          );

          expect(msg.replyToChannel, isFalse);
          expect(msg.parentMessage, isNotNull);
          expect(msg.message, equals('this is reply'));
          expect(msg.data, equals('data'));
          expect(msg.parentMessageId, equals(parentMessage.messageId));
        },
      );

      test(
        'should success with file message on root user message',
        () async {
          final parentMessage = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'message'),
            senderId: user.userId,
          );

          final params = FileMessageParams.withUrl(imageUrl, size: 123)
            ..parentMessageId = parentMessage.messageId
            ..data = 'data';

          final msg = await sendFileMessageViaAPI(
            channel: channel,
            params: params,
            senderId: user.userId,
          );

          expect(msg.url, equals(imageUrl));
          expect(msg.data, equals('data'));
          expect(msg.parentMessageId, equals(parentMessage.messageId));
        },
      );

      test(
        'should success with user message on root file message',
        () async {
          final params = FileMessageParams.withUrl(imageUrl, size: 123);
          final parentMessage = await sendFileMessageViaAPI(
            channel: channel,
            params: params,
            senderId: user.userId,
          );

          final replyParams = UserMessageParams(message: 'this is reply')
            ..parentMessageId = parentMessage.messageId;

          final msg = await sendUserMessageViaAPI(
            channel: channel,
            params: replyParams,
            senderId: user.userId,
          );

          expect(msg.message, equals('this is reply'));
          expect(msg.parentMessageId, equals(parentMessage.messageId));
        },
      );

      test(
        'should success with file message on root file message',
        () async {
          final params = FileMessageParams.withUrl(imageUrl, size: 123);
          final parentMessage = await sendFileMessageViaAPI(
            channel: channel,
            params: params,
            senderId: user.userId,
          );

          params
            ..parentMessageId = parentMessage.messageId
            ..uploadFile.url = 'http://www.naver.com'
            ..uploadFile.fileSize = 567;

          final msg = await sendFileMessageViaAPI(
            channel: channel,
            params: params,
            senderId: user.userId,
          );

          expect(msg.url, equals('http://www.naver.com'));
          expect(msg.size, equals(567));
          expect(msg.parentMessageId, equals(parentMessage.messageId));
        },
      );

      test(
        'should throw an exception with given invalid parent id',
        () async {
          final params = UserMessageParams(message: 'this is reply')
            ..parentMessageId = 1231231231
            ..data = 'data';

          expect(
            () => sendUserMessageViaAPI(
              channel: channel,
              params: params,
              senderId: user.userId,
            ),
            throwsA(isA<SBError>()),
          );
        },
      );

      test(
        'should success to apply thread info into given message',
        () async {
          final parentMessage = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'message'),
            senderId: user.userId,
          );

          final params = UserMessageParams(message: 'this is reply')
            ..parentMessageId = parentMessage.messageId
            ..data = 'data';

          late UserMessage message;
          late ThreadInfoUpdateEvent threadEvent;

          channel.sendUserMessage(
            params,
            onCompleted: (msg, error) {
              message = msg;
            },
          );

          final res = await event.waitFor(
            [
              EventFuncName.onThreadInfoUpdated,
              EventFuncName.onChannelChanged,
            ],
          );

          for (var each in res) {
            if (each.fncName == EventFuncName.onThreadInfoUpdated) {
              threadEvent = each.data.last as ThreadInfoUpdateEvent;
            }
          }
          expect(message, isNotNull);
          expect(threadEvent, isNotNull);

          parentMessage.applyThreadInfoUpdateEvent(threadEvent);
          expect(parentMessage.threadInfo!.updatedAt,
              equals(threadEvent.threadInfo.updatedAt));
          expect(parentMessage.threadInfo!.replyCount,
              equals(threadEvent.threadInfo.replyCount));
        },
      );

      test(
        'should success to apply thread info into given message',
        () async {
          final parentMessage = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'message'),
            senderId: user.userId,
          );

          final params = UserMessageParams(message: 'this is reply')
            ..parentMessageId = parentMessage.messageId
            ..data = 'data';

          late UserMessage message;
          late ThreadInfoUpdateEvent threadEvent;

          channel.sendUserMessage(
            params,
            onCompleted: (msg, error) {
              message = msg;
            },
          );

          final res = await event.waitFor(
            [
              EventFuncName.onThreadInfoUpdated,
              EventFuncName.onChannelChanged,
            ],
          );

          for (var each in res) {
            if (each.fncName == EventFuncName.onThreadInfoUpdated) {
              threadEvent = each.data.last as ThreadInfoUpdateEvent;
            }
          }
          expect(message, isNotNull);
          expect(threadEvent, isNotNull);

          parentMessage.applyThreadInfoUpdateEvent(threadEvent);
          expect(parentMessage.threadInfo!.updatedAt,
              equals(threadEvent.threadInfo.updatedAt));
          expect(parentMessage.threadInfo!.replyCount,
              equals(threadEvent.threadInfo.replyCount));
        },
      );
    },
  );

  group(
    'groupchannel get threaded message',
    () {
      //message (user)
      //- reply (user)
      //- reply (file)
      //message2 (file)
      //- reply (user)
      test(
        'should return replies and thread info',
        () async {
          final msg1 = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'this is test'),
            senderId: users[2].userId,
            shouldWaitForCallback: false,
          );

          final msg2 = await sendFileMessageViaAPI(
            channel: channel,
            params: FileMessageParams.withUrl(imageUrl),
            senderId: users[1].userId,
            shouldWaitForCallback: false,
          );

          final reply11 = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'reply 11')
              ..parentMessageId = msg1.messageId,
            senderId: user.userId,
            shouldWaitForCallback: false,
          );

          final reply12 = await sendFileMessageViaAPI(
            channel: channel,
            params: FileMessageParams.withUrl(imageUrl)
              ..parentMessageId = msg1.messageId,
            senderId: users[3].userId,
            shouldWaitForCallback: false,
          );

          final reply21 = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'second message reply')
              ..parentMessageId = msg2.messageId,
            senderId: user.userId,
            shouldWaitForCallback: false,
          );

          final params = ThreadedMessageListParams()
            ..replyType = ReplyType.all
            ..nextResultSize = 10
            ..previousResultSize = 10
            ..includeParentMessageText = true
            ..isInclusive = true
            ..replyType = ReplyType.all;

          final time = DateTime.now().millisecondsSinceEpoch;
          final respond =
              await msg1.getThreadedMessagesByTimestamp(time, params);
          expect(respond, isNotNull);
          expect(respond.parentMessage.messageId, equals(msg1.messageId));
          expect(respond.parentMessage.message, equals(msg1.message));
          expect(respond.parentMessage.threadInfo, isNotNull);
          expect(respond.parentMessage.threadInfo?.replyCount, equals(2));
          expect(respond.replies.length, equals(2));
          expect(respond.replies.first.messageId, equals(reply11.messageId));
          expect(respond.replies.last.messageId, equals(reply12.messageId));

          final respond2 =
              await msg2.getThreadedMessagesByTimestamp(time, params);
          expect(respond2, isNotNull);
          expect(respond2.parentMessage.messageId, equals(msg2.messageId));
          expect(respond2.parentMessage.threadInfo, isNotNull);
          expect(respond2.parentMessage.threadInfo?.replyCount, equals(1));
          expect(respond2.replies.length, equals(1));
          expect(respond2.replies.first.messageId, equals(reply21.messageId));
        },
      );
      test(
        'should return valid message when getting one message',
        () async {
          final message = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'test'),
            senderId: user.userId,
          );

          final reply = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'reply')
              ..replyToChannel = true
              ..parentMessageId = message.messageId,
            senderId: user.userId,
          );

          expect(reply.replyToChannel, isTrue);
          expect(reply.parentMessage, isNotNull);

          final params = MessageRetrievalParams(
            channelType: channel.channelType,
            channelUrl: channel.channelUrl,
            messageId: reply.messageId,
          )..includeParentMessageInfo = true;

          final result = await BaseMessage.getMessage(params);
          expect(result.replyToChannel, isTrue);
          expect(result.parentMessage, isNotNull);
          expect(result.messageId, equals(result.messageId));
          expect(result.message, equals(result.message));
          expect(result, isA<UserMessage>());
        },
      );
    },
  );

  group(
    'groupchannel get message list include threaded options',
    () {
      test(
        'should return replies and thread info',
        () async {
          final msg1 = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'this is test'),
            senderId: users[2].userId,
          );

          await sendFileMessageViaAPI(
            channel: channel,
            params: FileMessageParams.withUrl(imageUrl, size: 123),
            senderId: users[1].userId,
          );

          await sendFileMessageViaAPI(
            channel: channel,
            params: FileMessageParams.withUrl(imageUrl, size: 123)
              ..parentMessageId = msg1.messageId,
            senderId: users[3].userId,
          );

          final params = MessageListParams()
            ..replyType = ReplyType.all
            ..includeReplies = true
            ..includeParentMessageText = true
            ..includeThreadInfo = true
            ..previousResultSize = 10;

          final result = await channel.getMessagesByTimestamp(
            ExtendedInteger.max,
            params,
          );
          expect(result, isNotNull);
          expect(result.length, equals(3));
          expect(result.first.messageId, equals(msg1.messageId));
          expect(result.last.parentMessageId, equals(msg1.messageId));
        },
      );

      test(
        'should return correct replies and thread info on multiple threaded',
        () async {
          final msg1 = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'this is test'),
            senderId: users[2].userId,
          );

          final msg2 = await sendFileMessageViaAPI(
            channel: channel,
            params: FileMessageParams.withUrl(imageUrl, size: 123),
            senderId: users[1].userId,
          );

          await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'reply 11')
              ..parentMessageId = msg1.messageId,
            senderId: user.userId,
          );

          await sendFileMessageViaAPI(
            channel: channel,
            params: FileMessageParams.withUrl(imageUrl, size: 123)
              ..parentMessageId = msg1.messageId,
            senderId: users[3].userId,
          );

          await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'second message reply')
              ..parentMessageId = msg2.messageId,
            senderId: user.userId,
          );

          final msg3 = await sendUserMessageViaAPI(
            channel: channel,
            params: UserMessageParams(message: 'this is test'),
            senderId: user.userId,
          );

          await sendFileMessageViaAPI(
            channel: channel,
            params: FileMessageParams.withUrl(imageUrl, size: 123)
              ..parentMessageId = msg3.messageId,
            senderId: user.userId,
          );

          final params = MessageListParams()
            ..replyType = ReplyType.all
            ..includeReplies = true
            ..includeThreadInfo = true
            ..previousResultSize = 10;

          final result = await channel.getMessagesByTimestamp(
            ExtendedInteger.max,
            params,
          );
          expect(result, isNotNull);
          expect(result.length, equals(7));
          expect(result.first.messageId, equals(msg1.messageId));
          expect(result.first.threadInfo?.replyCount, equals(2));
          expect(result.last.parentMessageId, equals(msg3.messageId));
        },
      );
    },
  );
}
