import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/request/channel/group_channel_create_request.dart';
import 'package:sendbird_sdk/request/channel_operation/group_channel_invite_request.dart';
import 'package:sendbird_sdk/request/channel_operation/group_channel_join_request.dart';
import 'package:sendbird_sdk/request/channel_operation/group_channel_leave_request.dart';
import 'package:sendbird_sdk/request/messages/user_message_send_request.dart';
import 'package:sendbird_sdk/request/user/info_update_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/helpers/private_api_client.dart';
import '../../../common/prep.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';
import '../../../common/utils.dart';
import '../../../common/ws_tester.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('groupchannel message releated event', () {
    test(
        'onReceiveMessage & onChannelChanged should get called when receiving message',
        () async {
      final users = await testApiCreateUsers(3);
      final user = users.first;
      await testApiConnect(users.first.userId);

      await PrivateApiClient().send(UserInfoUpdateRequest(
        userId: users.last.userId,
        nickname: 'senderrr',
        fileInfo: FileInfo.fromUrl(
          url: 'https://google.com',
          mimeType: 'image/jpeg',
        ),
      ));

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      var channel = await createGroupChannel(params);

      final mParams = UserMessageParams(message: 'test');
      final event = ChannelEventListener();

      PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
        params: mParams,
      ));

      final results = await event.waitFor([
        EventFuncName.onMessageReceived,
        EventFuncName.onChannelChanged,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(2));
      final received = results.first.data.last as UserMessage;
      expect(received, isNotNull);
      expect(received.message, 'test');
      expect(received.sender, isNotNull);
      expect(received.sender?.nickname, equals('senderrr'));
      expect(received.sender?.profileUrl, equals('https://google.com'));
      expect(received.sender?.userId, equals(users.last.userId));
    });

    test(
        'onMentionReceived & onMessageReceived & onChannelChanged should call when receiving message with mention',
        () async {
      final users = await testApiCreateUsers(3);
      final user = users.first;
      await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      var channel = await createGroupChannel(params);

      final mParams = UserMessageParams(message: 'test')
        ..mentionType = MentionType.users
        ..mentionedUserIds = [user.userId];

      final event = ChannelEventListener();

      await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
        params: mParams,
      ));

      final results = await event.waitFor([
        EventFuncName.onMessageReceived,
        EventFuncName.onChannelChanged,
        EventFuncName.onMentionReceived,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(3));
      final received = results.first.data.last as UserMessage;
      expect(received, isNotNull);
      expect(received.message, 'test');
    });
  });
  group('groupchannel operation releated event', () {
    test(
        'should get onReceiveInvitation & onUserJoin called on invite & accept',
        () async {
      final users = await testApiCreateUsers(3);
      final user = users.first;
      await testApiConnect(users.first.userId);
      await SendbirdSdk().setChannelInvitationPreference(false);

      final params = GroupChannelParams()
        ..operatorUserIds = [users.last.userId]
        ..userIds = [users.last.userId];

      final channel = await PrivateApiClient()
          .send<GroupChannel>(GroupChannelCreateRequest(params));

      final event = ChannelEventListener();

      PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: [user.userId],
        channelUrl: channel.channelUrl,
        inviterId: users.last.userId,
      ));

      var results = await event.waitFor([
        EventFuncName.onUserReceivedInvitation,
      ]);

      final cbChannel = results.first.data.first as GroupChannel?;
      final invitees = results.first.data[1] as List<Member>?;
      final inviter = results.first.data.last as Member?;

      expect(results, isNotNull);
      expect(results.length, equals(1));
      expect(cbChannel, isNotNull);
      expect(cbChannel?.members.length, 2);
      expect(invitees, isNotNull);
      expect(invitees?.length, 1);
      expect(invitees?.first.state, MemberState.invited);
      expect(inviter, isNotNull);
      expect(inviter?.state, MemberState.joined);

      //check member state
      channel.acceptInvitation();

      results = await event.waitFor([
        EventFuncName.onUserJoined,
      ]);

      final cbChannel2 = results.first.data.first as GroupChannel?;
      final member = results.first.data.last as Member?;

      expect(results, isNotNull);
      expect(results.length, equals(1));
      expect(cbChannel2, isNotNull);
      expect(cbChannel2?.members.length, 2);
      expect(member, isNotNull);
      expect(member?.state, MemberState.joined);
    });

    test('onDeclineInvitation should get called on decline', () async {
      final users = await testApiCreateUsers(3);
      final user = users.first;
      await testApiConnect(users.first.userId);
      await SendbirdSdk().setChannelInvitationPreference(false);

      final params = GroupChannelParams()
        ..operatorUserIds = [users.last.userId]
        ..userIds = [users.last.userId];

      final channel = await PrivateApiClient()
          .send<GroupChannel>(GroupChannelCreateRequest(params));

      final event = ChannelEventListener();

      PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: [user.userId],
        channelUrl: channel.channelUrl,
        inviterId: users.last.userId,
      ));

      var results = await event.waitFor([
        EventFuncName.onUserReceivedInvitation,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));

      channel.declineInvitation();

      results = await event.waitFor([
        EventFuncName.onUserDeclinedInvitation,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test(
        'onChannelChanged & onChannelOperatorUpdated should get called on change channel',
        () async {
      final user = await testApiCreateUserAndConnect();
      final users = await testApiCreateUsers(2);
      final imageFile = await getAssetFrom('resources/image.jpeg');

      final params = GroupChannelParams()
        ..name = 'channel'
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();

      final event = ChannelEventListener();
      final channel = await GroupChannel.createChannel(params);

      final results = await event.waitFor([
        EventFuncName.onUserJoined,
        EventFuncName.onUserReceivedInvitation,
      ]);

      expect(channel, isNotNull);
      expect(channel.coverUrl, isNotNull);
      expect(channel.name, 'channel');
      expect(results, isNotNull);
      expect(results.length, equals(2));

      final event2 = ChannelEventListener();

      final updatedParams = GroupChannelParams()
        ..name = 'update name'
        ..channelUrl = channel.channelUrl
        ..operatorUserIds = users.map((e) => e.userId).toList()
        ..coverImage = FileInfo.fromData(
          file: imageFile,
          name: 'Cover Image',
          mimeType: 'image/jpeg',
        );

      channel.updateChannel(updatedParams);

      final results2 = await event2.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onChannelOperatorsUpdated,
      ]);

      expect(results2, isNotNull);
      expect(results2.length, equals(2));
    });

    test('onChannelDelete should get called when delete the channel', () async {
      final user = await testApiCreateUserAndConnect();

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      var channel = await createGroupChannel(params);

      final event = ChannelEventListener();

      channel.deleteChannel();

      final results = await event.waitFor([
        EventFuncName.onChannelDeleted,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });
  });

  group('groupchannel metadata event', () {
    test('onChannelMetaDataCreated should get call when creating meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);

      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      await channel.createMetaData(metaData);

      final results = await event.waitFor([
        EventFuncName.onMetaDataCreated,
      ], timeout: 10);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onMetaUpdated should get call when updating meta', () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);
      final event = ChannelEventListener();

      final metaData = {'key1': 'value1'};
      await channel.createMetaData(metaData);
      await channel.updateMetaData({'key1': 'value2'});

      final results = await event.waitFor([
        EventFuncName.onMetaDataUpdated,
      ], timeout: 10);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onMetaDeleted should get call when deleting meta', () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      await channel.createMetaData(metaData);
      await channel.deleteMetaData('key1');

      final results = await event.waitFor([EventFuncName.onMetaDataDeleted]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });
  });

  group('groupchannel metacounter event', () {
    test('onChannelMetaCounterCreated should get call when creating meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);

      await channel.createMetaCounters({'key1': 1});

      final event = ChannelEventListener();
      final results = await event.waitFor([
        EventFuncName.onMetaCountersCreated,
      ], timeout: 10);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onChannelMetaCounterUpdated should get call when updating meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);
      final event = ChannelEventListener();

      await channel.createMetaCounters({'key1': 1});
      await channel.updateMetaCounters({'key1': 2});

      var results = await event.waitFor([
        EventFuncName.onMetaCountersUpdated,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));

      channel.increaseMetaCounters({'key1': 4});

      results = await event.waitFor([
        EventFuncName.onMetaCountersUpdated,
      ]);
      expect(results, isNotNull);
      expect(results.length, equals(1));

      channel.decreaseMetaCounters({'key1': 5});

      results = await event.waitFor([
        EventFuncName.onMetaCountersUpdated,
      ]);
      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onChannelMetaCounterDeleted should get call when deleting meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);

      await channel.createMetaCounters({'key1': 2});
      channel.deleteMetaCounters('key1');

      final event = ChannelEventListener();
      final results = await event.waitFor([
        EventFuncName.onMetaCountersDeleted,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });
  });

  group('groupchannel moderation related events', () {
    test('onUserMuted & onUserUnmuted should get called on mute & unmute',
        () async {
      final users = await testApiCreateUsers(3);
      await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = users.map((e) => e.userId).toList()
        ..userIds = users.map((e) => e.userId).toList();

      final channel = await createGroupChannel(params);

      channel.muteUser(userId: users.last.userId);
      final event = ChannelEventListener();
      var results = await event.waitFor([
        EventFuncName.onUserMuted,
      ]);
      expect(results, isNotNull);
      expect(results.length, equals(1));

      channel.unmuteUser(userId: users.last.userId);
      results = await event.waitFor([
        EventFuncName.onUserUnmuted,
      ]);
      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onUserBanned & onUserUnbanned should get called on ban & unbanned',
        () async {
      final users = await testApiCreateUsers(3);
      await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = users.map((e) => e.userId).toList()
        ..userIds = users.map((e) => e.userId).toList();

      final channel = await createGroupChannel(params);

      channel.banUser(userId: users.last.userId);
      final event = ChannelEventListener();
      var results = await event.waitFor([
        EventFuncName.onUserBanned,
      ]);
      expect(results, isNotNull);
      expect(results.length, equals(1));

      channel.unbanUser(userId: users.last.userId);
      results = await event.waitFor([
        EventFuncName.onUserUnbanned,
      ]);
      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test(
        'onChannelFrozen & onChannelUnfrozen should get call when freeze & unfreeze channel',
        () async {
      final user = await testApiCreateUserAndConnect();

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      var channel = await createGroupChannel(params);

      final event = ChannelEventListener();

      channel.freeze();

      var results = await event.waitFor([
        EventFuncName.onChannelFrozen,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));

      channel.unfreeze();

      results = await event.waitFor([
        EventFuncName.onChannelUnfrozen,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onChannelHidden should get called when hide the channel', () async {
      final user = await testApiCreateUserAndConnect();

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      var channel = await createGroupChannel(params);

      final event = ChannelEventListener();
      await channel.hideChannel();

      expect(channel.isHidden, isTrue);
      final results = await event.waitFor([EventFuncName.onChannelHidden]);
      expect(results, isNotNull);
      expect(results.length, equals(1));
    });
  });

  group('groupchannel member count change event for broadcast', () {
    test('onChannelMemberCountChanged should get called', () async {
      await configureAppForMCNT(
        appId: testInfo.appId,
        eventInterval: 10000,
      );
      final event = ChannelEventListener();

      final otherUser = await testApiCreateUser();
      final user = await testApiCreateUserAndConnect();

      final params = GroupChannelParams()
        ..isBroadcast = true
        ..isPublic = true
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);

      await PrivateApiClient().send(GroupChannelJoinRequest(
        channelUrl: channel.channelUrl,
        userId: otherUser.userId,
      ));

      await event.waitFor([EventFuncName.onChannelMemberCountChanged]);

      await Future.delayed(Duration(seconds: 2));

      await PrivateApiClient().send(GroupChannelLeaveRequest(
        channelUrl: channel.channelUrl,
        userId: otherUser.userId,
      ));

      await event.waitFor([EventFuncName.onChannelMemberCountChanged]);
    });
  });

  group('groupchannel typing', () {
    late List<User> users;
    late GroupChannel channel;
    late ChannelEventListener event;
    late WSTester wsTester;

    setUp(() async {
      users = await testApiCreateUsers(2);
      await testApiConnect(users[0].userId);

      event = ChannelEventListener();
      final params = GroupChannelParams()
        ..operatorUserIds = users.map((e) => e.userId).toList()
        ..userIds = users.map((e) => e.userId).toList();
      channel = await createGroupChannel(params);

      wsTester = WSTester(
        appId: testInfo.appId,
        userId: users[1].userId,
        wsHost: sdk.getInternal().state.wsHost!,
      );
      await wsTester.connect();
    });

    tearDown(() {
      wsTester.close();
    });

    test('should trigger typing update callback when start', () async {
      wsTester.typing(channel, true);
      await event.waitFor([EventFuncName.onTypingStatusUpdated]);
    });

    test('should trigger typing update callback when end', () async {
      wsTester.typing(channel, true);
      await event.waitFor([EventFuncName.onTypingStatusUpdated]);
      wsTester.typing(channel, false);
      await event.waitFor([EventFuncName.onTypingStatusUpdated]);
    });
  });
}
