import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/helpers/private_api_client.dart';
import '../../../common/request.dart';
import '../../../common/user.dart';
import '../../../common/sdksetup.dart';
import '../../../common/utils.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
    final attr = {
      'acl': {'allow_exclusive_channel': true}
    };
    await PrivateApiClient().updateAppAttributes(testInfo.appId, attr);
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('groupchannel create', () {
    test('should create a new channel with given params', () async {
      GroupChannel channel;
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      var params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      channel = await createGroupChannel(params);
      expect(channel, isNotNull);
      expect(channel.isSuper, isFalse);
      expect(channel.isBroadcast, isFalse);
      expect(channel.isPublic, isFalse);
      expect(channel.isDistinct, isFalse);
      expect(channel.isEphemeral, isFalse);
      expect(channel.isStrict, isFalse);
    });

    test('should create a new channel with given image url info', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      GroupChannel? channel;

      var params = GroupChannelParams();
      params.coverImage =
          FileInfo.fromUrl(name: 'Name', url: imageUrl, mimeType: 'image/jpeg');

      try {
        channel = await createGroupChannel(params);
      } catch (e) {
        TestFailure('Exception: $e');
      }
      expect(channel, isNotNull);
      expect(channel?.coverUrl, equals(imageUrl));
    });

    test('should create a new channel with given image data', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      GroupChannel channel;
      var sentBytes = 0;
      var totalBytes = 0;

      final imageFile = await getAssetFrom('resources/image.jpeg');
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..coverImage = FileInfo.fromData(
          file: imageFile,
          name: 'Cover Image',
          mimeType: 'image/jpeg',
        );

      channel =
          await GroupChannel.createChannel(params, progress: (sent, total) {
        sentBytes = sent;
        totalBytes = total;
      });

      expect(channel, isNotNull);
      expect(sentBytes, equals(totalBytes));
      expect(channel.coverUrl, isNotEmpty);
      expect(sentBytes != 0, true);
    });
  });

  group('groupchannel update', () {
    test('should update group channel with changed parameters', () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..name = 'channel'
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId]
        ..isPublic = true;

      final channel = await createGroupChannel(params);

      final updateParams = GroupChannelParams(isUpdate: true)
        ..name = 'new channel';

      await channel.updateChannel(updateParams);

      expect(updateParams.toJson().length, equals(2));
      expect(channel.name, equals('new channel'));
      expect(channel.isPublic, isTrue);
      expect(channel.isBroadcast, isFalse);
      expect(channel.isSuper, isFalse);
      expect(channel.isDistinct, isFalse);
      expect(channel.isEphemeral, isFalse);
      expect(channel.isDiscoverable, isTrue);
      expect(channel.isStrict, isFalse);
    });

    test('should change given channel with Image Data', () async {
      final user = await testApiCreateUserAndConnect();
      final users = await testApiCreateUsers(2);
      final imageFile = await getAssetFrom('resources/image.jpeg');

      final params = GroupChannelParams()
        ..name = 'channel'
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];

      final channel = await createGroupChannel(params);

      expect(channel, isNotNull);
      expect(channel.coverUrl, isNotNull);
      expect(channel.name, 'channel');

      final event2 = ChannelEventListener();

      final updatedParams = GroupChannelParams()
        ..name = 'update name'
        ..channelUrl = channel.channelUrl
        ..operatorUserIds = users.map((e) => e.userId).toList()
        ..coverImage = FileInfo.fromData(
          file: imageFile,
          name: 'Cover Image',
          mimeType: 'image/jpeg',
        );

      var sentBytes = 0;
      var totalBytes = 0;

      channel.updateChannel(updatedParams, progress: (sent, total) {
        sentBytes = sent;
        totalBytes = total;
      });

      final results2 = await event2.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onChannelOperatorsUpdated,
      ]);

      expect(results2, isNotNull);
      expect(results2.length, equals(2));
      expect(sentBytes, equals(totalBytes));
      expect(sentBytes != 0, isTrue);
      expect(channel.name, equals('update name'));
    });

    test('should change given channel with image url', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      final params = GroupChannelParams()
        ..name = 'the name'
        ..operatorUserIds = [user.userId];

      final channel = await createGroupChannel(params);
      expect(channel, isNotNull);

      var newParams = GroupChannelParams()
        ..name = 'new name'
        ..coverImage = FileInfo.fromUrl(
            name: 'Name', url: imageUrl, mimeType: 'image/jpeg');

      await channel.updateChannel(newParams);
      expect(channel.name, 'new name');
      expect(channel.coverUrl, imageUrl);
    });

    test('should not update operatorIds if not set', () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..name = 'the name'
        ..operatorUserIds = [user.userId];

      final channel = await createGroupChannel(params);

      final newParams = GroupChannelParams.withChannel(channel);
      newParams.name = 'new name';

      await channel.updateChannel(newParams);
      expect(channel.name, 'new name');
      expect(channel.members.first.role, Role.chat_operator);
    });

    test('should throw an exception with invalid channel url', () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..name = 'the name'
        ..operatorUserIds = [user.userId];

      final channel = await createGroupChannel(params);

      params
        ..channelUrl = 'some_fake_url'
        ..name = 'update name';

      expect(() => channel.updateChannel(params),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel delete', () {
    test('should delete given channel', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      var params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);
      expect(channel, isNotNull);

      await channel.deleteChannel();
      //channel should not be valid anymore
    });
  });

  group('groupchannel get', () {
    test('should return a channel with given channel url', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      var params = GroupChannelParams();
      final channel1 = await createGroupChannel(params);
      expect(channel1, isNotNull);

      final channel2 = await GroupChannel.getChannel(channel1.channelUrl);
      expect(channel2, isNotNull);

      expect(channel1.channelUrl, equals(channel2.channelUrl));
    });
  });
}
