import 'package:sendbird_sdk/request/feature/group_channel_read_request.dart';
import 'package:sendbird_sdk/request/reaction/channel_reaction_add_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:sendbird_sdk/utils/extensions.dart';
import 'package:test/test.dart';
import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/helpers/private_api_client.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late User user;
  late List<User> users;
  late ChannelEventListener event;
  late GroupChannel channel;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    users = await testApiCreateUsers(4);
    user = await testApiConnect(users.first.userId);
    event = ChannelEventListener();
    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..isPublic = true
      ..isDiscoverable = true
      ..userIds = users.map((e) => e.userId).toList();

    channel = await createGroupChannel(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('groupchannel refresh', () {
    test('should return up to dated channel object', () async {
      channel.sendUserMessageWithText('test');
      await event.waitFor([EventFuncName.onChannelChanged]);

      final updated = await GroupChannel.refresh(channel.channelUrl);
      expect(updated.lastMessage, isA<UserMessage>());
    });
  });

  group('groupchannel add operators', () {
    test('should success to add valid operators', () async {
      channel.addOperators(users.map((e) => e.userId).toList());
      await event.waitFor([EventFuncName.onChannelOperatorsUpdated]);
      channel.members.forEach((e) {
        expect(e.role, Role.chat_operator);
      });
    });

    test(
      'should throw an exception when userIds is empty',
      () async {
        expect(() => channel.addOperators([]),
            throwsA(isA<InvalidParameterError>()));
      },
    );
  });

  group('groupchannel remove operators', () {
    test('should success to remove valid operators', () async {
      channel.addOperators(users.map((e) => e.userId).toList());
      await event.waitFor([EventFuncName.onChannelOperatorsUpdated]);
      channel.removeOperators([users.last.userId]);
      await event.waitFor([EventFuncName.onChannelOperatorsUpdated]);
      final nonOperator = channel.members
          .firstWhere((element) => element.userId == users.last.userId);
      expect(nonOperator, isNotNull);
      expect(nonOperator.role, Role.none);
    });

    test('should throw an exception with invalid operators', () async {
      expect(() => channel.removeOperators([]), throwsA(isA<SBError>()));
    });
  });

  group('groupchannel remove all operators', () {
    test('should success to remove all operators', () async {
      channel.addOperators(users.map((e) => e.userId).toList());
      await event.waitFor([EventFuncName.onChannelOperatorsUpdated]);
      channel.removeAllOperators();
      await event.waitFor([EventFuncName.onChannelOperatorsUpdated]);
      channel.members.forEach((e) {
        expect(e.role, Role.none);
      });
    });
  });

  group('groupchannel get message list', () {
    var messageId = 0;
    var msgId = 0;
    var timestamp = 0;
    setUp(() async {
      final uParams = UserMessageParams(message: 'test');
      final fParams = FileMessageParams.withUrl(imageUrl, size: 123)
        ..data = 'file data';

      final message = await sendUserMessageViaAPI(
        channel: channel,
        params: uParams,
        senderId: user.userId,
      );
      messageId = message.messageId;

      await sendFileMessageViaAPI(
        channel: channel,
        params: fParams,
        senderId: users.last.userId,
      );

      final msg1 = await sendUserMessageViaAPI(
        channel: channel,
        params: uParams,
        senderId: user.userId,
      );
      timestamp = msg1.createdAt;

      final msg = await sendUserMessageViaAPI(
        channel: channel,
        params: uParams,
        senderId: user.userId,
      );
      msgId = msg.messageId;
      await sendFileMessageViaAPI(
        channel: channel,
        params: fParams,
        senderId: users.last.userId,
      );
    });

    test(
        'should return reply message with replyToChannel option to true and parentMessage Information',
        () async {
      final reply = await sendUserMessageViaAPI(
        channel: channel,
        params: UserMessageParams(message: 'reply')
          ..replyToChannel = true
          ..parentMessageId = messageId,
        senderId: user.userId,
      );
      final params = MessageListParams()
        ..isInclusive = true
        ..nextResultSize = 0
        ..previousResultSize = 10
        ..replyType = ReplyType.all
        ..includeParentMessageInfo = true;

      final result = await channel.getMessagesById(reply.messageId, params);

      expect(result[5].replyToChannel, isTrue);
      expect(result[5].message, equals('reply'));
      expect(result[5].parentMessage, isNotNull);
      expect(result[5].parentMessage, isA<UserMessage>());
    });

    test('should return mix of message types at given timestamp', () async {
      final params = MessageListParams()
        ..isInclusive = true
        ..nextResultSize = 0
        ..previousResultSize = 10;

      final result = await channel.getMessagesByTimestamp(timestamp, params);
      expect(result.length, equals(3));
      //check all properties
      expect(result.first, isA<UserMessage>());
      final u1 = result.first as UserMessage;
      expect(u1.message, equals('test'));
      expect(result[1], isA<FileMessage>());
      final u2 = result[1] as FileMessage;
      expect(u2.url, equals(imageUrl));
      expect(u2.size, equals(123));
      expect(result.last, isA<UserMessage>());
    });

    test('should throw an exception with given invalid timestamp', () async {
      final params = MessageListParams()
        ..isInclusive = true
        ..nextResultSize = 0
        ..previousResultSize = 10;

      expect(
        () => channel.getMessagesByTimestamp(-1, params),
        throwsA(isA<SBError>()),
      );
    });

    test('should return mix of message types at given message id', () async {
      final params = MessageListParams()
        ..isInclusive = true
        ..nextResultSize = 10
        ..previousResultSize = 0;

      final result = await channel.getMessagesById(msgId, params);
      expect(result.length, equals(2));
      expect(result.first, isA<UserMessage>());
      expect(result.last, isA<FileMessage>());
    });

    test('should throw an exception with given invalid message id', () async {
      final params = MessageListParams()
        ..isInclusive = true
        ..nextResultSize = 0
        ..previousResultSize = 10;

      expect(
        () => channel.getMessagesById(-1, params),
        throwsA(isA<SBError>()),
      );
    });
  });

  group('groupchannel get a message', () {
    test('should return message with reactions', () async {
      final message = await sendUserMessageViaAPI(
        channel: channel,
        params: UserMessageParams(message: 'test'),
        senderId: user.userId,
      );

      await PrivateApiClient().send<ReactionEvent>(
        ChannelReactionAddRequest(
            channelType: channel.channelType,
            channelUrl: channel.channelUrl,
            messageId: message.messageId,
            userId: users.first.userId,
            key: 'Happy'),
      );

      await PrivateApiClient().send<ReactionEvent>(
        ChannelReactionAddRequest(
            channelType: channel.channelType,
            channelUrl: channel.channelUrl,
            messageId: message.messageId,
            userId: users[1].userId,
            key: 'Happy'),
      );

      final params = MessageRetrievalParams(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: message.messageId,
      )..includeReactions = true;

      final result = await BaseMessage.getMessage(params);
      expect(result.reactions?.length, isNot(equals(0)));
    });

    test('should return valid message', () async {
      final message = await sendUserMessageViaAPI(
        channel: channel,
        params: UserMessageParams(message: 'test'),
        senderId: user.userId,
      );

      final params = MessageRetrievalParams(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: message.messageId,
      );

      final result = await BaseMessage.getMessage(params);
      expect(result.replyToChannel, isFalse);
      expect(result.parentMessage, isNull);
      expect(result.messageId, equals(message.messageId));
      expect(result.message, equals(message.message));
      expect(result, isA<UserMessage>());
    });

    test('should throw an exception with invalid params', () {
      final params = MessageRetrievalParams(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: -1,
      );

      expect(() => BaseMessage.getMessage(params),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel reset my history', () {
    test('should reset chat messages at given time', () async {
      final uParams = UserMessageParams(message: 'test');
      final fParams = FileMessageParams.withUrl(imageUrl, size: 123);

      await sendUserMessageViaAPI(
        channel: channel,
        params: uParams,
        senderId: user.userId,
      );

      await sendFileMessageViaAPI(
        channel: channel,
        params: fParams,
        senderId: users.last.userId,
      );

      await sendUserMessageViaAPI(
        channel: channel,
        params: uParams,
        senderId: user.userId,
      );

      final params = MessageListParams()
        ..isInclusive = true
        ..previousResultSize = 10
        ..nextResultSize = 0;

      var result = await channel.getMessagesByTimestamp(
        ExtendedInteger.max,
        params,
      );
      expect(result.length, equals(3));

      await channel.resetMyHistory();

      result = await channel.getMessagesByTimestamp(
        ExtendedInteger.max,
        params,
      );
      expect(result.length, equals(0));
    });
  });

  group('groupchannel join', () {
    test('should able to join a given channel', () async {
      final extraUser = (await testApiCreateUsers(1)).first;
      await testApiConnect(extraUser.userId);
      channel.join();
      await event.waitFor([EventFuncName.onUserJoined]);
      expect(channel.joinedAt, isNot(equals(0)));
    });
  });

  group('groupchannel leave', () {
    test('should able to leave a given joined channel', () async {
      channel.leave();
      await event.waitFor([EventFuncName.onUserLeaved]);
      expect(channel.invitedAt, equals(0));
      expect(channel.joinedAt, equals(0));
    });
  });

  group('groupchannel inviteUsers', () {
    test(
      'should able to inviteUsers when inviteUsers exists',
      () async {
        expect(channel.memberCount, equals(4));
        var testUsers = await testApiCreateUsers(2);
        await channel.inviteUsers(testUsers.map((e) => e.userId).toList());
        await Future.delayed(Duration(seconds: 1));
        expect(channel.memberCount, equals(6));
      },
    );
    test(
      'should throw an exception when userIds is empty',
      () async {
        var emptyUsers = <User>[];
        expect(
            () => channel.inviteUsers(emptyUsers.map((e) => e.userId).toList()),
            throwsA(isA<InvalidParameterError>()));
      },
    );
  });

  group('groupchannel getMember', () {
    test('should return user object with given joined user id', () async {
      expect(channel.getMember(users[2].userId), isNotNull);
      expect(channel.getMember(user.userId), isNotNull);
    });

    test('should return null with not joined user id', () async {
      final extraUser = (await testApiCreateUsers(1)).first;
      expect(channel.getMember(extraUser.userId), isNull);
    });
  });

  group('groupchannel get read members', () {
    test('should return 0 read members', () async {
      channel.sendUserMessageWithText('test');
      await event.waitFor([EventFuncName.onChannelChanged]);
      expect(
        channel.getReadMembers(channel.lastMessage!).length,
        equals(0),
      );
    });

    test('should return correct read members', () async {
      channel.sendUserMessageWithText('test');
      await event.waitFor([EventFuncName.onChannelChanged]);

      PrivateApiClient().send(GroupChannelMarkAsReadApiRequest(
        channelUrls: [channel.channelUrl],
        userId: users[1].userId,
      ));
      await event.waitFor([EventFuncName.onReadReceiptUpdated]);

      PrivateApiClient().send(GroupChannelMarkAsReadApiRequest(
        channelUrls: [channel.channelUrl],
        userId: users[3].userId,
      ));
      await event.waitFor([EventFuncName.onReadReceiptUpdated]);

      final readMembers = channel
          .getReadMembers(channel.lastMessage!)
          .map((e) => e.userId)
          .toList();
      expect(readMembers.length, equals(2));
      expect(readMembers.contains(users[1].userId), isTrue);
      expect(readMembers.contains(users[3].userId), isTrue);
    });
  });

  group('groupchannel get unread members', () {
    test('should return all unread members', () async {
      channel.sendUserMessageWithText('test');
      await event.waitFor([EventFuncName.onChannelChanged]);

      expect(
        channel.getUnreadMembers(channel.lastMessage!).length,
        equals(3),
      );
    });

    test('should return 0 unread members', () async {
      channel.sendUserMessageWithText('test');
      await event.waitFor([EventFuncName.onChannelChanged]);

      PrivateApiClient().send(GroupChannelMarkAsReadApiRequest(
        channelUrls: [channel.channelUrl],
        userId: users[1].userId,
      ));
      await event.waitFor([EventFuncName.onReadReceiptUpdated]);

      PrivateApiClient().send(GroupChannelMarkAsReadApiRequest(
        channelUrls: [channel.channelUrl],
        userId: users[3].userId,
      ));
      await event.waitFor([EventFuncName.onReadReceiptUpdated]);

      final readMembers = channel.getUnreadMembers(channel.lastMessage!);
      expect(readMembers.length, equals(1));
      expect(readMembers.first.userId, equals(users[2].userId));
    });
  });

  group('groupchannel get read status', () {
    test('should return all read members status', () async {
      channel.sendUserMessageWithText('test');
      await event.waitFor([EventFuncName.onChannelChanged]);

      final statuses = channel.getReadStatus(false);
      expect(statuses.length, equals(3));
    });
  });

  group('groupchannel my push trigger option', () {
    test('should succeed to set push trigger option', () async {
      channel.setMyPushTriggerOption(GroupChannelPushTriggerOption.all);
      await event.waitFor([EventFuncName.onChannelChanged]);
      expect(
        channel.myPushTriggerOption,
        equals(GroupChannelPushTriggerOption.all),
      );

      channel.setMyPushTriggerOption(GroupChannelPushTriggerOption.mentionOnly);
      await event.waitFor([EventFuncName.onChannelChanged]);
      expect(
        channel.myPushTriggerOption,
        equals(GroupChannelPushTriggerOption.mentionOnly),
      );
    });

    test('should succeed to get my push trigger option', () async {
      channel.setMyPushTriggerOption(GroupChannelPushTriggerOption.off);
      await event.waitFor([EventFuncName.onChannelChanged]);
      final result = await channel.getMyPushTriggerOption();
      expect(result, equals(GroupChannelPushTriggerOption.off));
    });
  });

  group('groupchannel set my count preference', () {
    test('should succeed with valid params', () async {
      channel.setMyCountPreference(CountPreference.mentionOnly);
      await event.waitFor([EventFuncName.onChannelChanged]);
      expect(channel.myCountPreference, equals(CountPreference.mentionOnly));

      channel.setMyCountPreference(CountPreference.messageOnly);
      await event.waitFor([EventFuncName.onChannelChanged]);
      expect(channel.myCountPreference, equals(CountPreference.messageOnly));
    });
  });

  group('groupchannel notify screenshot take', () {
    test('should succeed normally', () async {
      await channel.notifyScreenshotWasTaken();
    });
  });
}
