import 'package:test/test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late List<User> users;
  late User user;
  late ChannelEventListener event;
  late GroupChannel channel;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    users = await testApiCreateUsers(2);
    user = await testApiConnect(users.first.userId);
    event = ChannelEventListener();
    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = users.map((e) => e.userId).toList();
    //TODO: flaky when waiting events
    channel = await createGroupChannel(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('groupchannel ban user', () {
    test('should success given valid user', () async {
      channel.banUser(userId: users.last.userId);

      final result = await event.waitFor([
        EventFuncName.onUserBanned,
        EventFuncName.onUserLeaved,
      ]);

      expect(result.length, equals(2));
    });
    test('should throw an exception when userId is empty', () async {
      expect(() => channel.banUser(userId: ''),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with given invalid user', () async {
      expect(() => channel.banUser(userId: 'fake_id'), throwsA(isA<SBError>()));
    });
  });

  group('groupchannel unban user', () {
    test('should success given banned valid user', () async {
      channel.banUser(userId: users.last.userId);

      var result = await event.waitFor([
        EventFuncName.onUserBanned,
        EventFuncName.onUserLeaved,
      ]);
      expect(result.length, equals(2));

      channel.unbanUser(userId: users.last.userId);
      result = await event.waitFor([
        EventFuncName.onUserUnbanned,
      ]);
      expect(result.length, equals(1));
    });

    test('should fthrow an exceptionail when userId is empty', () async {
      expect(() => channel.unbanUser(userId: ''),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with given not banned user', () async {
      expect(() => channel.unbanUser(userId: users.last.userId),
          throwsA(isA<SBError>()));
    });

    test('should throw an exception with given invalid user', () async {
      expect(
        () => channel.unbanUser(userId: 'fake_id'),
        throwsA(isA<SBError>()),
      );
    });
  });

  group('groupchannel mute user', () {
    test('should success to mute valid user', () async {
      channel.muteUser(userId: users.last.userId);
      final result = await event.waitFor([EventFuncName.onUserMuted]);
      expect(result.length, equals(1));
    });

    test('should throw an exception when userId is empty', () async {
      expect(() => channel.muteUser(userId: ''),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with given invalid user', () async {
      expect(
          () => channel.muteUser(userId: 'fake_id'), throwsA(isA<SBError>()));
    });
  });

  group('groupchannel unmute user', () {
    test('should success to unmute valid user', () async {
      channel.muteUser(userId: users.last.userId);
      var result = await event.waitFor([EventFuncName.onUserMuted]);
      expect(result.length, equals(1));

      channel.unmuteUser(userId: users.last.userId);
      result = await event.waitFor([EventFuncName.onUserUnmuted]);
      expect(result.length, equals(1));
    });

    test('should throw an exception with given not muted user', () async {
      expect(() => channel.unmuteUser(userId: users.last.userId),
          throwsA(isA<SBError>()));
    });

    test('should throw an exception when userId is empty', () async {
      expect(() => channel.unmuteUser(userId: ''),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with given invalid user', () async {
      expect(
          () => channel.unmuteUser(userId: 'fake_id'), throwsA(isA<SBError>()));
    });
  });

  group('groupchannel get my mute info', () {
    test('shuold return value when umuted state', () async {
      final res = await channel.getMyMuteInfo();
      expect(res.isMuted, isFalse);
      expect(res.description, '');
      expect(res.endAt, 0);
      expect(res.startAt, 0);
      expect(res.remainingDuration, 0);
    });

    test('shuold return value when muted state', () async {
      await channel.muteUser(
        userId: user.userId,
        description: 'mute you',
        seconds: 100,
      );

      final res = await channel.getMyMuteInfo();
      expect(res.isMuted, isTrue);
      expect(res.description, 'mute you');
      expect(res.endAt, isNot(equals(0)));
      expect(res.startAt, isNot(equals(0)));
      expect(res.remainingDuration, isNot(equals(0)));
    });
  });

  group('groupchannel freeze channel', () {
    test('should success to freeze valid channel', () async {
      channel.freeze();

      final result = await event.waitFor([EventFuncName.onChannelFrozen]);
      expect(result.length, equals(1));
      expect(channel.isFrozen, isTrue);
    });

    test('should throw an exception with invalid channel', () async {
      final fakeChannel = GroupChannel(channelUrl: 'channelurl');
      expect(() => fakeChannel.freeze(), throwsA(isA<SBError>()));
    });

    test('should throw an exception on non-operator attemps', () async {
      await sdk.disconnect();

      await testApiConnect(users.last.userId);
      expect(() => channel.freeze(), throwsA(isA<SBError>()));
    });
  });

  group('groupchannel unfreeze channel', () {
    test('should success to unfreeze given channel', () async {
      channel.freeze();
      var result = await event.waitFor([EventFuncName.onChannelFrozen]);
      expect(result.length, equals(1));
      expect(channel.isFrozen, isTrue);

      channel.unfreeze();
      result = await event.waitFor([EventFuncName.onChannelUnfrozen]);
      expect(result.length, equals(1));
      expect(channel.isFrozen, isFalse);
    });

    test('should do nothing with given unfrozen channel', () async {
      await channel.unfreeze();
    });
  });

  group('groupchannel hide channel', () {
    test('should success to hide with given channel', () async {
      channel.hideChannel();
      final result = await event.waitFor([EventFuncName.onChannelHidden]);
      expect(result.length, equals(1));
      expect(channel.isHidden, isTrue);
    });

    test('should success when hidePreviousMessage set to true', () async {
      channel.sendUserMessage(UserMessageParams(message: 'test'));
      channel.hideChannel(hidePreviousMessage: true);
      await event.waitFor([EventFuncName.onChannelHidden]);
      expect(channel.unreadMentionCount, equals(0));
      expect(channel.unreadMessageCount, equals(0));
    });

    // test('should throw an exception on non-operator attemps', () async {
    //   await sdk.disconnect();

    //   await testApiConnect(users.last.userId);
    //   expect(() => channel.hideChannel(), throwsA(isA<SBError>()));
    // });
  });

  group('groupchannel unhide channel', () {
    test('should success with given hidden channel', () async {
      channel.hideChannel();
      var result = await event.waitFor([EventFuncName.onChannelHidden]);
      expect(result.length, equals(1));
      expect(channel.isHidden, isTrue);

      channel.unhideChannel();
      result = await event.waitFor([EventFuncName.onChannelChanged]);
      expect(result.length, equals(1));
      expect(channel.isHidden, isFalse);
    });

    test('should success by sending message if autounhide is activated',
        () async {
      channel.hideChannel();
      var result = await event.waitFor([EventFuncName.onChannelHidden]);
      expect(result.length, equals(1));
      expect(channel.isHidden, isTrue);

      channel.sendUserMessageWithText('text', onCompleted: null);
      result = await event.waitFor([EventFuncName.onChannelChanged]);
      expect(result.length, equals(1));
      expect(channel.isHidden, isFalse);
      expect(channel.hiddenState, GroupChannelHiddenState.unhidden);
    });
  });
}
