import 'dart:async';
import 'dart:ui';

import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../../common/callback.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';
import '../../../common/utils.dart';
import '../../../common/common.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late ChannelEventListener event;
  late GroupChannel channel;
  late FileMessage message;
  late User user;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    user = await testApiCreateUserAndConnect();
    event = ChannelEventListener();

    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = [user.userId];
    channel = await createGroupChannel(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('groupchannel send file message', () {
    test('should succeed with replyToChannel True', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      final mParams = FileMessageParams.withFile(imageFile)
        ..thumbnailSizes = [
          Size(10, 10),
          Size(20, 20),
        ]
        ..replyToChannel = true;

      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.url, isNotNull);
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
          expect(message.thumbnails, isNotNull);
          expect(message.thumbnails?.length, equals(2));
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('should succeed with normal file', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      final mParams = FileMessageParams.withFile(imageFile)
        ..thumbnailSizes = [
          Size(10, 10),
          Size(20, 20),
        ];

      final completor = Completer();

      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) {
          expect(message, isNotNull);
          expect(message.secureUrl, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.url, isNotNull);
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
          expect(message.thumbnails, isNotNull);
          expect(message.thumbnails?.length, equals(2));
          completor.complete();
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);
      await completor.future;

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('should succeed to send file message via api with thumbnails',
        () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      final params = FileMessageParams.withFile(imageFile)
        ..thumbnailSizes = [
          Size(10, 10),
          Size(20, 20),
        ];

      final wait = Completer();
      channel.sendFileMessage(params, onCompleted: (msg, err) {
        expect(msg.url, isNotEmpty);
        expect(msg.thumbnails, isNotEmpty);
        expect(msg.thumbnails?.first, isA<Thumbnail>());
        wait.complete();
      });

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.paused);
      await wait.future;
    });

    test('should return file message object to json ', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      final mParams = FileMessageParams.withFile(imageFile)
        ..thumbnailSizes = [
          Size(10, 10),
          Size(20, 20),
        ];

      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) {
          final json = message.toJson();
          expect(json['type'], equals('image/jpeg'));
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('should succeed with url params', () async {
      final mParams = FileMessageParams.withUrl(imageUrl,
          size: 123, mimeType: 'image/jpeg');

      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.url, equals(imageUrl));
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);
      message =
          (results.first.data.first as GroupChannel).lastMessage as FileMessage;

      expect(results, isNotNull);
      expect(results.length, equals(1));
      expect(message.type, equals('image/jpeg'));
      expect(message.url, equals(imageUrl));
      expect(message.size, equals(123));
    });
  });

  group('groupchannel update file message', () {
    test('should succeed with given valid file message', () async {
      final mParams = FileMessageParams.withUrl(imageUrl, size: 123);

      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.url, equals(imageUrl));
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      var results = await event.waitFor([EventFuncName.onChannelChanged]);
      expect(results, isNotNull);
      expect(results.length, equals(1));
      final msg = (results.first.data.first as GroupChannel).lastMessage!;

      mParams
        ..data = 'update data'
        ..customType = 'update custom';

      channel.updateFileMessage(msg.messageId, mParams);

      final res = await event.waitFor([EventFuncName.onMessageUpdated]);
      final msg2 = res.first.data.last as FileMessage;
      expect(msg2.data, equals('update data'));
      expect(msg2.customType, equals('update custom'));
    });

    test('should throw an exception with given invalid file message id',
        () async {
      final mParams = FileMessageParams.withUrl(imageUrl, size: 123);

      expect(() => channel.updateFileMessage(0, mParams),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with invalid params', () async {
      final mParams = FileMessageParams.withUrl(imageUrl, size: 123);
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.url, equals(imageUrl));
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      await event.waitFor([EventFuncName.onChannelChanged]);

      expect(() => channel.updateFileMessage(-1, mParams),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel resend file message', () {
    final mParams = FileMessageParams.withUrl(imageUrl, size: 123);

    test('should succeed with failed file message', () async {
      await sdk.disconnect();

      final message = FileMessage.fromParams(params: mParams, channel: channel);
      message
        ..sendingStatus = MessageSendingStatus.failed
        ..errorCode = ErrorCode.connectionRequired;

      event = ChannelEventListener();
      await testApiConnect(user.userId);
      channel = await GroupChannel.getChannel(channel.channelUrl);
      channel.resendFileMessage(message,
          params: mParams, onCompleted: (msg, err) {});
      await event.waitFor([EventFuncName.onChannelChanged]);
    });

    test('should throw an exception with already sent file message', () async {
      final pending = channel.sendFileMessage(mParams);
      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
      message =
          (results.first.data.first as GroupChannel).lastMessage as FileMessage;

      expect(() => channel.resendFileMessage(message, params: mParams),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with not resendable message', () async {
      final fake = FileMessage(
        messageId: 123,
        url: 'url',
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      );

      final fake2 = FileMessage(
        channelType: channel.channelType,
        channelUrl: '123',
        url: 'url',
        messageId: 1231,
        sendingStatus: MessageSendingStatus.failed,
      );

      final fake3 = FileMessage(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        url: 'url',
        messageId: 1234,
        sendingStatus: MessageSendingStatus.failed,
        errorCode: ErrorCode.invalidInitialization,
      );

      expect(() => channel.resendFileMessage(fake, params: mParams),
          throwsA(isA<InvalidParameterError>()));
      expect(() => channel.resendFileMessage(fake2, params: mParams),
          throwsA(isA<InvalidParameterError>()));
      expect(() => channel.resendFileMessage(fake3, params: mParams),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel copy file message', () {
    final mParams = FileMessageParams.withUrl(imageUrl, size: 123);

    test('should succeed with given other channel', () async {
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (msg, error) {
          message = msg;
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);

      expect(results, isNotNull);
      expect(results.length, equals(1));

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel2 = await createGroupChannel(params);

      final wait = Completer();
      channel.copyMessage(message, channel2, onCompleted: (res, err) {
        final copied = res as FileMessage;
        expect(copied, isA<FileMessage>());
        expect(err, isNull);
        expect(copied, isNotNull);
        expect(message.url, equals(copied.url));
        expect(message.sender?.userId, equals(copied.sender?.userId));
        wait.complete();
      });
      await event.waitFor([EventFuncName.onChannelChanged]);
      await wait.future;
    });

    test('should succeed with given same channel', () async {
      final pending = channel.sendFileMessage(mParams);
      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);
      expect(results, isNotNull);
      expect(results.length, equals(1));
      final message =
          (results.first.data.first as GroupChannel).lastMessage as FileMessage;

      final wait = Completer();
      channel.copyMessage(message, channel, onCompleted: (res, err) {
        final copied = res as FileMessage;
        expect(copied, isA<FileMessage>());
        expect(err, isNull);
        expect(copied, isNotNull);
        expect(message.url, equals(copied.url));
        expect(message.sender?.userId, equals(copied.sender?.userId));
        wait.complete();
      });

      await event.waitFor([EventFuncName.onChannelChanged]);
      await wait.future;
    });

    test('should throw an exception with other channel message', () async {
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (msg, error) {
          message = msg;
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);

      expect(results, isNotNull);
      expect(results.length, equals(1));

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel2 = await createGroupChannel(params);

      expect(() => channel2.copyMessage(message, channel),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel file message cancel', () {
    test('should succeed properly', () async {
      final file = await getAssetFrom('resources/video.mp4');
      final params = FileMessageParams.withFile(file);

      final completer = Completer();
      final pending = channel.sendFileMessage(
        params,
        onCompleted: (message, error) {
          expect(error, isNotNull);
          expect(error, isA<OperationCancelError>());
          completer.complete();
        },
      );
      await Future.delayed(Duration(seconds: 1));
      final res = channel.cancelUploadingFileMessage(pending.requestId!);
      await completer.future;
      expect(res, isTrue);
    });
  });
}
