import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/request/messages/message_delete_request.dart';
import 'package:sendbird_sdk/request/messages/user_message_send_request.dart';
import 'package:sendbird_sdk/request/messages/user_message_update_request.dart';
import 'package:sendbird_sdk/request/reaction/channel_reaction_add_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/common.dart';
import '../../../common/helpers/private_api_client.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late GroupChannel channel;
  late List<User> users;

  setUpAll(() async {
    testInfo = await testCreateApp();
    users = await testApiCreateUsers(3);
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    final params = GroupChannelParams()
      ..operatorUserIds = [users.first.userId]
      ..userIds = users.map((e) => e.userId).toList();
    channel = await createGroupChannelViaAPI(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('groupchannel message change log', () {
    test('should return without including timestamp and reply message',
        () async {
      final params = UserMessageParams(message: 'message');

      final msg1 = await PrivateApiClient()
          .send<UserMessage>(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params,
        senderId: users.last.userId,
      ));

      final msg1Reply =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: UserMessageParams(message: 'reply')
          ..parentMessageId = msg1.messageId
          ..replyToChannel = true,
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelUserMessageUpdateApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg1Reply.messageId,
        params: params..message = 'reply-update',
        senderId: users.last.userId,
      ));

      await testApiConnect(users.first.userId);

      final logParams = MessageChangeLogParams()
        ..replyType = ReplyType.none
        ..includeParentMessageInfo = true;
      final res = await channel.getMessageChangeLogs(
        params: logParams,
      );
      expect(res, isNotNull);
      expect(res.updatedMessages.length, equals(0));
    });

    test('should return without including reply message', () async {
      final params = UserMessageParams(message: 'message');
      final timestamp = DateTime.now().millisecondsSinceEpoch;

      final msg1 = await PrivateApiClient()
          .send<UserMessage>(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params,
        senderId: users.last.userId,
      ));

      final msg1Reply =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: UserMessageParams(message: 'reply')
          ..parentMessageId = msg1.messageId
          ..replyToChannel = true,
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelUserMessageUpdateApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg1Reply.messageId,
        params: params..message = 'reply-update',
        senderId: users.last.userId,
      ));

      await testApiConnect(users.first.userId);

      final logParams = MessageChangeLogParams()
        ..replyType = ReplyType.none
        ..includeParentMessageInfo = true;
      final res = await channel.getMessageChangeLogs(
        timestamp: timestamp,
        params: logParams,
      );
      expect(res.updatedMessages.length, equals(0));
    });

    test(
        'should return updated message with parent message information and replyToChannel option',
        () async {
      final params = UserMessageParams(message: 'message');
      final timestamp = DateTime.now().millisecondsSinceEpoch;

      final msg1 = await PrivateApiClient()
          .send<UserMessage>(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params,
        senderId: users.last.userId,
      ));

      final msg1Reply =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: UserMessageParams(message: 'reply')
          ..parentMessageId = msg1.messageId
          ..replyToChannel = true,
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelUserMessageUpdateApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg1Reply.messageId,
        params: params..message = 'reply-update',
        senderId: users.last.userId,
      ));

      await testApiConnect(users.first.userId);

      final logParams = MessageChangeLogParams()
        ..replyType = ReplyType.all
        ..includeParentMessageInfo = true;
      final res = await channel.getMessageChangeLogs(
        timestamp: timestamp,
        params: logParams,
      );

      expect(res.updatedMessages[0], isNotNull);
      expect(res.updatedMessages[0].replyToChannel, isTrue);
      expect(res.updatedMessages[0].parentMessage, isNotNull);
      expect(res.updatedMessages[0].message, equals('reply-update'));
    });

    test('should return proper # of log', () async {
      final params = UserMessageParams(message: 'message');
      final params2 = UserMessageParams(message: 'message2');
      final params3 = UserMessageParams(message: 'message3');
      final timestamp = DateTime.now().millisecondsSinceEpoch;

      final msg1 =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params,
        senderId: users.last.userId,
      ));

      final msg2 =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params2,
        senderId: users.last.userId,
      ));

      final msg3 =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params3,
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelUserMessageUpdateApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg1.messageId,
        params: params..message = 'message1-1',
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelUserMessageUpdateApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg2.messageId,
        params: params2..message = 'message2-1',
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelMessageDeleteRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg3.messageId,
      ));

      await testApiConnect(users.first.userId);

      final logParams = MessageChangeLogParams();
      final res = await channel.getMessageChangeLogs(
        timestamp: timestamp,
        params: logParams,
      );

      expect(res, isNotNull);
      expect(res.deletedMessageIds.length, equals(1));
      expect(res.updatedMessages.length, equals(2));
      expect(res.hasMore, isFalse);
    });

    test('should return with relevent properties', () async {
      final params = UserMessageParams(message: 'message');
      final params2 = UserMessageParams(message: 'message2')
        ..metaArrays = [
          MessageMetaArray(key: 'key1', value: ['value1'])
        ];
      final params3 = UserMessageParams(message: 'message3');
      final timestamp = DateTime.now().millisecondsSinceEpoch;

      final msg1 =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params,
        senderId: users.last.userId,
      ));

      final msg2 =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params2,
        senderId: users.last.userId,
      ));

      final msg3 =
          await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        params: params3,
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelUserMessageUpdateApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg1.messageId,
        params: params..message = 'message1-1',
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelReactionAddRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg1.messageId,
        key: 'thumb-up',
        userId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelUserMessageUpdateApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg2.messageId,
        params: params2..message = 'message2-1',
        senderId: users.last.userId,
      ));

      await PrivateApiClient().send(ChannelMessageDeleteRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        messageId: msg3.messageId,
      ));

      await testApiConnect(users.first.userId);

      final logParams = MessageChangeLogParams()
        ..includeMetaArray = true
        ..includeReactions = true;
      final res = await channel.getMessageChangeLogs(
        timestamp: timestamp,
        params: logParams,
      );

      expect(res, isNotNull);
      expect(res.deletedMessageIds.length, equals(1));
      expect(res.updatedMessages.length, equals(2));
      expect(res.hasMore, isFalse);

      expect(res.updatedMessages.first.reactions?.length, equals(1));
      expect(res.updatedMessages.last.metaArrays?.length, equals(1));
    });
  });
}
