import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/core/message/scheduled_file_message.dart';
import 'package:sendbird_sdk/core/message/scheduled_user_message.dart';
import 'package:sendbird_sdk/params/scheduled_file_message_params.dart';
import 'package:sendbird_sdk/params/scheduled_file_message_update_params.dart';
import 'package:sendbird_sdk/params/scheduled_message_total_count_params.dart';
import 'package:sendbird_sdk/params/scheduled_user_message_update_params.dart';
import 'package:sendbird_sdk/query/scheduled_message/scheduled_message_list_query.dart';
import 'package:sendbird_sdk/params/scheduled_message_retrieval_params.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/helpers/private_api_client.dart';
import '../../../common/request.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';
import '../../../common/utils.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late ChannelEventListener event;
  late GroupChannel channel, superGroupChannel;
  late User user;

  setUpAll(() async {
    testInfo = await testCreateApp();

    await PrivateApiClient().send(
      AppAttributeUpdateRequest(
        appId: testInfo.appId,
        attributeName: 'scheduled_message',
        value: {'enabled': true},
      ),
    );
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    user = await testApiCreateUserAndConnect();
    event = ChannelEventListener();
    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = [user.userId];
    final params2 = GroupChannelParams()
      ..isSuper = true
      ..operatorUserIds = [user.userId]
      ..userIds = [user.userId];
    channel = await createGroupChannel(params);
    superGroupChannel = await createGroupChannel(params2);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('groupchannel update scheduled user message', () {
    test(
        'should succeed updating scheduled user message before 10 minutes the message is sent',
        () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isA<ScheduledUserMessage>());
      expect(scheduledMessage.message, equals('scheduled test'));

      final uParams =
          ScheduledUserMessageUpdateParams(message: 'update message');

      final updateScheduledMessage = await channel.updateScheduledUserMessage(
        params: uParams,
        scheduledMessageid: scheduledMessage.messageId,
      );

      expect(updateScheduledMessage, isA<ScheduledUserMessage>());
      expect(updateScheduledMessage.message, equals('update message'));
    });

    test('should succeed updating scheduled_at in 10 minutes', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isA<ScheduledUserMessage>());
      expect(scheduledMessage.message, equals('scheduled test'));

      final uParams = ScheduledUserMessageUpdateParams(
        message: 'update message',
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
      );

      final updateScheduledMessage = await channel.updateScheduledUserMessage(
        params: uParams,
        scheduledMessageid: scheduledMessage.messageId,
      );

      expect(updateScheduledMessage, isA<ScheduledUserMessage>());
      expect(updateScheduledMessage.message, equals('update message'));
    });

    test('should fail updating scheduled_at in 3 minutes', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isA<ScheduledUserMessage>());
      expect(scheduledMessage.message, equals('scheduled test'));

      final uParams = ScheduledUserMessageUpdateParams(
        message: 'update message',
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 180000,
      );

      expect(
          () => channel.updateScheduledUserMessage(
                params: uParams,
                scheduledMessageid: scheduledMessage.messageId,
              ),
          throwsA(isA<BadRequestError>()));
    });

    //TODO fix circleCI to not timeout after 10 min
    // test(
    //     'should fail updating scheduled user message before 3 minutes before the message is sent',
    //     () async {
    //   final mParams = ScheduledUserMessageParams(
    //     scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
    //     message: 'scheduled test',
    //     translationTargetLanguages: ['KR', 'ENG'],
    //   );

    //   final scheduledMessage =
    //       await channel.createScheduledUserMessage(mParams);

    //   expect(scheduledMessage, isA<ScheduledUserMessage>());
    //   expect(scheduledMessage.message, equals('scheduled test'));

    //   await Future.delayed(Duration(minutes: 7));

    //   final uParams =
    //       ScheduledUserMessageUpdateParams(message: 'update message');

    //   expect(
    //       () => channel.updateScheduledUserMessage(
    //             params: uParams,
    //             scheduledMessageid: scheduledMessage.messageId,
    //           ),
    //       throwsA(isA<BadRequestError>()));
    // }, timeout: Timeout.none);

    test(
        'should fail updating scheduled user message with incorrect scheduledMessageId',
        () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isA<ScheduledUserMessage>());
      expect(scheduledMessage.message, equals('scheduled test'));

      final uParams =
          ScheduledUserMessageUpdateParams(message: 'update message');

      expect(
          () => channel.updateScheduledUserMessage(
                params: uParams,
                scheduledMessageid: -1,
              ),
          throwsA(isA<BadRequestError>()));
    });

    test('should succeed updating scheduled file message with correct params',
        () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
      );

      final scheduledMessage =
          await channel.createScheduledFileMessage(mParams);

      final uParams =
          ScheduledFileMessageUpdateParams(fileName: 'update file name');

      final updateScheduledMessage = await channel.updateScheduledFileMessage(
        params: uParams,
        scheduledMessageid: scheduledMessage.messageId,
      );

      expect(updateScheduledMessage, isA<ScheduledFileMessage>());
      expect(updateScheduledMessage.file.name, equals('update file name'));
    });
  });

  group('groupchannel send scheduled user message', () {
    test('should succeed sending 10 minutes from now', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isA<ScheduledUserMessage>());
    });
    test('should fail sending 3 minutes from now', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 180000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      expect(() => channel.createScheduledUserMessage(mParams),
          throwsA(isA<BadRequestError>()));
    });
    test('should fail when user is muted', () async {
      //TODO ServerSide Fix
    });
    test('should fail when user is banned', () async {
      await channel.banUser(userId: user.userId, seconds: 600);

      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      expect(() => channel.createScheduledUserMessage(mParams),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('super groupchannel send scheduled user message', () {
    test('should succeed sending 10 minutes from now', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      final scheduledMessage =
          await superGroupChannel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isA<ScheduledUserMessage>());
    });
    test('should fail sending 3 minutes from now', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 180000,
        message: 'scheduled test',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      expect(() => superGroupChannel.createScheduledUserMessage(mParams),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('groupchannel send scheduled file message', () {
    test('should succeed sending in 10 minutes', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
      );

      final scheduledMessage =
          await channel.createScheduledFileMessage(mParams);

      expect(scheduledMessage, isA<ScheduledFileMessage>());
    });
    test('should fail sending in 3 minutes', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 180000,
      );

      expect(() => channel.createScheduledFileMessage(mParams),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('super groupchannel send scheduled file message', () {
    test('should succeed sending in 10 minutes', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
      );

      final scheduledMessage =
          await superGroupChannel.createScheduledFileMessage(mParams);

      expect(scheduledMessage, isA<ScheduledFileMessage>());
    });
    test('should fail sending in 3 minutes', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 180000,
      );

      expect(() => superGroupChannel.createScheduledFileMessage(mParams),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('groupchannel cancel scheduled message', () {
    test('should succeed with correct params', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isNotNull);

      await channel.cancelScheduledMessage(scheduledMessage.scheduledMessageId);

      final params = ScheduledMessageRetrievalParams(
        channelUrl: channel.channelUrl,
        scheduledMessageId: scheduledMessage.scheduledMessageId,
      );

      final viewScheduledMessage =
          await BaseMessage.getScheduledMessage(params);

      expect(viewScheduledMessage.scheduledInfo!.scheduledStatus,
          equals(ScheduledStatus.canceled));
    });

    test('should fail with incorrect scheduledId', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isNotNull);

      expect(() => channel.cancelScheduledMessage(-1),
          throwsA(isA<BadRequestError>()));
    });

    //TODO fix circleCI to not timeout after 10 min
    // test('should fail when cancelling before 3 min sending', () async {
    //   final mParams = ScheduledUserMessageParams(
    //     scheduledAt: DateTime.now().millisecondsSinceEpoch + 306000,
    //     message: 'scheduled test',
    //   );

    //   final scheduledMessage =
    //       await channel.createScheduledUserMessage(mParams);

    //   expect(scheduledMessage, isNotNull);

    //   await Future.delayed(Duration(minutes: 2));

    //   expect(
    //       () => channel
    //           .cancelScheduledMessage(scheduledMessage.scheduledMessageId),
    //       throwsA(isA<BadRequestError>()));
    // }, timeout: Timeout.none);
  });

  group('groupchannel get scheduled message list', () {
    test('should succeed with default parameter', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test 1',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      await channel.createScheduledUserMessage(mParams);

      final mParams2 = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test 2',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      await channel.createScheduledUserMessage(mParams2);

      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams3 = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
      );

      await channel.createScheduledFileMessage(mParams3);

      final res = await ScheduledMessageListQuery().loadNext();

      expect(res, isA<List<BaseMessage>>());
    });

    test('should succeed with correct channelUrl', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test 1',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      await channel.createScheduledUserMessage(mParams);

      final mParams2 = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test 2',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      await channel.createScheduledUserMessage(mParams2);

      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams3 = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
      );

      await channel.createScheduledFileMessage(mParams3);

      final res =
          await ScheduledMessageListQuery(channelUrl: channel.channelUrl)
              .loadNext();

      expect(res, isA<List<BaseMessage>>());
    });

    test('should fail with incorrect channelUrl', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test 1',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      await channel.createScheduledUserMessage(mParams);

      final mParams2 = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test 2',
        translationTargetLanguages: ['KR', 'ENG'],
      );

      await channel.createScheduledUserMessage(mParams2);

      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams3 = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
      );

      await channel.createScheduledFileMessage(mParams3);

      expect(() => ScheduledMessageListQuery(channelUrl: 'na').loadNext(),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('groupchannel get total scheduled messages count', () {
    test('should succeed with correct params', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isNotNull);

      final totalCount = await sdk.getTotalScheduledMessageCount();

      expect(totalCount, equals(1));
    });
    test('should fail with incorrect channelUrl', () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isNotNull);

      var param = TotalScheduledMessageCountParams(
        channelUrl: 'na',
        channelType: ChannelType.group,
      );

      expect(() => sdk.getTotalScheduledMessageCount(params: param),
          throwsA(isA<BadRequestError>()));
    });
  });

  group('group channel get scheduled message', () {
    test('should succeed getting scheduled user message with correct params',
        () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isNotNull);

      final params = ScheduledMessageRetrievalParams(
        channelUrl: channel.channelUrl,
        scheduledMessageId: scheduledMessage.scheduledMessageId,
      );

      final viewScheduledMessage =
          await BaseMessage.getScheduledMessage(params);

      expect(viewScheduledMessage, isA<ScheduledUserMessage>());
    });

    test(
        'should fail getting scheduled user message with incorrect scheduled message id',
        () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isNotNull);

      final params = ScheduledMessageRetrievalParams(
        channelUrl: channel.channelUrl,
        scheduledMessageId: -1,
      );

      expect(() => BaseMessage.getScheduledMessage(params),
          throwsA(isA<BadRequestError>()));
    });

    test('should fail getting scheduled user message with incorrect channelUrl',
        () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isNotNull);

      final params = ScheduledMessageRetrievalParams(
        channelUrl: 'na',
        scheduledMessageId: scheduledMessage.scheduledMessageId,
      );

      expect(() => BaseMessage.getScheduledMessage(params),
          throwsA(isA<BadRequestError>()));
    });

    test(
        'should fail getting scheduled user message with incorrect channelUrl and incorrect scheduled message id',
        () async {
      final mParams = ScheduledUserMessageParams(
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
        message: 'scheduled test',
      );

      final scheduledMessage =
          await channel.createScheduledUserMessage(mParams);

      expect(scheduledMessage, isNotNull);

      final params = ScheduledMessageRetrievalParams(
        channelUrl: 'na',
        scheduledMessageId: -1,
      );

      expect(() => BaseMessage.getScheduledMessage(params),
          throwsA(isA<BadRequestError>()));
    });

    test('should succeed getting scheduled file message with correct params',
        () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      var mParams = ScheduledFileMessageParams.withFile(
        imageFile,
        scheduledAt: DateTime.now().millisecondsSinceEpoch + 600000,
      );

      final scheduledMessage =
          await channel.createScheduledFileMessage(mParams);

      final params = ScheduledMessageRetrievalParams(
        channelUrl: channel.channelUrl,
        scheduledMessageId: scheduledMessage.scheduledMessageId,
      );

      final viewScheduledMessage =
          await BaseMessage.getScheduledMessage(params);

      expect(viewScheduledMessage, isA<ScheduledFileMessage>());
    });
  });
}
