import 'dart:async';

import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late ChannelEventListener event;
  late GroupChannel channel;
  late UserMessage message;
  late User user;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    user = await testApiCreateUserAndConnect();
    event = ChannelEventListener();
    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = [user.userId];
    channel = await createGroupChannel(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('groupchannel send user message', () {
    test('should succeed with normal text', () async {
      final mParams = UserMessageParams(message: 'test');

      final pending = channel.sendUserMessage(
        mParams,
        onCompleted: (message, error) {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.message, equals('test'));
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
        },
      );
      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('should succeed with valid params', () async {
      final mParams = UserMessageParams(message: 'test');
      mParams.metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1']),
        MessageMetaArray(key: 'key2', value: ['value2', 'value22'])
      ];
      mParams.data = 'data';
      mParams.customType = 'custom';
      mParams.mentionType = MentionType.users;
      mParams.pushOption = PushNotificationDeliveryOption.suppress;

      final pending = channel.sendUserMessage(
        mParams,
        onCompleted: (message, error) {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.message, equals('test'));
          expect(message.metaArrays, isNotNull);
          expect(message.metaArrays?.first.key, equals('key1'));
          expect(message.data, equals('data'));
          expect(message.customType, equals('custom'));
          expect(message.mentionType, equals(MentionType.users));
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
        },
      );
      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);

      final results = await event.waitFor([EventFuncName.onChannelChanged]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('should throw an exception with empty string', () async {
      final mParams = UserMessageParams(message: '');
      expect(
        () => channel.sendUserMessage(mParams),
        throwsA(isA<SBError>()),
      );
    });

    test('should throw error without connection', () async {
      await sdk.disconnect();

      final wait = Completer();
      channel.sendUserMessageWithText(
        'test',
        onCompleted: (message, error) {
          expect(error, isNotNull);
          expect(error, isA<ConnectionRequiredError>());
          wait.complete();
        },
      );

      await wait.future;
    });
  });

  group('groupchannel update user message', () {
    test('should succeed with given valid user message', () async {
      final mParams = UserMessageParams(message: 'test');

      channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) {},
      );

      final result = await event.waitFor([EventFuncName.onChannelChanged]);
      message =
          (result.first.data.first as GroupChannel).lastMessage as UserMessage;

      mParams.message = 'test updated';

      channel.updateUserMessage(message.messageId, mParams);
      final results = await event.waitFor([EventFuncName.onMessageUpdated]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
      final updated = results.first.data.last as UserMessage;
      expect(updated.message, 'test updated');
    });

    test('should throw an exception with given invalid user message id',
        () async {
      final params = UserMessageParams(message: 'test updated');
      expect(() => channel.updateUserMessage(-1, params),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel delete user message', () {
    test('should succeed with valid sent user message', () async {
      final mParams = UserMessageParams(message: 'test');
      channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) {},
      );

      final result = await event.waitFor([EventFuncName.onChannelChanged]);
      message =
          (result.first.data.first as GroupChannel).lastMessage as UserMessage;

      channel.deleteMessage(message.messageId);
      final result2 = await event.waitFor([EventFuncName.onMessageDeleted]);
      expect(result2.first.data.last, equals(message.messageId));
    });

    test('should throw an exception with invalid message id', () async {
      expect(() => channel.deleteMessage(123), throwsA(isA<SBError>()));
    });

    test('should throw an exception with invalid negative integer message id',
        () async {
      expect(() => channel.deleteMessage(-1),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel resend user message', () {
    test('should succeed with failed user message', () async {
      //NOTE: problem is async operations could still be running at the point
      //disconnect is called. How do I handle?
      await sdk.disconnect();

      final mParams = UserMessageParams(message: 'test')..data = 'data';

      var wait = Completer();
      message = channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) {
          message = msg;
          wait.complete();
        },
      );

      await wait.future;
      expect(message.sendingStatus, equals(MessageSendingStatus.failed));

      await testApiConnect(user.userId);
      channel = await GroupChannel.getChannel(channel.channelUrl);
      event = ChannelEventListener();
      channel.resendUserMessage(message, onCompleted: (msg, err) {});
      await event.waitFor([EventFuncName.onChannelChanged]);
    });

    test('should throw an exception when message retrieve an error message',
        () async {
      final mParams = UserMessageParams(message: 'test')..data = 'data';

      var wait = Completer();
      message = channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) {
          message = msg..errorCode = 800101; //Errorcode.connectionRequired
          wait.complete();
        },
      );

      await wait.future;

      expect(() => channel.resendUserMessage(message),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with already sent user message', () async {
      final mParams = UserMessageParams(message: 'test');
      channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) {},
      );

      final result = await event.waitFor([EventFuncName.onChannelChanged]);
      message =
          (result.first.data.first as GroupChannel).lastMessage as UserMessage;

      expect(() => channel.resendUserMessage(message, onCompleted: null),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with invalid user message', () async {
      final fakeMessage = UserMessage(
        translations: <String, String>{},
        messageId: 123,
        message: 'message',
        channelType: ChannelType.group,
        channelUrl: 'url',
        sender: Sender.fromUser(user, channel)!,
      );
      expect(() => channel.resendUserMessage(fakeMessage, onCompleted: null),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel copy user message', () {
    test('should succeed with given other channel', () async {
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final otherChannel = await createGroupChannel(params);

      final mParams = UserMessageParams(message: 'test');
      message = channel.sendUserMessage(mParams);
      final result = await event.waitFor([EventFuncName.onChannelChanged]);
      message =
          (result.first.data.first as GroupChannel).lastMessage as UserMessage;

      final wait = Completer();
      channel.copyMessage(message, otherChannel, onCompleted: (copied, err) {
        expect(err, isNull);
        expect(copied, isNotNull);
        expect(message.message, equals(copied.message));
        expect(message.sender?.userId, equals(copied.sender?.userId));
        wait.complete();
      });

      await event.waitFor([EventFuncName.onChannelChanged]);
      await wait.future;
    });

    test('should succeed with given same channel', () async {
      final mParams = UserMessageParams(message: 'test');

      message = channel.sendUserMessage(mParams);
      final result = await event.waitFor([EventFuncName.onChannelChanged]);
      message =
          (result.first.data.first as GroupChannel).lastMessage as UserMessage;

      final wait = Completer();
      channel.copyMessage(message, channel, onCompleted: (copied, err) {
        expect(err, isNull);
        expect(copied, isNotNull);
        expect(message.message, equals(copied.message));
        expect(message.sender?.userId, equals(copied.sender?.userId));
        wait.complete();
      });

      await event.waitFor([EventFuncName.onChannelChanged]);
      await wait.future;
    });

    test('should throw an exception with other channel message', () async {
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final otherChannel = await createGroupChannel(params);

      final mParams = UserMessageParams(message: 'test');
      otherChannel.sendUserMessage(mParams);
      final result = await event.waitFor([EventFuncName.onChannelChanged]);
      message =
          (result.first.data.first as GroupChannel).lastMessage as UserMessage;

      expect(() => channel.copyMessage(message, otherChannel),
          throwsA(isA<SBError>()));
    });

    test('should throw an exception with invalid message', () async {
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final otherChannel = await createGroupChannel(params);

      final fake = UserMessage(
        translations: <String, String>{},
        messageId: 123,
        message: 'message',
        channelType: ChannelType.group,
        channelUrl: 'url',
        sender: Sender.fromUser(user, channel)!,
      );
      expect(() => channel.copyMessage(fake, otherChannel),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel translate user message', () {
    test('should succeed with target languages and user message', () async {
      final mParams = UserMessageParams(message: 'test');

      message = channel.sendUserMessage(mParams);
      final result = await event.waitFor([EventFuncName.onChannelChanged]);
      message =
          (result.first.data.first as GroupChannel).lastMessage as UserMessage;

      final translated = await channel.translateUserMessage(message, ['ko']);
      expect(translated.translations.isEmpty, isFalse);
      expect(translated.translations.length, equals(1));
    });

    test('should throw an exception with invalid message', () async {
      final fake = UserMessage(
        translations: <String, String>{},
        messageId: 0,
        message: 'message',
        channelType: ChannelType.group,
        channelUrl: 'url',
        sender: Sender.fromUser(user, channel)!,
      );
      expect(() => channel.translateUserMessage(fake, ['en']),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw an exception with invalid params', () async {
      final mParams = UserMessageParams(message: 'test');

      var wait = Completer();
      message = channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) {
          message = msg;
          wait.complete();
        },
      );

      await wait.future;

      expect(() => channel.translateUserMessage(message, []),
          throwsA(isA<InvalidParameterError>()));
    });
  });
}
