import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late GroupChannel channel;
  late ChannelEventListener event;
  late UserMessage message;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    final user = await testApiCreateUserAndConnect();
    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = [user.userId];
    channel = await createGroupChannel(params);
    event = ChannelEventListener();
    final mParams = UserMessageParams(message: 'test');
    channel.sendUserMessage(mParams);

    final results = await event.waitFor([EventFuncName.onChannelChanged]);
    expect(results, isNotNull);
    expect(results.length, equals(1));
    message =
        (results.first.data.first as GroupChannel).lastMessage as UserMessage;
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('groupchannel create message meta array', () {
    test('should success with valid meta arrays', () async {
      final result = await channel.addMessageMetaArray(message, [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2'])
      ]);

      final array = result.metaArrays!;
      expect(array.length, equals(1));
      expect(array.first.key, equals('key1'));
      expect(array.first.value, equals(['value1', 'value2']));
    });

    test('should success with multiple meta arrays', () async {
      final result = await channel.addMessageMetaArray(message, [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6'])
      ]);
      final array = result.metaArrays ?? [];
      expect(array.length, equals(3));
      expect(array.first.key, equals('key1'));
      expect(array.first.value, equals(['value1', 'value2']));
      expect(array.last.key, equals('key3'));
      expect(array.last.value, equals(['value5', 'value6']));
    });
  });

  group('groupchannel get meta arrays', () {
    test('should return corresponding meta arrays with given keys', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      final array = result.metaArrays ?? [];
      expect(array.length, equals(3));
      expect(result.getMetaArrays(['key1']).length, equals(1));
      expect(result.getMetaArrays(['key2']).first.value,
          equals(['value3', 'value4']));
    });

    test('should throw an exception with invalid params', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      expect(() => result.getMetaArrays([]), throwsA(isA<SBError>()));
    });
  });

  group('groupchannel remove message meta array', () {
    test('should success to remove meta array values', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];
      final removeArrays = [
        MessageMetaArray(key: 'key1', value: ['value1']),
      ];
      var result = await channel.addMessageMetaArray(message, metaArrays);
      expect(result.metaArrays?.length, equals(3));

      result = await channel.removeMessageMetaArray(message, removeArrays);
      final array = result.metaArrays ?? [];
      expect(array, isNotNull);
      expect(array.length, equals(3));
      expect(array.first.key, equals('key1'));
      expect(array.first.value.length, equals(1));
      expect(array.first.value, equals(['value2']));
    });

    test('should success to remove meta array multiple values ', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];
      final removeArrays = [
        MessageMetaArray(key: 'key1', value: ['value1']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];
      var result = await channel.addMessageMetaArray(message, metaArrays);
      expect(result.metaArrays?.length, equals(3));

      result = await channel.removeMessageMetaArray(message, removeArrays);
      final array = result.metaArrays ?? [];
      expect(array.length, equals(3));
      expect(array.first.key, equals('key1'));
      expect(array.first.value.length, equals(1));
      expect(array.first.value, equals(['value2']));
      expect(array.last.key, equals('key3'));
      expect(array.last.value.length, equals(0));
    });
  });

  group('groupchannel delete message meta array', () {
    test('should success to delete meta arrays', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      expect(result.metaArrays?.length, equals(3));

      result = await channel.deleteMessageMetaArrayKeys(message, [
        'key1',
        'key2',
        'key3',
      ]);
      expect(result.metaArrays, isNull);
    });

    test('should success to delete partial meta arrays', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      expect(result.metaArrays?.length, equals(3));

      result =
          await channel.deleteMessageMetaArrayKeys(message, ['key2', 'key3']);
      final array = result.metaArrays ?? [];
      expect(array.length, equals(1));
      expect(array.first.key, equals('key1'));
    });

    test('should throw an exception with invalid params', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      expect(result.metaArrays?.length, equals(3));

      expect(() => channel.deleteMessageMetaArrayKeys(message, []),
          throwsA(isA<InvalidParameterError>()));
    });
  });
}
