import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late GroupChannel channel;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    final user = await testApiCreateUserAndConnect();
    var params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = [user.userId];
    channel = await createGroupChannel(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('groupchannel meta counter create', () {
    test('should success and return created meta counter', () async {
      final event = ChannelEventListener();
      final metaCounters = {'key1': 1};
      channel.createMetaCounters(metaCounters);
      final result = await event.waitFor([EventFuncName.onMetaCountersCreated]);

      //wait for a certain event happen
      expect((result[0].data.last as Map<String, int>)['key1'], equals(1));
    });

    test('should throw an exception with null or empty parameter', () async {
      expect(() => channel.createMetaCounters({}),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel meta counter get', () {
    test('should success and return counters with given keys', () async {
      final event = ChannelEventListener();
      final metaCounters = {'key1': 3, 'key2': 6, 'key3': 21};
      channel.createMetaCounters(metaCounters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);
      final result = await channel.getMetaCounters(['key2', 'key3']);
      final expected = {'key2': 6, 'key3': 21};
      expect(result, expected);
    });

    test('should throw an exception with given invalid params', () async {
      final event = ChannelEventListener();
      final metaCounters = {'key1': 3, 'key2': 6, 'key3': 21};
      channel.createMetaCounters(metaCounters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);
      expect(() => channel.getMetaCounters([]),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should success and return all meta counters', () async {
      final event = ChannelEventListener();
      final metaCounters = {'key1': 3, 'key2': 6};
      channel.createMetaCounters(metaCounters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);
      final result = await channel.getAllMetaCounters();
      expect(result, metaCounters);
    });

    test('should success and return empty', () async {
      final result = await channel.getAllMetaCounters();
      expect(result, {});
    });
  });

  group('groupchannel meta counter update', () {
    test('should success and return updated meta counter', () async {
      final event = ChannelEventListener();
      final metaConters = {'key1': 1};
      channel.createMetaCounters(metaConters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);

      final updated = {'key1': 5};
      final result = await channel.updateMetaCounters(updated);

      expect(result['key1'], equals(5));
    });

    test('should success and return increased meta counter', () async {
      final event = ChannelEventListener();
      final metaConters = {'key1': 1};
      channel.createMetaCounters(metaConters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);

      final updated = {'key1': 5};
      channel.increaseMetaCounters(updated);
      final result = await event.waitFor([EventFuncName.onMetaCountersUpdated]);

      expect((result[0].data.last as Map<String, int>)['key1'], equals(6));
    });

    test('should throw an exception with invalid params on increated meta data',
        () async {
      final event = ChannelEventListener();
      final metaConters = {'key1': 1};
      channel.createMetaCounters(metaConters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);
      expect(() => channel.increaseMetaCounters({}),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should success and return decreased meta counter', () async {
      final event = ChannelEventListener();
      final metaConters = {'key1': 1};
      channel.createMetaCounters(metaConters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);

      final updated = {'key1': 5};
      channel.decreaseMetaCounters(updated);
      final result = await event.waitFor([EventFuncName.onMetaCountersUpdated]);

      expect((result[0].data.last as Map<String, int>)['key1'], equals(-4));
    });

    test('should throw an exception with invalid params on decrease meta data',
        () async {
      final event = ChannelEventListener();
      final metaConters = {'key1': 1};
      channel.createMetaCounters(metaConters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);
      expect(() => channel.decreaseMetaCounters({}),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should throw error with null or empty parameter', () async {
      expect(channel.updateMetaCounters({}),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel meta counter delete', () {
    test('should success and delete given meta counters key', () async {
      final event = ChannelEventListener();
      final metaCounters = {'key1': 1, 'key2': 3};
      channel.createMetaCounters(metaCounters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);
      channel.deleteMetaCounters('key2');
      await event.waitFor([EventFuncName.onMetaCountersDeleted]);
      final result = await channel.getAllMetaCounters();
      final expected = {'key1': 1};
      expect(result, expected);
    });

    test('should delete all', () async {
      final event = ChannelEventListener();
      final metaCounters = {'key1': 1, 'key2': 3};
      channel.createMetaCounters(metaCounters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);
      channel.deleteAllMetaCounters();
      await event.waitFor([EventFuncName.onMetaCountersDeleted]);
      final result = await channel.getAllMetaCounters();
      expect(result.isEmpty, isTrue);
    });

    test('should fail with non exist meta counters key', () async {
      final event = ChannelEventListener();
      final metaCounters = {'key1': 1, 'key2': 3};
      channel.createMetaCounters(metaCounters);
      await event.waitFor([EventFuncName.onMetaCountersCreated]);
      expect(
        channel.deleteMetaCounters('key3'),
        throwsA(isA<BadRequestError>()),
      );
    });

    test(
        'should throw an exception and generated error with null or empty parameter',
        () async {
      expect(
        channel.deleteMetaCounters(''),
        throwsA(isA<InvalidParameterError>()),
      );
    });
  });
}
