import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late GroupChannel channel;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    final user = await testApiCreateUserAndConnect();
    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = [user.userId];
    channel = await createGroupChannel(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('groupchannel meta data create', () {
    test('should success and return created meta data', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      channel.createMetaData(metaData);
      final result = await event.waitFor([EventFuncName.onMetaDataCreated]);
      //wait for a certain event happen
      expect((result[0].data.last as Map<String, String>)['key1'],
          equals('value1'));
    });
    test('should throw an exception with null or empty parameter', () async {
      expect(() => channel.createMetaData({}),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel meta data update', () {
    test('should success and return updated meta data', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);

      final updatedData = {'key1': 'value2'};

      channel.updateMetaData(updatedData);
      final result = await event.waitFor([EventFuncName.onMetaDataUpdated]);
      expect((result[0].data.last as Map<String, String>)['key1'],
          equals('value2'));
    });

    test('should throw an exception with null or empty parameter', () async {
      expect(() => channel.updateMetaData({}),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('groupchannel meta data delete', () {
    test('should success and delete given metadata key', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      channel.deleteMetaData('key2');
      await event.waitFor([EventFuncName.onMetaDataDeleted]);
      final result = await channel.getAllMetaData();
      final expected = {'key1': 'value1'};
      expect(result, expected);
    });

    test('should success and delete all metadata', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      channel.deleteAllMetaData();
      await event.waitFor([EventFuncName.onMetaDataDeleted]);
      final result = await channel.getAllMetaData();
      expect(result.isEmpty, isTrue);
    });

    test('should throw an exception with non exist metadata key', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);

      expect(
        () => channel.deleteMetaData('key3'),
        throwsA(isA<BadRequestError>()),
      );
    });

    test('should throw an exception with null or empty parameter', () async {
      expect(
        () => channel.deleteMetaData(''),
        throwsA(isA<InvalidParameterError>()),
      );
    });
  });

  group('groupchannel meta data get', () {
    test('should success and return metadata with given keys', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      final result = await channel.getMetaData(['key2', 'key3']);
      final expected = {'key2': 'value2', 'key3': 'value3'};
      expect(result, expected);
    });

    test('should throw an exception with given invalid params', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      expect(
          () => channel.getMetaData([]), throwsA(isA<InvalidParameterError>()));
    });

    test('should success and return all meta data', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      final result = await channel.getAllMetaData();
      expect(result, metaData);
    });

    test('should success and return empty', () async {
      final result = await channel.getAllMetaData();
      expect(result, {});
    });
  });

  group('groupchannel meta data cached', () {
    test('should return proper value after create', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      final cached = channel.getCachedMetaData();
      expect(cached.length, equals(1));
      expect(cached['key1'], equals('value1'));
    });

    test('should return proper value after update', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      channel.updateMetaData({'key1': 'value2'});
      await event.waitFor([EventFuncName.onMetaDataUpdated]);
      final cached = channel.getCachedMetaData();
      expect(cached.length, equals(1));
      expect(cached['key1'], equals('value2'));
    });

    test('should return proper value after delete', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      var cached = channel.getCachedMetaData();
      expect(cached.length, equals(2));
      expect(cached['key1'], equals('value1'));
      expect(cached['key2'], equals('value2'));
      channel.deleteMetaData('key1');
      await event.waitFor([EventFuncName.onMetaDataDeleted]);
      cached = channel.getCachedMetaData();
      expect(cached.length, equals(1));
      expect(cached['key1'], isNull);
    });

    test('should return proper value after getall', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      await channel.getAllMetaData();
      final cached = channel.getCachedMetaData();
      expect(cached.length, equals(2));
      expect(cached['key1'], equals('value1'));
      expect(cached['key2'], equals('value2'));
    });

    test('should return proper value after removal', () async {
      final event = ChannelEventListener();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      channel.createMetaData(metaData);
      await event.waitFor([EventFuncName.onMetaDataCreated]);
      var cached = channel.getCachedMetaData();
      expect(cached.length, equals(2));
      expect(cached['key1'], equals('value1'));
      expect(cached['key2'], equals('value2'));
      channel.deleteAllMetaData();
      await event.waitFor([EventFuncName.onMetaDataDeleted]);
      cached = channel.getCachedMetaData();
      expect(cached.length, equals(0));
    });
  });
}
