import 'package:sendbird_sdk/request/messages/user_message_send_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../../common/callback.dart';
import '../../../common/helpers/private_api_client.dart';
import '../../../common/prep.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';
import '../../../common/utils.dart';
import '../../../common/ws_tester.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('openchannel message releated event', () {
    test(
        'onReceiveMessage & onChannelChanged should get called when receiving message',
        () async {
      final users = await testApiCreateUsers(3);
      final user = users.first;
      await testApiConnect(users.first.userId);

      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      var channel = await OpenChannel.createChannel(params);
      await channel.enter();

      final mParams = UserMessageParams(message: 'test');
      final event = ChannelEventListener();
      PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
        params: mParams,
      ));

      final results = await event.waitFor([
        EventFuncName.onMessageReceived,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
      final received = results.first.data.last as UserMessage;
      expect(received, isNotNull);
      expect(received.message, 'test');
    });
  });

  group('openchannel operation releated event', () {
    test('onChannelOperatorUpdated should get called on change operators',
        () async {
      final user = await testApiCreateUserAndConnect();
      final users = await testApiCreateUsers(2);
      final imageFile = await getAssetFrom('resources/image.jpeg');

      final params = OpenChannelParams()
        ..name = 'channel'
        ..operatorUserIds = [user.userId];

      final event = ChannelEventListener();
      final channel = await OpenChannel.createChannel(params);
      await channel.enter();

      final updatedParams = OpenChannelParams()
        ..name = 'update name'
        ..channelUrl = channel.channelUrl
        ..operatorUserIds = users.map((e) => e.userId).toList()
        ..coverImage = FileInfo.fromData(
          file: imageFile,
          name: 'Cover Image',
          mimeType: 'image/jpeg',
        );

      channel.updateChannel(updatedParams);

      final results2 = await event.waitFor([
        EventFuncName.onChannelOperatorsUpdated,
        EventFuncName.onChannelChanged,
      ]);

      expect(results2, isNotNull);
      expect(results2.length, equals(2));
    });

    test('onChannelDelete should get called when delete the channel', () async {
      final user = await testApiCreateUserAndConnect();

      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      var channel = await OpenChannel.createChannel(params);
      await channel.enter();

      final event = ChannelEventListener();
      channel.deleteChannel();

      final results = await event.waitFor([
        EventFuncName.onChannelDeleted,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });
  });

  group('openchannel metadata event', () {
    test('onChannelMetaDataCreated should get call when creating meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel = await OpenChannel.createChannel(params);
      await channel.enter();

      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      channel.createMetaData(metaData);

      final results = await event.waitFor([EventFuncName.onMetaDataCreated]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onChannelMetaDataUpdated should get call when updating meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel = await OpenChannel.createChannel(params);
      await channel.enter();

      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      await channel.createMetaData(metaData);
      channel.updateMetaData({'key1': 'value2'});

      final results = await event.waitFor([EventFuncName.onMetaDataUpdated]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onChannelMetaDataDeleted should get call when deleting meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel = await OpenChannel.createChannel(params);
      await channel.enter();

      final event = ChannelEventListener();
      final metaData = {'key1': 'value1'};
      await channel.createMetaData(metaData);
      channel.deleteMetaData('key1');

      final results = await event.waitFor([EventFuncName.onMetaDataDeleted]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });
  });

  group('openchannel metacounter event', () {
    test('onChannelMetaCounterCreated should get call when creating meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final event = ChannelEventListener();
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel = await OpenChannel.createChannel(params);
      await channel.enter();
      channel.createMetaCounters({'key1': 1});

      final results =
          await event.waitFor([EventFuncName.onMetaCountersCreated]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onChannelMetaCounterUpdated should get call when updating meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final event = ChannelEventListener();
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel = await OpenChannel.createChannel(params);
      await channel.enter();
      await channel.createMetaCounters({'key1': 1});
      channel.updateMetaCounters({'key1': 2});

      var results = await event.waitFor([EventFuncName.onMetaCountersUpdated]);

      expect(results, isNotNull);
      expect(results.length, equals(1));

      channel.increaseMetaCounters({'key1': 4});

      results = await event.waitFor([
        EventFuncName.onMetaCountersUpdated,
      ]);
      expect(results, isNotNull);
      expect(results.length, equals(1));

      channel.decreaseMetaCounters({'key1': 5});

      results = await event.waitFor([
        EventFuncName.onMetaCountersUpdated,
      ]);
      expect(results, isNotNull);
      expect(results.length, equals(1));
    });

    test('onChannelMetaCounterDeleted should get call when deleting meta',
        () async {
      final user = await testApiCreateUserAndConnect();
      final event = ChannelEventListener();
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel = await OpenChannel.createChannel(params);
      await channel.enter();

      await channel.createMetaCounters({'key1': 2});
      channel.deleteMetaCounters('key1');

      final results =
          await event.waitFor([EventFuncName.onMetaCountersDeleted]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
    });
  });

  group('openchannel member count change event', () {
    test('onChannelParticipantCountChanged should get called', () async {
      await configureAppForMCNT(
        appId: testInfo.appId,
        eventInterval: 5000,
        threshold: 1,
      );
      final event = ChannelEventListener();

      final users = await testApiCreateUsers(2);
      final user = await testApiCreateUserAndConnect();
      //NOTE: need to find a way to cache or merge in place
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel = await OpenChannel.createChannel(params);
      // channel.enter();

      final wsTester = WSTester(
        appId: testInfo.appId,
        userId: users[0].userId,
        wsHost: sdk.getInternal().state.wsHost!,
      );
      await wsTester.connect();
      await wsTester.enter(channel);

      channel.enter();

      //callback when someone enters (threshold)
      await event.waitFor([EventFuncName.onChannelParticipantCountChanged]);

      // final event2 = ChannelEventListener();
      // //callback when interval is met
      // await event2.waitFor([EventFuncName.onChannelParticipantCountChanged]);

      wsTester.exit(channel);

      await event.waitFor([EventFuncName.onChannelParticipantCountChanged]);

      await wsTester.close();
    }, timeout: Timeout(Duration(seconds: 180)));
  });

  // group('OpenChannel moderation related events', () {
  //   test('onUserMuted & onUserUnmuted should get called on mute & unmute',
  //       () async {
  //     final users = await testApiCreateUsers(3);
  //     await testApiConnect(users.first.userId);

  //     final params = OpenChannelParams()
  //       ..operatorUserIds = users.map((e) => e.userId).toList();

  //     final channel = await OpenChannel.createChannel(params);
  //     await channel.enter();

  //     channel.muteUser(userId: users.last.userId);
  //     final event = ChanneleventTester();
  //     var results = await event.waitFor([
  //       eventFncName.onUserMuted,
  //     ]);
  //     expect(results, isNotNull);
  //     expect(results.length, equals(1));

  //     channel.unmuteUser(userId: users.last.userId);
  //     results = await event.waitFor([
  //       eventFncName.onUserUnmuted,
  //     ]);
  //     expect(results, isNotNull);
  //     expect(results.length, equals(1));
  //   });

  //   test('onUserBanned & onUserUnbanned should get called on ban & unbanned',
  //       () async {
  //     final users = await testApiCreateUsers(3);
  //     await testApiConnect(users.first.userId);

  //     final params = OpenChannelParams()
  //       ..operatorUserIds = users.map((e) => e.userId).toList();

  //     final channel = await OpenChannel.createChannel(params);

  //     channel.banUser(userId: users.last.userId);
  //     final event = ChanneleventTester();
  //     var results = await event.waitFor([
  //       eventFncName.onUserBanned,
  //     ]);
  //     expect(results, isNotNull);
  //     expect(results.length, equals(1));

  //     channel.unbanUser(userId: users.last.userId);
  //     results = await event.waitFor([
  //       eventFncName.onUserUnbanned,
  //     ]);
  //     expect(results, isNotNull);
  //     expect(results.length, equals(1));
  //   });
  // });
}
