// Copyright © 2020 SENDBIRD.COM. All rights reserved.

import 'package:test/test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/common.dart';
import '../../../common/callback.dart';
import '../../../common/user.dart';
import '../../../common/sdksetup.dart';
import '../../../common/utils.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('openchannel create', () {
    test('should create a new channel with given params', () async {
      OpenChannel channel;
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      var params = OpenChannelParams()..operatorUserIds = [user.userId];
      channel = await OpenChannel.createChannel(params);
      expect(channel, isNotNull);
    });

    test('should create a new channel with given image url info', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      OpenChannel? channel;

      var params = OpenChannelParams();
      params.coverImage =
          FileInfo.fromUrl(name: 'Name', url: imageUrl, mimeType: 'image/jpeg');

      try {
        channel = await OpenChannel.createChannel(params);
      } catch (e) {
        TestFailure('Exception: $e');
      }
      expect(channel, isNotNull);
      expect(channel?.coverUrl, equals(imageUrl));
    });

    test('should create a new channel with given image data', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      OpenChannel channel;
      var sentBytes = 0;
      var totalBytes = 0;

      final imageFile = await getAssetFrom('resources/image.jpeg');
      final params = OpenChannelParams()
        ..operatorUserIds = [user.userId]
        ..coverImage = FileInfo.fromData(
          file: imageFile,
          name: 'Cover Image',
          mimeType: 'image/jpeg',
        );

      channel =
          await OpenChannel.createChannel(params, progress: (sent, total) {
        sentBytes = sent;
        totalBytes = total;
      });

      expect(channel, isNotNull);
      expect(sentBytes, equals(totalBytes));
      expect(channel.coverUrl, isNotEmpty);
      expect(sentBytes != 0, true);
    });
  });

  group('openchannel update', () {
    test('should change given channel with Image Data', () async {
      final user = await testApiCreateUserAndConnect();
      final imageFile = await getAssetFrom('resources/image.jpeg');
      final params = OpenChannelParams()
        ..name = 'channel'
        ..operatorUserIds = [user.userId];

      final channel = await OpenChannel.createChannel(params);
      await channel.enter();

      expect(channel, isNotNull);
      expect(channel.coverUrl, isNotNull);
      expect(channel.name, 'channel');

      var sentBytes = 0;
      var totalBytes = 0;

      final event = ChannelEventListener();

      params
        ..name = 'update name'
        ..coverImage = FileInfo.fromData(
          file: imageFile,
          name: 'Cover Image',
          mimeType: 'image/jpeg',
        );

      channel.updateChannel(params, progress: (sent, total) {
        sentBytes = sent;
        totalBytes = total;
      });

      final results = await event.waitFor([
        EventFuncName.onChannelChanged,
      ]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
      expect(sentBytes, equals(totalBytes));
      expect(sentBytes != 0, isTrue);
    });

    test('should change given channel with image url', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      final params = OpenChannelParams()
        ..name = 'the name'
        ..operatorUserIds = [user.userId];

      var channel = await OpenChannel.createChannel(params);
      expect(channel, isNotNull);

      var newParams = OpenChannelParams()
        ..name = 'new name'
        ..coverImage = FileInfo.fromUrl(
          name: 'Name',
          url: imageUrl,
          mimeType: 'image/jpeg',
        );

      channel = await channel.updateChannel(newParams);
      expect(channel.name, 'new name');
      expect(channel.coverUrl, imageUrl);
    });

    test('should throw an exception when channels are different', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      final params = OpenChannelParams()
        ..name = 'the name'
        ..operatorUserIds = [user.userId];
      final newParams = OpenChannelParams()..channelUrl = 'newUrl';

      var channel = await OpenChannel.createChannel(params);
      expect(channel, isNotNull);

      expect(() => channel.updateChannel(newParams),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('openchannel delete', () {
    test('should delete given channel', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      var params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel = await OpenChannel.createChannel(params);
      expect(channel, isNotNull);

      await channel.deleteChannel();
      //channel should not be valid anymore
    });
  });

  group('openchannel get', () {
    test('should return a channel with given channel url', () async {
      final user = await testApiCreateUserAndConnect();
      expect(user, isNotNull);

      var params = OpenChannelParams();
      final channel1 = await OpenChannel.createChannel(params);
      expect(channel1, isNotNull);

      final channel2 = await OpenChannel.getChannel(channel1.channelUrl);
      expect(channel2, isNotNull);

      expect(channel1.channelUrl, equals(channel2.channelUrl));
    });
  });
}
