import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../../common/callback.dart';
import '../../../common/common.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late List<User> users;
  late User user;
  late ChannelEventListener event;
  late OpenChannel channel;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);

    users = await testApiCreateUsers(2);
    user = await testApiConnect(users.first.userId);
    event = ChannelEventListener();
    final params = OpenChannelParams()..operatorUserIds = [user.userId];
    channel = await createOpenChannel(params: params, user: user);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('openchannel ban user', () {
    test('should success to ban valid user', () async {
      channel.banUser(userId: users.last.userId);
      final result = await event.waitFor([EventFuncName.onUserBanned]);
      expect(result.length, equals(1));
    });

    test('should throw an exception with given invalid user', () async {
      expect(() => channel.banUser(userId: 'fake_id'), throwsA(isA<SBError>()));
    });
  });

  group('openchannel unban user', () {
    test('should success to ban valid user', () async {
      channel.banUser(userId: users.last.userId);
      var result = await event.waitFor([EventFuncName.onUserBanned]);
      expect(result.length, equals(1));

      channel.unbanUser(userId: users.last.userId);
      result = await event.waitFor([EventFuncName.onUserUnbanned]);
      expect(result.length, equals(1));
    });

    test('should throw an exception with given not banned user', () async {
      expect(() => channel.unbanUser(userId: users.last.userId),
          throwsA(isA<SBError>()));
    });

    test('should throw an exception with given invalid user', () async {
      expect(
        () => channel.unbanUser(userId: 'fake_id'),
        throwsA(isA<SBError>()),
      );
    });
  });

  group('openchannel mute user', () {
    test('should success to mute valid user', () async {
      channel.muteUser(userId: users.last.userId);
      final result = await event.waitFor([EventFuncName.onUserMuted]);
      expect(result.length, equals(1));
    });

    test('should throw an exception with given invalid user', () async {
      expect(
        () => channel.muteUser(userId: 'fake_id'),
        throwsA(isA<SBError>()),
      );
    });
  });

  group('openchannel unmute user', () {
    test('should success to unmute valid user', () async {
      channel.muteUser(userId: users.last.userId);
      var result = await event.waitFor([EventFuncName.onUserMuted]);
      expect(result.length, equals(1));

      channel.unmuteUser(userId: users.last.userId);
      result = await event.waitFor([EventFuncName.onUserUnmuted]);
      expect(result.length, equals(1));
    });

    test('should throws an exception with given not muted user', () async {
      expect(
        () => channel.unmuteUser(userId: users.last.userId),
        throwsA(isA<SBError>()),
      );
    });
    test('should throw an exception with given invalid user', () async {
      expect(
        () => channel.unmuteUser(userId: 'fake_id'),
        throwsA(isA<SBError>()),
      );
    });
  });

  group('openchannel get my mute info', () {
    test('shuold return value when umuted state', () async {
      final res = await channel.getMyMuteInfo();
      expect(res.isMuted, isFalse);
      expect(res.description, '');
      expect(res.endAt, 0);
      expect(res.startAt, 0);
    });

    test('shuold return value when muted state', () async {
      await channel.muteUser(
        userId: user.userId,
        description: 'mute you',
        seconds: 100,
      );

      final res = await channel.getMyMuteInfo();
      expect(res.isMuted, isTrue);
      expect(res.description, 'mute you');
      expect(res.endAt, isNot(equals(0)));
      expect(res.startAt, isNot(equals(0)));
    });
  });
}
