import 'dart:async';

import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/callback.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';
import '../../../common/utils.dart';
import '../../../common/common.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late ChannelEventListener event;
  late OpenChannel channel;
  late FileMessage message;
  late User user;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    user = await testApiCreateUserAndConnect();
    event = ChannelEventListener();
    final params = OpenChannelParams()..operatorUserIds = [user.userId];
    channel = await OpenChannel.createChannel(params);
    channel.enter();
    await event.waitFor([EventFuncName.onUserEntered]);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('openchannel send file message', () {
    test('should succeed replyToChannel with true', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      final mParams = FileMessageParams.withFile(imageFile)
        ..replyToChannel = true;
      //this parameter is not included

      final wait = Completer();
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) async {
          expect(message.replyToChannel, isTrue);
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      //returns filemessage with replyToChannel null

      //gets the file message which extends basemessage where we can check replyToChannel exists or not

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);
      await wait.future;
    });

    test('should succeed with normal file', () async {
      final imageFile = await getAssetFrom('resources/image.jpeg');
      final mParams = FileMessageParams.withFile(imageFile);

      final wait = Completer();
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) async {
          expect(message.replyToChannel, isFalse);
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.url, isNotNull);
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);
      await wait.future;
    });

    test('should succeed with url params', () async {
      final mParams = FileMessageParams.withUrl(imageUrl, size: 123);

      final wait = Completer();
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) async {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.url, equals(imageUrl));
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);
      await wait.future;
    });
  });

  group('openchannel update file message', () {
    final mParams = FileMessageParams.withUrl(imageUrl, size: 123);

    test('should succeed with given valid file message', () async {
      var wait = Completer();
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.url, equals(imageUrl));
          expect(message.sendingStatus, MessageSendingStatus.succeeded);
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);
      await wait.future;

      mParams
        ..data = 'update data'
        ..customType = 'update custom';

      channel.updateFileMessage(message.messageId, mParams);
      await event.waitFor([EventFuncName.onMessageUpdated]);
    });

    test('should throw an exception with given invalid file message id',
        () async {
      expect(() => channel.updateFileMessage(0, mParams),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('openchannel resend file message', () {
    final mParams = FileMessageParams.withUrl(imageUrl, size: 123);

    test('should succeed with failed file message', () async {
      await sdk.disconnect();

      final message = FileMessage.fromParams(params: mParams, channel: channel);
      message
        ..sendingStatus = MessageSendingStatus.failed
        ..errorCode = ErrorCode.connectionRequired;

      event = ChannelEventListener();
      await testApiConnect(user.userId);
      channel = await OpenChannel.getChannel(channel.channelUrl);
      channel.enter();
      await event.waitFor([EventFuncName.onUserEntered]);

      final wait = Completer();
      channel.resendFileMessage(message, params: mParams,
          onCompleted: (msg, err) async {
        await Future.delayed(Duration(seconds: 1));
        wait.complete();
      });
      await wait.future;
    });

    test('should throw an exception with already sent file message', () async {
      final wait = Completer();
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (message, error) async {
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );
      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);
      await wait.future;

      expect(() => channel.resendFileMessage(message, params: mParams),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('openchannel copy file message', () {
    final mParams = FileMessageParams.withUrl(imageUrl, size: 123);

    test('should succeed with given other channel', () async {
      var wait = Completer();
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);
      await wait.future;

      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel2 = await OpenChannel.createChannel(params);
      channel2.enter();
      await event.waitFor([EventFuncName.onUserEntered]);

      wait = Completer();
      channel.copyMessage(message, channel2, onCompleted: (res, err) async {
        final copied = res as FileMessage;
        expect(copied, isA<FileMessage>());
        expect(err, isNull);
        expect(copied, isNotNull);
        expect(message.url, equals(copied.url));
        expect(message.sender?.userId, equals(copied.sender?.userId));
        await Future.delayed(Duration(seconds: 1));
        wait.complete();
      });

      await wait.future;
    });

    test('should succeed with given same channel', () async {
      var wait = Completer();
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );
      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);
      await wait.future;

      wait = Completer();
      channel.copyMessage(message, channel, onCompleted: (res, err) async {
        final copied = res as FileMessage;
        expect(copied, isA<FileMessage>());
        expect(err, isNull);
        expect(copied, isNotNull);
        expect(message.url, equals(copied.url));
        expect(message.sender?.userId, equals(copied.sender?.userId));
        await Future.delayed(Duration(seconds: 1));
        wait.complete();
      });
      await wait.future;
    });

    test('should throw an exception with other channel message', () async {
      var wait = Completer();
      final pending = channel.sendFileMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, MessageSendingStatus.pending);
      await wait.future;

      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final channel2 = await OpenChannel.createChannel(params);
      channel2.enter();
      await event.waitFor([EventFuncName.onUserEntered]);

      expect(() => channel2.copyMessage(message, channel),
          throwsA(isA<InvalidParameterError>()));
    });
  });
}
