import 'dart:async';

import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/callback.dart';
import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late ChannelEventListener event;
  late User user;
  late OpenChannel channel;
  late UserMessage message;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    event = ChannelEventListener();
    user = await testApiCreateUserAndConnect();
    final params = OpenChannelParams()..operatorUserIds = [user.userId];
    channel = await OpenChannel.createChannel(params);
    channel.enter();
    await event.waitFor([EventFuncName.onUserEntered]);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('openchannel send user message', () {
    test('should succeed with normal text with replyToChannel to be true',
        () async {
      //include replyToChannel to be true
      final mParams = UserMessageParams(message: 'test')..replyToChannel = true;

      final wait = Completer();
      final pending = channel.sendUserMessage(
        mParams,
        onCompleted: (message, error) async {
          expect(message.replyToChannel, isTrue);
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      expect(pending, isNotNull);
      expect(pending.sendingStatus, equals(MessageSendingStatus.pending));
      await wait.future;
    });

    test('should succeed with normal text', () async {
      final mParams = UserMessageParams(message: 'test');

      final wait = Completer();
      final pending = channel.sendUserMessage(
        mParams,
        onCompleted: (message, error) async {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.message, equals('test'));
          expect(message.sendingStatus, equals(MessageSendingStatus.succeeded));
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );
      expect(pending, isNotNull);
      expect(pending.sendingStatus, equals(MessageSendingStatus.pending));
      await wait.future;
    });

    test('should succeed with valid params', () async {
      final mParams = UserMessageParams(message: 'test');
      mParams.metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1']),
        MessageMetaArray(key: 'key2', value: ['value2', 'value22'])
      ];
      mParams.data = 'data';
      mParams.customType = 'custom';
      mParams.mentionType = MentionType.users;
      mParams.pushOption = PushNotificationDeliveryOption.suppress;

      final wait = Completer();
      final pending = channel.sendUserMessage(
        mParams,
        onCompleted: (message, error) async {
          expect(message, isNotNull);
          expect(message.messageId, isNotNull);
          expect(message.message, equals('test'));
          expect(message.metaArrays, isNotNull);
          expect(message.metaArrays?.first.key, equals('key1'));
          expect(message.data, equals('data'));
          expect(message.customType, equals('custom'));
          expect(message.mentionType, equals(MentionType.users));
          expect(message.sendingStatus, equals(MessageSendingStatus.succeeded));
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );
      expect(pending, isNotNull);
      expect(pending.sendingStatus, equals(MessageSendingStatus.pending));
      await wait.future;
    });

    test('should throw error without connection', () async {});
  });

  group('openchannel update user message', () {
    test('should succeed with given valid user message', () async {
      final mParams = UserMessageParams(message: 'test');

      final wait = Completer();
      channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      await wait.future;

      mParams.message = 'test updated';

      channel.updateUserMessage(message.messageId, mParams);
      final results = await event.waitFor([EventFuncName.onMessageUpdated]);

      expect(results, isNotNull);
      expect(results.length, equals(1));
      final updated = results.first.data.last as UserMessage;
      expect(updated.message, 'test updated');
    });

    test('should throw an exception with given invalid user message id',
        () async {
      final params = UserMessageParams(message: 'test updated');
      expect(() => channel.updateUserMessage(-1, params),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('openchannel resend user message', () {
    test('should succeed with failed user message', () async {
      //NOTE: problem is async operations could still be running at the point
      //disconnect is called. How do I handle?
      await sdk.disconnect();

      final mParams = UserMessageParams(message: 'test');

      var wait = Completer();
      message = channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      await wait.future;
      expect(message.sendingStatus, equals(MessageSendingStatus.failed));

      await testApiConnect(user.userId);
      await channel.enter();
      // await event.waitFor([EventFuncName.onUserEntered]);

      wait = Completer();
      channel.resendUserMessage(message, onCompleted: (msg, err) {
        message = msg;
        wait.complete();
      });
      await wait.future;
    });

    test('should throw an exception with already sent user message', () async {
      final mParams = UserMessageParams(message: 'test');

      final wait = Completer();
      channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );
      await wait.future;

      expect(() => channel.resendUserMessage(message, onCompleted: null),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('openchannel copy user message', () {
    test('should succeed with given other channel', () async {
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final otherChannel = await OpenChannel.createChannel(params);
      await otherChannel.enter();

      final mParams = UserMessageParams(message: 'test');

      var wait = Completer();
      channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );
      await wait.future;

      wait = Completer();
      channel.copyMessage(message, otherChannel,
          onCompleted: (copied, err) async {
        expect(err, isNull);
        expect(copied, isNotNull);
        expect(message.message, equals(copied.message));
        expect(message.sender?.userId, equals(copied.sender?.userId));
        await Future.delayed(Duration(seconds: 1));
        wait.complete();
      });

      await wait.future;
    });

    test('should succeed with given same channel', () async {
      final mParams = UserMessageParams(message: 'test');

      var wait = Completer();
      channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) {
          message = msg;
          wait.complete();
        },
      );
      await wait.future;

      wait = Completer();
      channel.copyMessage(message, channel, onCompleted: (copied, err) async {
        expect(err, isNull);
        expect(copied, isNotNull);
        expect(message.message, equals(copied.message));
        expect(message.sender?.userId, equals(copied.sender?.userId));
        await Future.delayed(Duration(seconds: 1));
        wait.complete();
      });
      await wait.future;
    });

    test('should throw an exception with other channel message', () async {
      final params = OpenChannelParams()..operatorUserIds = [user.userId];
      final otherChannel = await OpenChannel.createChannel(params);
      await otherChannel.enter();

      final mParams = UserMessageParams(message: 'test');

      final wait = Completer();
      otherChannel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );
      await wait.future;

      expect(() => channel.copyMessage(message, otherChannel),
          throwsA(isA<SBError>()));
    });
  });

  group('openchannel translate user message', () {
    test('should succeed with target languages and user message', () async {
      final mParams = UserMessageParams(message: 'test');

      final wait = Completer();
      channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );
      await wait.future;

      final translated = await channel.translateUserMessage(message, ['ko']);
      expect(translated.translations.isEmpty, isFalse);
      expect(translated.translations.length, equals(1));
    });

    test('should throw an exception with invalid params', () async {
      final mParams = UserMessageParams(message: 'test');

      var wait = Completer();
      message = channel.sendUserMessage(
        mParams,
        onCompleted: (msg, error) async {
          message = msg;
          await Future.delayed(Duration(seconds: 1));
          wait.complete();
        },
      );

      await wait.future;

      expect(() => channel.translateUserMessage(message, []),
          throwsA(isA<InvalidParameterError>()));
    });
  });
}
