import 'dart:async';

import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late OpenChannel channel;
  late UserMessage message;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    final user = await testApiCreateUserAndConnect();

    final params = OpenChannelParams()..operatorUserIds = [user.userId];
    channel = await OpenChannel.createChannel(params);
    await channel.enter();

    final mParams = UserMessageParams(message: 'test');

    final wait = Completer();
    final pending = channel.sendUserMessage(
      mParams,
      onCompleted: (msg, error) {
        expect(msg, isNotNull);
        expect(msg.messageId, isNotNull);
        expect(msg.message, equals('test'));
        expect(msg.sendingStatus, MessageSendingStatus.succeeded);
        message = msg;
        wait.complete();
      },
    );
    expect(pending, isNotNull);
    expect(pending.sendingStatus, MessageSendingStatus.pending);

    await wait.future;
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('openchannel create message meta array', () {
    test('should success with valid meta arrays', () async {
      final result = await channel.addMessageMetaArray(message, [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2'])
      ]);

      final array = result.metaArrays ?? [];
      expect(array.length, equals(1));
      expect(array.first.key, equals('key1'));
      expect(array.first.value, equals(['value1', 'value2']));
    });

    test('should success with multiple meta arrays', () async {
      final result = await channel.addMessageMetaArray(message, [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6'])
      ]);

      final array = result.metaArrays ?? [];
      expect(array.length, equals(3));
      expect(array.first.key, equals('key1'));
      expect(array.first.value, equals(['value1', 'value2']));
      expect(array.last.key, equals('key3'));
      expect(array.last.value, equals(['value5', 'value6']));
    });
  });

  group('openchannel remove message meta array', () {
    test('should success to remove meta array values', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];
      final removeArrays = [
        MessageMetaArray(key: 'key1', value: ['value1']),
      ];
      var result = await channel.addMessageMetaArray(message, metaArrays);
      result = await channel.removeMessageMetaArray(message, removeArrays);
      final array = result.metaArrays ?? [];
      expect(array, isNotNull);
      expect(array.length, equals(3));
      expect(array.first.key, equals('key1'));
      expect(array.first.value.length, equals(1));
      expect(array.first.value, equals(['value2']));
    });

    test('should success to remove meta array multiple values ', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];
      final removeArrays = [
        MessageMetaArray(key: 'key1', value: ['value1']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      result = await channel.removeMessageMetaArray(message, removeArrays);
      final array = result.metaArrays ?? [];
      expect(array, isNotNull);
      expect(array.length, equals(3));
      expect(array.first.key, equals('key1'));
      expect(array.first.value.length, equals(1));
      expect(array.first.value, equals(['value2']));
      expect(array.last.key, equals('key3'));
      expect(array.last.value.length, equals(0));
    });
  });

  group('openchannel delete message meta array', () {
    test('should success to delete meta arrays', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      var array = result.metaArrays ?? [];
      expect(array.length, equals(3));

      result = await channel.deleteMessageMetaArrayKeys(message, [
        'key1',
        'key2',
      ]);
      array = result.metaArrays ?? [];
      expect(array.length, equals(1));
    });

    test('should success to delete partial meta arrays', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      var array = result.metaArrays ?? [];
      expect(array.length, equals(3));

      result =
          await channel.deleteMessageMetaArrayKeys(message, ['key2', 'key3']);
      array = result.metaArrays ?? [];
      expect(array.length, equals(1));
      expect(array.first.key, equals('key1'));
    });

    test('should throw an exception with invalid params', () async {
      final metaArrays = [
        MessageMetaArray(key: 'key1', value: ['value1', 'value2']),
        MessageMetaArray(key: 'key2', value: ['value3', 'value4']),
        MessageMetaArray(key: 'key3', value: ['value5', 'value6']),
      ];

      var result = await channel.addMessageMetaArray(message, metaArrays);
      expect(result.metaArrays?.length, equals(3));

      expect(() => channel.deleteMessageMetaArrayKeys(message, []),
          throwsA(isA<InvalidParameterError>()));
    });
  });
}
