import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late OpenChannel channel;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    final user = await testApiCreateUserAndConnect();
    final params = OpenChannelParams()..operatorUserIds = [user.userId];
    channel = await OpenChannel.createChannel(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});
  group('openchannel meta counter create', () {
    test('should success and return created meta counter', () async {
      final metaCounters = {'key1': 1};
      final result = await channel.createMetaCounters(metaCounters);
      //wait for a certain event happen
      expect(result, equals(metaCounters));
    });

    test(
        'should throw an exception and generated error with null or empty parameter',
        () async {
      expect(channel.createMetaCounters({}),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('openchannel meta counter get', () {
    test('should success and return counters with given keys', () async {
      final metaCounters = {'key1': 3, 'key2': 6, 'key3': 21};
      await channel.createMetaCounters(metaCounters);
      final result = await channel.getMetaCounters(['key2', 'key3']);
      final expected = {'key2': 6, 'key3': 21};
      expect(result, expected);
    });

    test('should success and return all meta counters', () async {
      final metaCounters = {'key1': 3, 'key2': 6};
      await channel.createMetaCounters(metaCounters);
      final result = await channel.getAllMetaCounters();
      expect(result, metaCounters);
    });

    test('should success and return empty', () async {
      final result = await channel.getAllMetaCounters();
      expect(result, {});
    });
  });

  group('openchannel meta counter update', () {
    test('should success and return updated meta counter', () async {
      final metaConters = {'key1': 1};
      await channel.createMetaCounters(metaConters);

      final updated = {'key1': 5};
      final result = await channel.updateMetaCounters(updated);

      expect(result, updated);
    });

    test('should success and return increased meta counter', () async {
      final metaConters = {'key1': 1};
      await channel.createMetaCounters(metaConters);

      final updated = {'key1': 5};
      final expected = {'key1': 6};
      final result = await channel.increaseMetaCounters(updated);

      expect(result, expected);
    });

    test('should success and return decreased meta counter', () async {
      final metaConters = {'key1': 1};
      await channel.createMetaCounters(metaConters);

      final updated = {'key1': 5};
      final expected = {'key1': -4};
      final result = await channel.decreaseMetaCounters(updated);

      expect(result, expected);
    });

    test('should throw error with null or empty parameter', () async {
      expect(channel.updateMetaCounters({}),
          throwsA(isA<InvalidParameterError>()));
    });
  });

  group('openchannel meta counter delete', () {
    test('should success and delete given meta counters key', () async {
      final metaCounters = {'key1': 1, 'key2': 3};
      await channel.createMetaCounters(metaCounters);
      await channel.deleteMetaCounters('key2');
      final result = await channel.getAllMetaCounters();
      final expected = {'key1': 1};
      expect(result, expected);
    });

    test('should throw an exception with non exist meta counters key',
        () async {
      final metaCounters = {'key1': 1, 'key2': 3};
      await channel.createMetaCounters(metaCounters);
      expect(
        channel.deleteMetaCounters('key3'),
        throwsA(isA<BadRequestError>()),
      );
    });

    test(
        'should throw an exception and generated error with null or empty parameter',
        () async {
      expect(
        channel.deleteMetaCounters(''),
        throwsA(isA<InvalidParameterError>()),
      );
    });
  });
}
