import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../../../common/sdksetup.dart';
import '../../../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late OpenChannel channel;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    final user = await testApiCreateUserAndConnect();
    final params = OpenChannelParams()..operatorUserIds = [user.userId];
    channel = await OpenChannel.createChannel(params);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('openchannel meta data create', () {
    test('should success and return created meta data', () async {
      final metaData = {'key1': 'value1'};
      final result = await channel.createMetaData(metaData);
      //wait for a certain event happen
      expect(result, equals(metaData));
    });
    test(
        'should throw an exception and generated error with null or empty parameter',
        () async {
      expect(channel.createMetaData({}),
          throwsA(isInstanceOf<InvalidParameterError>()));
    });
  });

  group('openchannel meta data update', () {
    test('should success and return updated meta data', () async {
      final metaData = {'key1': 'value1'};
      await channel.createMetaData(metaData);

      final updatedData = {'key1': 'value2'};
      final result = await channel.updateMetaData(updatedData);

      expect(result, updatedData);
    });

    test(
        'should throw an exception and generated error with null or empty parameter',
        () async {
      expect(channel.updateMetaData({}),
          throwsA(isInstanceOf<InvalidParameterError>()));
    });
  });

  group('openchannel meta data delete', () {
    test('should success and delete given metadata key', () async {
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await channel.createMetaData(metaData);
      await channel.deleteMetaData('key2');
      final result = await channel.getAllMetaData();
      final expected = {'key1': 'value1'};
      expect(result, expected);
    });

    test('should throw an exception with non exist metadata key', () async {
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await channel.createMetaData(metaData);
      expect(
        channel.deleteMetaData('key3'),
        throwsA(isInstanceOf<BadRequestError>()),
      );
    });

    test(
        'should throw an exception and generated error with null or empty parameter',
        () async {
      expect(
        channel.deleteMetaData(''),
        throwsA(isInstanceOf<InvalidParameterError>()),
      );
    });
  });

  group('openchannel meta data get', () {
    test('should success and return metadata with given keys', () async {
      final metaData = {'key1': 'value1', 'key2': 'value2', 'key3': 'value3'};
      await channel.createMetaData(metaData);
      final result = await channel.getMetaData(['key2', 'key3']);
      final expected = {'key2': 'value2', 'key3': 'value3'};
      expect(result, expected);
    });

    test('should success and return all meta data', () async {
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await channel.createMetaData(metaData);
      final result = await channel.getAllMetaData();
      expect(result, metaData);
    });

    test('should success and return empty', () async {
      final result = await channel.getAllMetaData();
      expect(result, {});
    });
  });

  group('openchannel meta data cached', () {
    test('should return proper value after create', () async {
      final metaData = {'key1': 'value1'};
      await channel.createMetaData(metaData);
      final cached = channel.getCachedMetaData();
      expect(cached.length, equals(1));
      expect(cached['key1'], equals('value1'));
    });

    test('should return proper value after update', () async {
      final metaData = {'key1': 'value1'};
      await channel.createMetaData(metaData);
      await channel.updateMetaData({'key1': 'value2'});
      final cached = channel.getCachedMetaData();
      expect(cached.length, equals(1));
      expect(cached['key1'], equals('value2'));
    });

    test('should return proper value after delete', () async {
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await channel.createMetaData(metaData);
      var cached = channel.getCachedMetaData();
      expect(cached.length, equals(2));
      expect(cached['key1'], equals('value1'));
      expect(cached['key2'], equals('value2'));
      await channel.deleteMetaData('key1');
      cached = channel.getCachedMetaData();
      expect(cached.length, equals(1));
      expect(cached['key1'], isNull);
    });

    test('should return proper value after getall', () async {
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await channel.createMetaData(metaData);
      await channel.getAllMetaData();
      final cached = channel.getCachedMetaData();
      expect(cached.length, equals(2));
      expect(cached['key1'], equals('value1'));
      expect(cached['key2'], equals('value2'));
    });

    test('should return proper value after removeall', () async {
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await channel.createMetaData(metaData);
      var cached = channel.getCachedMetaData();
      expect(cached.length, equals(2));
      expect(cached['key1'], equals('value1'));
      expect(cached['key2'], equals('value2'));
      await channel.deleteAllMetaData();
      cached = channel.getCachedMetaData();
      expect(cached.length, equals(0));
    });
  });
}
