import 'dart:async';

import 'package:sendbird_sdk/events/poll_update_event.dart';
import 'package:sendbird_sdk/events/poll_vote_event.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:stack_trace/stack_trace.dart';

import 'utils.dart';

class EventFuncName {
  static String onChannelDeleted = 'onChannelDeleted';
  static String onChannelHidden = 'onChannelHidden';
  static String onChannelFrozen = 'onChannelFrozen';
  static String onChannelUnfrozen = 'onChannelUnfrozen';
  static String onChannelChanged = 'onChannelChanged';
  static String onChannelMemberCountChanged = 'onChannelMemberCountChanged';
  static String onChannelParticipantCountChanged =
      'onChannelParticipantCountChanged';
  static String onChannelOperatorsUpdated = 'onChannelOperatorsUpdated';
  static String onMetaDataCreated = 'onMetaDataCreated';
  static String onMetaDataUpdated = 'onMetaDataUpdated';
  static String onMetaDataDeleted = 'onMetaDataDeleted';
  static String onMetaCountersCreated = 'onMetaCountersCreated';
  static String onMetaCountersUpdated = 'onMetaCountersUpdated';
  static String onMetaCountersDeleted = 'onMetaCountersDeleted';
  static String onDeliveryReceiptUpdated = 'onDeliveryReceiptUpdated';
  static String onMessageReceived = 'onMessageReceived';
  static String onMessageUpdated = 'onMessageUpdated';
  static String onMentionReceived = 'onMentionReceived';
  static String onMessageDeleted = 'onMessageDeleted';
  static String onReadReceiptUpdated = 'onReadReceiptUpdated';
  static String onTypingStatusUpdated = 'onTypingStatusUpdated';
  static String onThreadInfoUpdated = 'onThreadInfoUpdated';
  static String onUserJoined = 'onUserJoined';
  static String onUserLeaved = 'onUserLeaved';
  static String onUserEntered = 'onUserEntered';
  static String onUserExited = 'onUserExited';
  static String onUserMuted = 'onUserMuted';
  static String onUserUnmuted = 'onUserUnmuted';
  static String onUserBanned = 'onUserBanned';
  static String onUserUnbanned = 'onUserUnbanned';
  static String onUserReceivedInvitation = 'onUserReceivedInvitation';
  static String onUserDeclinedInvitation = 'onUserDeclinedInvitation';
  static String onReactionUpdated = 'onReactionUpdated';

  static String onSessionTokenRequired = 'onSessionTokenRequired';
  static String onSessionClosed = 'onSessionClosed';
  static String onSessionExpired = 'onSessionExpired';
  static String onSessionRefreshed = 'onSessionRefreshed';
  static String onSessionError = 'onSessionError';

  static String onReconnectionStarted = 'onReconnectionStarted';
  static String onReconnectionSucceeded = 'onReconnectionSucceeded';
  static String onReconnectionFailed = 'onReconnectionFailed';
  static String onReconnectionCanceled = 'onReconnectionCanceled';

  static String onPollUpdated = 'onPollUpdated';
  static String onPollVoted = 'onPollVoted';
}

class EventFuncResult {
  String fncName;
  List<dynamic> data;

  EventFuncResult(this.fncName, this.data);
}

class EventListener {
  String id = testApiRandomString();
  List<List<dynamic>> parameters = [];
  StreamController<String> streamController =
      StreamController<String>.broadcast(sync: true);
  List<EventFuncResult> results = [];
  List<String> eventNames = [];

  String get callerName =>
      Trace.current(1).frames[0].member?.split('.')[1] ?? 'unknown';

  Future<List<EventFuncResult>> waitFor(
    List<String> names, {
    int timeout = 10,
    bool cleanPrevious = true,
  }) async {
    if (cleanPrevious) {
      parameters = [];
      results = [];
    }

    final fncNames = List.from(names);
    final stream = streamController.stream;

    if (results.isNotEmpty && fncNames.isNotEmpty) {
      final hasAllEvents = List.from(fncNames)
          .map((e) => results.indexWhere((res) => res.fncName == e) != -1)
          .reduce((value, element) => value && element);

      if (hasAllEvents) {
        return results;
      }
    }

    //handle if events are already existed
    await for (String fncName in stream.timeout(Duration(seconds: timeout))) {
      // print('[CallBack] Call $fncName');
      var index = fncNames.indexOf(fncName);
      if (index >= 0) {
        fncNames.removeAt(index);
        var data = [];
        if (parameters.isNotEmpty) data = parameters.removeLast();
        // print('[CallBack] adding $fncName $data');
        results.add(EventFuncResult(fncName, data));
      }

      if (fncNames.isEmpty) break;
    }

    return results;
  }
}

class ChannelEventListener with EventListener, ChannelEventHandler {
  ChannelEventListener() {
    SendbirdSdk().addChannelEventHandler(id, this);
    // SendbirdSdk().addSessionHandler(this);
  }

  factory ChannelEventListener.waitFor(List<String> eventNames) {
    final listener = ChannelEventListener();
    listener.eventNames = List<String>.from(eventNames);
    return listener;
  }

  void cleanUp() {
    SendbirdSdk().removeChannelEventHandler(id);
    // SendbirdSdk().removeSessionHandler();
  }

  /// This method is to handle api vs socket event concurrently and because
  /// sometimes socket events come first before api response is get processed
  void addResult(EventFuncResult result) {
    final index = eventNames.indexWhere((e) => e == result.fncName);
    if (index != -1) {
      results.add(result);
      eventNames.removeAt(index);
    }
  }

  @override
  void onMessageReceived(BaseChannel channel, BaseMessage message) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, message]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, message]));
  }

  @override
  void onMessageUpdated(BaseChannel channel, BaseMessage message) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, message]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, message]));
  }

  @override
  void onMentionReceived(BaseChannel channel, BaseMessage message) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, message]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, message]));
  }

  @override
  void onMessageDeleted(BaseChannel channel, int messageId) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, messageId]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, messageId]));
  }

  @override
  void onReadReceiptUpdated(GroupChannel channel) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel]));
  }

  @override
  void onDeliveryReceiptUpdated(GroupChannel channel) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel]));
  }

  @override
  void onTypingStatusUpdated(GroupChannel channel) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel]));
  }

  @override
  void onUserReceivedInvitation(
    GroupChannel channel,
    List<User> invitees,
    User? inviter,
  ) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, invitees, inviter]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, invitees, inviter]));
  }

  @override
  void onUserDeclinedInvitation(
    GroupChannel channel,
    User invitee,
    User? inviter,
  ) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, invitee, inviter]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, invitee, inviter]));
  }

  @override
  void onUserJoined(GroupChannel channel, User user) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, user]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, user]));
  }

  @override
  void onUserLeaved(GroupChannel channel, User user) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, user]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, user]));
  }

  @override
  void onUserEntered(OpenChannel channel, User user) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, user]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, user]));
  }

  @override
  void onUserExited(OpenChannel channel, User user) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, user]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, user]));
  }

  @override
  void onUserMuted(BaseChannel channel, User user) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, user]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, user]));
  }

  @override
  void onUserUnmuted(BaseChannel channel, User user) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, user]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, user]));
  }

  @override
  void onUserBanned(BaseChannel channel, User user) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, user]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, user]));
  }

  @override
  void onUserUnbanned(BaseChannel channel, User user) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, user]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, user]));
  }

  @override
  void onChannelChanged(BaseChannel channel) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel]));
  }

  @override
  void onChannelMemberCountChanged(List<BaseChannel> channels) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add(channels);
    streamController.add(name);
    addResult(EventFuncResult(name, [channels]));
  }

  @override
  void onChannelParticipantCountChanged(List<OpenChannel> channels) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add(channels);
    streamController.add(name);
    addResult(EventFuncResult(name, [channels]));
  }

  @override
  void onChannelHidden(BaseChannel channel) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel]));
  }

  @override
  void onChannelDeleted(String channelUrl, ChannelType channelType) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channelUrl]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channelUrl]));
  }

  @override
  void onChannelFrozen(BaseChannel channel) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel]));
  }

  @override
  void onChannelUnfrozen(BaseChannel channel) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel]));
  }

  @override
  void onChannelOperatorsUpdated(BaseChannel channel) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel]));
  }

  @override
  void onMetaDataCreated(
    BaseChannel channel,
    Map<String, String> metaData,
  ) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, metaData]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, metaData]));
  }

  @override
  void onMetaDataUpdated(
    BaseChannel channel,
    Map<String, String> metaData,
  ) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, metaData]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, metaData]));
  }

  @override
  void onMetaDataDeleted(BaseChannel channel, List<String> metaDataKeys) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, metaDataKeys]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, metaDataKeys]));
  }

  @override
  void onMetaCountersCreated(
      BaseChannel channel, Map<String, int> metaCounters) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, metaCounters]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, metaCounters]));
  }

  @override
  void onMetaCountersUpdated(
      BaseChannel channel, Map<String, int> metaCounters) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, metaCounters]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, metaCounters]));
  }

  @override
  void onMetaCountersDeleted(
      BaseChannel channel, List<String> metaCounterKeys) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, metaCounterKeys]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, metaCounterKeys]));
  }

  @override
  void onThreadInfoUpdated(BaseChannel channel, ThreadInfoUpdateEvent event) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, event]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, event]));
  }

  @override
  void onReactionUpdated(BaseChannel channel, ReactionEvent event) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([channel, event]);
    streamController.add(name);
    addResult(EventFuncResult(name, [channel, event]));
  }

  @override
  void onPollUpdated(PollUpdateEvent event) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([event]);
    streamController.add(name);
    addResult(EventFuncResult(name, [event]));
  }

  @override
  void onPollVoted(PollVoteEvent event) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';

    parameters.add([event]);
    streamController.add(name);
    addResult(EventFuncResult(name, [event]));
  }
}

class SessionEventListener with EventListener, SessionEventHandler {
  SessionEventListener() {
    SendbirdSdk().addSessionEventHandler(this);
  }

  void cleanUp() {
    SendbirdSdk().removeSessionEventHandler();
  }

  @override
  void onSessionTokenRequired(Function(String) successCb, Function failCb) {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    parameters.add([successCb, failCb]);
    streamController.add(name);
  }

  /// Called when the SDK can't refresh the session.
  /// App should force a user to a login page to connect again.
  @override
  void onSessionClosed() {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    streamController.add(name);
  }

  /// Called when session key is expired.
  @override
  void onSessionExpired() {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    streamController.add(name);
  }

  /// Called when SDK refreshed the session key.
  @override
  void onSessionRefreshed() {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    streamController.add(name);
  }

  /// Called when the SDK run into an error while refreshing the session key
  @override
  void onSessionError(SBError error) {
    parameters.add([error]);
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    streamController.add(name);
  }
}

class ConnectionEventListener with EventListener, ConnectionEventHandler {
  ConnectionEventListener() {
    SendbirdSdk().addConnectionEventHandler(id, this);
  }

  void cleanUp() {
    SendbirdSdk().removeConnectionEventHandler(id);
  }

  @override
  void onReconnectionStarted() {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    streamController.add(name);
  }

  @override
  void onReconnectionSucceeded() {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    streamController.add(name);
  }

  @override
  void onReconnectionFailed() {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    streamController.add(name);
  }

  @override
  void onReconnectionCanceled() {
    final name = Trace.current().frames[0].member?.split('.')[1] ?? 'unknown';
    streamController.add(name);
  }
}
