import 'package:sendbird_sdk/request/channel/group_channel_create_request.dart';
import 'package:sendbird_sdk/request/messages/file_message_send_request.dart';
import 'package:sendbird_sdk/request/messages/user_message_send_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import 'callback.dart';
import 'environment.dart';
import 'helpers/private_api_client.dart';

String imageUrl = '''https://upload.wikimedia.org/wikipedia/commons/thumb/b/b6/
          Image_created_with_a_mobile_phone.png/1200px-Image_created_with_a_mobile_phone.png
      ''';

Environment env = Environment(type: EnvironmentType.staging);

Future<OpenChannel> createOpenChannel(
    {required OpenChannelParams params, required User user}) async {
  final event = ChannelEventListener();
  final params = OpenChannelParams()..operatorUserIds = [user.userId];
  final channel = await OpenChannel.createChannel(params);
  channel.enter();
  await event.waitFor([EventFuncName.onUserEntered]);
  event.cleanUp();
  return channel;
}

Future<GroupChannel> createGroupChannel(
  GroupChannelParams params, {
  List<String> additionalEvents = const [],
  bool shouldWait = true,
}) async {
  try {
    final joins =
        List.filled(params.userIds?.length ?? 0, EventFuncName.onUserJoined);
    final ops = params.operatorUserIds?.isNotEmpty == true &&
            params.isBroadcast == false &&
            params.isSuper == false
        ? <String>[EventFuncName.onChannelOperatorsUpdated]
        : <String>[];
    final invitations = (params.userIds?.length ?? 0) > 1 &&
            params.isBroadcast == false &&
            params.isSuper == false
        ? <String>[EventFuncName.onUserReceivedInvitation]
        : <String>[];

    final eventNames = invitations + ops + joins + additionalEvents;
    final event = ChannelEventListener.waitFor(shouldWait ? eventNames : []);
    final channel = await GroupChannel.createChannel(params);

    if (shouldWait) {
      await event.waitFor(eventNames, cleanPrevious: false);
      event.cleanUp();
    }

    return channel;
  } catch (e) {
    rethrow;
  }
}

Future<GroupChannel> createGroupChannelViaAPI(GroupChannelParams params) async {
  final channel = await PrivateApiClient()
      .send<GroupChannel>(GroupChannelCreateRequest(params));
  return channel;
}

Future<UserMessage> sendUserMessageViaAPI({
  required BaseChannel channel,
  required UserMessageParams params,
  required String senderId,
  bool shouldWaitForCallback = true,
}) async {
  final eventNames = [
    EventFuncName.onMessageReceived,
    if (channel is GroupChannel) EventFuncName.onChannelChanged,
  ];

  final event = ChannelEventListener.waitFor(eventNames);
  //NOTE: async order ? if this async is resolved after socket came, it stucks
  //becase event waitFor couldn't catch events
  final msg = await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
    channelType: channel.channelType,
    channelUrl: channel.channelUrl,
    params: params,
    senderId: senderId,
  ));

  if (SendbirdSdk().currentUser == null) return msg;

  if (shouldWaitForCallback) {
    await event.waitFor(eventNames, cleanPrevious: false);
    event.cleanUp();
    return msg;
  } else {
    event.cleanUp();
    return msg;
  }
}

Future<FileMessage> sendFileMessageViaAPI({
  required BaseChannel channel,
  required FileMessageParams params,
  required String senderId,
  bool shouldWaitForCallback = true,
}) async {
  final eventNames = [
    EventFuncName.onMessageReceived,
    EventFuncName.onChannelChanged,
  ];
  final event = ChannelEventListener.waitFor(eventNames);
  final msg = await PrivateApiClient().send(ChannelFileMessageSendApiRequest(
    channelType: channel.channelType,
    channelUrl: channel.channelUrl,
    params: params,
    senderId: senderId,
  ));

  if (SendbirdSdk().currentUser == null) return msg;

  if (channel is GroupChannel && shouldWaitForCallback) {
    final results = await event.waitFor(eventNames, cleanPrevious: false);
    event.cleanUp();
    return (results.last.data.first as GroupChannel).lastMessage as FileMessage;
  } else {
    event.cleanUp();
    return msg;
  }
}

Map<String, dynamic> generatePushPayload({
  required String appId,
  String? sessionKey,
  required BaseMessage message,
  required String channelName,
  required User recipient,
}) {
  var alert = '', category = '', type = '';

  if (message is UserMessage) {
    alert = message.message;
    category = 'sendbird_mesg';
    type = 'MESG';
  } else if (message is FileMessage) {
    alert = 'file';
    category = 'sendbird_file';
    type = 'FILE';
  } else if (message is AdminMessage) {
    alert = message.message;
    category = 'sendbird_admin';
    type = 'ADMM';
  }

  return {
    'apn': {
      'alert': alert,
      'badge': 0,
      'category': category,
      'mutable-content': 1,
      'sound': 'test.swf',
    },
    'sendbird': {
      'app_id': appId,
      'channel': {
        'channel_url': message.channelUrl,
        'channel_type': message.channelType.urlString,
        'name': channelName,
      },
      'type': type,
      'unread_message_count': 10,
      'created_at': message.createdAt,
      'recipient': {
        'id': recipient.userId,
        'name': recipient.nickname,
      },
      'session_key': {
        'key': sessionKey,
        'topics': ['delivery_receipt']
      }
    }
  };
}
