import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:sendbird_sdk/services/network/api_client.dart';

import '../common.dart';

import '../utils.dart';

class PrivateApiClient extends ApiClient {
  static final PrivateApiClient _singleton = PrivateApiClient._internal();
  factory PrivateApiClient() {
    _singleton.client
      ..isLocal = false
      ..bypassAuth = true;
    return _singleton;
  }

  PrivateApiClient._internal();

  Future<Map> createApplication() async {
    final body = {
      'name': 'TEST_APP',
      'plan': 'enterprise',
      'use_new_ops': true,
      'unread_cnt_subscription_all': true,
      'unread_cnt_subscription_custom_type': [
        'CUSTOM_TYPE_A',
        'CUSTOM_TYPE_B',
        'CUSTOM_TYPE_C'
      ],
      'use_group_channel_changelogs': true,
      'translation_vendor': true,
      'push_enabled': true,
      'plan_premiums': ['delivery_receipt'],
      'enable': true
    };
    final headers = {
      'Cache-Control': 'no-cache',
      'Authorization': env.authHeader!,
    };
    final res = await client.post(
      url: '/v3/applications/',
      body: body,
      headers: headers,
    );
    return res;
  }

  Future<void> updateAppAttributes(
    String appId,
    Map<String, dynamic> attributes,
  ) async {
    await client.put(
      url: '/v3/secret/$appId/attrs',
      body: attributes,
      headers: {
        'Secret-Key': env.secret!,
      },
    );
  }

  Future<void> updateAppAttribute({
    required String appId,
    required String attributeName,
    dynamic value,
  }) async {
    await client.put(
      url: '/v3/secret/$appId/attrs/$attributeName',
      body: {'value': value},
      headers: {
        'Secret-Key': env.secret!,
      },
    );
  }

  Future<void> updateAppPlanAttributes(
    String appId,
    Map<String, dynamic> attributes,
  ) async {
    await client.post(
      url: '/v3/secret/$appId/plan',
      body: attributes,
      headers: {
        'Secret-Key': env.secret!,
      },
    );
  }

  Future<void> updateReactionEnable(String appId, bool enable) async {
    await client.put(
      url: '/v3/secret/$appId/attrs/reactions',
      body: {'value': enable},
      headers: {
        'Secret-Key': env.secret!,
      },
    );
  }

  Future<User> createUser({
    String userId = '',
    String nickname = '',
    String profileUrl = '',
    bool accessToken = false,
    bool issueSessionToken = false,
    int? sessionTokenExpiresAt,
  }) async {
    if (userId.isEmpty) {
      userId = testApiRandomString();
    }
    if (nickname.isEmpty) {
      nickname = testApiRandomString();
    }

    final body = {
      'user_id': userId,
      'nickname': nickname,
      'profile_url': profileUrl,
      'is_access_token': accessToken,
      'issue_session_token': issueSessionToken,
      if (sessionTokenExpiresAt != null)
        'session_token_expires_at': sessionTokenExpiresAt,
    };

    final res = await client.post(url: '/v3/users', body: body);
    return User.fromJson(res);
  }

  Future<EmojiCategory> createEmojiCategory({
    required String categoryName,
    required String categoryUrl,
  }) async {
    final body = {
      'emoji_categories': [
        {
          'name': categoryName,
          'url': categoryUrl,
        }
      ]
    };

    final res = await client.post(url: '/v3/emoji_categories', body: body);
    final ary = List.from(res['emoji_categories']);
    return EmojiCategory.fromJson(ary.first);
  }

  Future<Emoji> createEmoji({
    required int categoryId,
    required String key,
    required String emojiUrl,
  }) async {
    final body = {
      'emoji_category_id': categoryId,
      'emojis': [
        {
          'key': key,
          'url': emojiUrl,
        }
      ]
    };

    final res = await client.post(url: '/v3/emojis', body: body);
    return Emoji.fromJson(res['emojis'].first);
  }

  Future<void> clearOGTagUrlCache(List<String> urls) async {
    final body = {'urls': urls};
    await client.put(url: '/v3/internal/clear_og_tag_cache', body: body);
  }

  Future<AdminMessage> sendAdminMessage({
    required ChannelType channelType,
    required String channelURL,
    required String message,
  }) async {
    final body = {
      'message_type': 'ADMM',
      'message': message,
    };
    final res = await client.post(
        url: '/v3/${channelType.urlString}/$channelURL/messages', body: body);

    return BaseMessage.msgFromJson<AdminMessage>(res)!;
  }

  Future<bool> setChannelInvitationPreference(
    String userId,
    bool autoAccept,
  ) async {
    final res = await client.put(
      url: '/v3/users/$userId/channel_invitation_preference',
      body: {'auto_accept': autoAccept},
    );

    return res['auto_accept'];
  }
}
