import 'package:sendbird_sdk/core/models/emoji.dart';
import 'package:sendbird_sdk/core/models/user.dart';
import 'package:sendbird_sdk/request/abstract/api_request.dart';
import 'package:sendbird_sdk/services/network/http_client.dart';
import 'common.dart';
import 'utils.dart';

const create_app = '/v3/applications/';
const clear_og_cache = '/v3/internal/clear_og_tag_cache';
const app_plan_attributes = '/v3/secret/{#}/plan';
const app_attributes = '/v3/secret/{#}/attrs';
const app_attribute_name = '/v3/secret/{#}/attrs/{#}';
const app_reaction = '/v3/secret/{#}/attrs/reactions';
const create_user = '/v3/users';
const create_emoji_category = '/v3/emoji_categories';
const create_emoji = '/v3/emojis';

class AppCreateRequest extends ApiRequest {
  @override
  final method = Method.post;

  AppCreateRequest() : super(bypassAuth: true) {
    url = 'applications';
    body = {
      'name': 'TEST_APP',
      'plan': 'enterprise',
      'use_new_ops': true,
      'unread_cnt_subscription_all': true,
      'unread_cnt_subscription_custom_type': [
        'CUSTOM_TYPE_A',
        'CUSTOM_TYPE_B',
        'CUSTOM_TYPE_C'
      ],
      'use_group_channel_changelogs': true,
      'translation_vendor': true,
      'push_enabled': true,
      'plan_premiums': ['delivery_receipt'],
      'enable': true
    };
    headers = {
      'Cache-Control': 'no-cache',
      'Authorization': env.authHeader!,
    };
  }

  @override
  Future response(res) async {
    return res;
  }
}

class AppAttributesUpdateRequest extends ApiRequest {
  @override
  final method = Method.put;

  AppAttributesUpdateRequest(
    String appId,
    Map<String, dynamic> attributes,
  ) : super(token: env.token, bypassAuth: true) {
    url = 'secret/$appId/attrs';
    body = attributes;
    headers = {'Secret-Key': env.secret!};
  }
}

class AppAttributeUpdateRequest extends ApiRequest {
  @override
  final method = Method.put;

  AppAttributeUpdateRequest({
    required String appId,
    required String attributeName,
    required dynamic value,
  }) : super(token: env.token, bypassAuth: true) {
    url = 'secret/$appId/attrs/$attributeName';
    body = {'value': value};
    headers = {'Secret-Key': env.secret!};
  }
}

class AppPlanAttributesUpdateRequest extends ApiRequest {
  @override
  final method = Method.post;

  AppPlanAttributesUpdateRequest(
    String appId,
    Map<String, dynamic> attributes,
  ) : super(token: env.token, bypassAuth: true) {
    url = 'secret/$appId/plan';
    body = attributes;
    headers = {'Secret-Key': env.secret!};
  }
}

class AppReactionEnableRequest extends ApiRequest {
  @override
  final method = Method.put;

  AppReactionEnableRequest(String appId, bool enabled)
      : super(token: env.token, bypassAuth: true) {
    url = 'secret/$appId/attrs/reactions';
    body = {'value': enabled};
    headers = {'Secret-Key': env.secret!};
  }
}

class AppUserCreateRequest extends ApiRequest {
  @override
  final method = Method.post;

  AppUserCreateRequest({
    String userId = '',
    String nickname = '',
    String profileUrl = '',
    bool accessToken = false,
    bool issueSessionToken = false,
    int? sessionTokenExpiresAt,
  }) : super(token: env.token, bypassAuth: true) {
    if (userId.isEmpty) {
      userId = testApiRandomString();
    }
    if (nickname.isEmpty) {
      nickname = testApiRandomString();
    }

    url = '';
    body = {
      'user_id': userId,
      'nickname': nickname,
      'profile_url': profileUrl,
      'is_access_token': accessToken,
      'issue_session_token': issueSessionToken,
      if (sessionTokenExpiresAt != null)
        'session_token_expires_at': sessionTokenExpiresAt,
    };
  }

  @override
  Future<User> response(res) async {
    return User.fromJson(res);
  }
}

class AppEmojiCreateRequest extends ApiRequest {
  @override
  final method = Method.post;

  AppEmojiCreateRequest({
    required int categoryId,
    required String key,
    required String emojiUrl,
  }) : super(token: env.token, bypassAuth: true) {
    url = 'emojis';
    body = {
      'emoji_category_id': categoryId,
      'emojis': [
        {
          'key': key,
          'url': emojiUrl,
        }
      ]
    };
  }

  @override
  Future<Emoji> response(res) async {
    return Emoji.fromJson(res['emojis'].first);
  }
}

class AppEmojiCategoryCreateRequest extends ApiRequest {
  @override
  final method = Method.post;

  AppEmojiCategoryCreateRequest({
    required String categoryName,
    required String categoryUrl,
  }) : super(token: env.token, bypassAuth: true) {
    url = 'emoji_categories';
    body = {
      'emoji_categories': [
        {
          'name': categoryName,
          'url': categoryUrl,
        }
      ]
    };
  }

  @override
  Future<EmojiCategory> response(res) async {
    final ary = List.from(res['emoji_categories']);
    return EmojiCategory.fromJson(ary.first);
  }
}

class AppOGTagClearRequest extends ApiRequest {
  @override
  final method = Method.put;

  AppOGTagClearRequest(List<String> urls)
      : super(token: env.token, bypassAuth: true) {
    url = 'internal/clear_og_tag_cache';
    body = {'urls': urls};
  }
}
