import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

void main() {
  group('message type', () {
    test('file message toJson should generate json', () {
      final json = {
        'url': 'http://google.com',
        'type': 'image/jpeg',
        'channel_url':
            'sendbird_group_channel_31435542_3004b5cc323e70f336aabc9ca147d2eedd6b21a8'
      };

      final msg = FileMessage.fromJson(json);
      expect(msg.channelUrl, isNotNull);
      expect(msg.type, equals('image/jpeg'));
      expect(msg.url, equals('http://google.com'));
    });

    test('user message toJson should generate json', () {
      final sender = Sender(
        userId: 'user1',
        nickname: 'user1',
        connectionStatus: UserConnectionStatus.offline,
        metaData: {'aa': 'bb'},
        preferredLanguages: ['ko', 'en'],
      );

      final message = UserMessage(
        translations: <String, String>{},
        channelType: ChannelType.group,
        channelUrl: '123',
        requestId: '123',
        messageId: 123,
        message: 'message',
        sender: sender,
        sendingStatus: MessageSendingStatus.succeeded,
        mentionedUsers: [
          User(
            userId: '12345',
            nickname: 'user1',
            profileUrl: 'https://naver.com',
          ),
          User(
            userId: '12346',
            nickname: 'user1',
            profileUrl: 'https://naver.com',
          ),
        ],
        mentionType: MentionType.users,
        requestedMentionUserIds: ['12345', '12346'],
        createdAt: 913121,
        updatedAt: 192312,
        parentMessageId: 121,
        parentMessageText: 'parent',
        threadInfo: ThreadInfo(replyCount: 0, updatedAt: 123),
        metaArrays: [
          MessageMetaArray(key: 'key', value: ['value1']),
          MessageMetaArray(key: 'key2', value: ['value2']),
        ],
        customType: 'custom',
        messageSurvivalSeconds: 123,
        isOperatorMessage: false,
        reactions: [
          Reaction(key: 'happy', userIds: ['1234', '5678'], updatedAt: 0),
          Reaction(key: 'sad', userIds: ['1234', '5678'], updatedAt: 0),
        ],
      );

      final res = message.toJson();
      expect(res['channel_type'], equals('group'));
      expect(res['channel_url'], equals('123'));
      expect(res['message_id'], equals(123));
      expect(res['sending_status'], equals('succeeded'));
      expect(
          res['mentioned_users'],
          equals([
            {
              'user_id': '12345',
              'nickname': 'user1',
              'profile_url': 'https://naver.com',
              'is_online': null,
              'last_seen_at': null,
              'is_active': true,
              'preferred_languages': null,
              'friend_discovery_key': null,
              'friend_name': null,
              'discovery_keys': null,
              'metadata': {},
              'require_auth_for_profile_image': false,
              'session_token': null
            },
            {
              'user_id': '12346',
              'nickname': 'user1',
              'profile_url': 'https://naver.com',
              'is_online': null,
              'last_seen_at': null,
              'is_active': true,
              'preferred_languages': null,
              'friend_discovery_key': null,
              'friend_name': null,
              'discovery_keys': null,
              'metadata': {},
              'require_auth_for_profile_image': false,
              'session_token': null
            }
          ]));
      expect(
          res['sorted_metaarray'],
          equals([
            {
              'key': 'key',
              'value': ['value1']
            },
            {
              'key': 'key2',
              'value': ['value2']
            }
          ]));
      expect(
          res['reactions'],
          equals([
            {
              'key': 'happy',
              'user_ids': ['1234', '5678'],
              'updated_at': 0,
            },
            {
              'key': 'sad',
              'user_ids': ['1234', '5678'],
              'updated_at': 0,
            }
          ]));
    });
    // test('admin message toJson should generate json', () {});
  });

  group('channel type', () {
    test('group channel toJson should generate json', () {
      final sender = Sender(
        userId: 'user1',
        nickname: 'user1',
        connectionStatus: UserConnectionStatus.offline,
        metaData: {'aa': 'bb'},
        preferredLanguages: ['ko', 'en'],
      );

      final msg = UserMessage(
        translations: <String, String>{},
        channelType: ChannelType.group,
        channelUrl: '123',
        requestId: '123',
        messageId: 123,
        message: 'message',
        sender: sender,
      );

      final channel = GroupChannel(
        channelUrl: '123',
        lastMessage: msg,
        isSuper: false,
        isStrict: false,
        isBroadcast: false,
        isPublic: true,
        isDistinct: true,
        isDiscoverable: true,
        unreadMessageCount: 10,
        unreadMentionCount: 1,
        members: [
          Member(
            userId: '12345',
            nickname: 'user1',
            profileUrl: 'https://naver.com',
            state: MemberState.joined,
            role: Role.chat_operator,
          ),
          Member(
            userId: '12346',
            nickname: 'user1',
            profileUrl: 'https://naver.com',
            state: MemberState.invited,
            role: Role.none,
          ),
        ],
        memberCount: 2,
        joinedMemberCount: 1,
        myPushTriggerOption: GroupChannelPushTriggerOption.mentionOnly,
        myMemberState: MemberState.joined,
        myRole: Role.chat_operator,
        myMutedState: MuteState.unmuted,
        myCountPreference: CountPreference.mentionOnly,
        isHidden: false,
        data: 'data',
        customType: 'custom',
        name: 'channel name',
      );

      final res = channel.toJson();

      expect(res['channel_url'], equals('123'));
      expect((res['last_message'] as Map)['message'], equals('message'));
      expect(res['is_super'], isFalse);
      expect(res['is_strict'], isFalse);
      expect(res['is_public'], isTrue);
      expect(res['is_discoverable'], isTrue);
      expect(res['is_broadcast'], isFalse);
      expect(res['unread_message_count'], equals(10));
      expect(res['unread_mention_count'], equals(1));
      expect(res['member_count'], equals(2));
      expect(res['joined_member_count'], equals(1));
      expect(res['push_trigger_option'], equals('mention_only'));
      expect(res['my_role'], equals('operator'));
      expect(res['count_preference'], equals('unread_mention_count_only'));
      expect(
          res['members'],
          equals([
            {
              'user_id': '12345',
              'nickname': 'user1',
              'profile_url': 'https://naver.com',
              'is_online': null,
              'last_seen_at': null,
              'is_active': true,
              'preferred_languages': null,
              'friend_discovery_key': null,
              'friend_name': null,
              'discovery_keys': null,
              'metadata': {},
              'require_auth_for_profile_image': false,
              'session_token': null,
              'state': 'joined',
              'is_blocked_by_me': false,
              'is_blocking_me': false,
              'is_muted': false,
              'role': 'operator'
            },
            {
              'user_id': '12346',
              'nickname': 'user1',
              'profile_url': 'https://naver.com',
              'is_online': null,
              'last_seen_at': null,
              'is_active': true,
              'preferred_languages': null,
              'friend_discovery_key': null,
              'friend_name': null,
              'discovery_keys': null,
              'metadata': {},
              'require_auth_for_profile_image': false,
              'session_token': null,
              'state': 'invited',
              'is_blocked_by_me': false,
              'is_blocking_me': false,
              'is_muted': false,
              'role': 'none'
            }
          ]));
    });
    // test('open channel toJson should generate json', () {});
  });
}
