import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:sendbird_sdk/utils/string_utils.dart';

import '../common/common.dart';
import '../common/sdksetup.dart';
import '../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late User user;
  late List<User> users;
  late List<GroupChannel> channels;

  setUpAll(() async {
    testInfo = await testCreateApp();
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDownAll(() async {
    await sdk.disconnect();
  });

  group('group channel list query', () {
    setUpAll(() async {
      final u1 = await testApiCreateUser(userId: 'bob', nickname: 'bob');
      final u2 = await testApiCreateUser(userId: 'john', nickname: 'john');
      final u3 =
          await testApiCreateUser(userId: 'danielle', nickname: 'danielle');
      users = [u1, u2, u3];
      user = await testApiConnect(u1.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isPublic = true
        ..isDiscoverable = true
        ..name = 'channel 1'
        ..customType = 'custom 1'
        ..userIds = [u1.userId, u2.userId];
      final channel = await createGroupChannel(params);

      final params2 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isSuper = true
        ..name = 'channel 2'
        ..customType = 'custom 1'
        ..userIds = [u2.userId];
      final channel2 = await createGroupChannel(params2);

      final params3 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..name = 'sendbird 1'
        ..customType = 'sendbird 1'
        ..userIds = users.map((e) => e.userId).toList();
      final channel3 = await createGroupChannel(params3);

      final params4 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..name = 'sendbird 2'
        ..customType = 'sendbird 2'
        ..userIds = [u1.userId, u2.userId];
      final channel4 = await createGroupChannel(params4);

      final params5 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..name = 'channel sendbird'
        ..customType = 'custom 2'
        ..userIds = [u1.userId, u3.userId];
      final channel5 = await createGroupChannel(params5);

      final params6 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..name = 'sendbird channel 1'
        ..customType = 'custom 3'
        ..userIds = users.map((e) => e.userId).toList();
      final channel6 = await createGroupChannel(params6);

      channels = [channel, channel2, channel3, channel4, channel5, channel6];
    });

    test('should return channel with all properties set properly', () async {
      final query = GroupChannelListQuery();
      query.limit = 1;
      query.publicChannelFilter = PublicChannelFilter.public;
      query.customTypes = ['custom 1'];
      final result = await query.loadNext();
      expect(result.length, equals(1));

      final channel = result.first;
      expect(channel.isPublic, isTrue);
      expect(channel.isDiscoverable, isTrue);
      expect(channel.isBroadcast, isFalse);
      expect(channel.hiddenState, GroupChannelHiddenState.unhidden);
      expect(channel.myCountPreference, CountPreference.all);
      expect(channel.myMemberState, MemberState.joined);
      expect(channel.isHidden, isFalse);
      expect(channel.isFrozen, isFalse);
      expect(channel.isStrict, isFalse);
      expect(channel.isEphemeral, isFalse);
      expect(channel.lastMessage, isNull);
      expect(channel.members.length, equals(2));
      expect(channel.name, equals('channel 1'));
      expect(channel.customType, equals('custom 1'));
    });

    test('should return all existing group channels', () async {
      final query = GroupChannelListQuery();
      query.limit = 5;
      final result = await query.loadNext();
      expect(result.length, equals(5));
    });

    test('should return correct channels with userIdExact filter', () async {
      final query = GroupChannelListQuery();
      query.userIdsExactlyIn = [users[0].userId, users[1].userId];
      final result = await query.loadNext();
      expect(result.length, equals(3));
    });

    test('should return correct channels with userId include filter', () async {
      final query = GroupChannelListQuery();
      query.setUserIdsIncludeFilter(
        [user.userId, users.last.userId],
        GroupChannelListQueryType.or,
      );
      final result = await query.loadNext();
      expect(result.length, equals(6));

      final query2 = GroupChannelListQuery();
      query2.setUserIdsIncludeFilter(
        [user.userId, users.last.userId],
        GroupChannelListQueryType.and,
      );
      final result2 = await query2.loadNext();
      expect(result2.length, equals(3));
    });

    test('should return correct channels with custom type start filter',
        () async {
      final query = GroupChannelListQuery();
      query.customTypeStartWith = 'sendbird';
      final result = await query.loadNext();
      expect(result.length, equals(2));
    });

    test('should return correct channels with custom types filter', () async {
      final query = GroupChannelListQuery();
      query.customTypes = ['sendbird 1', 'custom 1'];
      final result = await query.loadNext();
      expect(result.length, equals(3));
    });

    test('should return correct channels with super channel filter', () async {
      final query = GroupChannelListQuery();
      query.superChannelFilter = SuperChannelFilter.superChannel;
      final result = await query.loadNext();
      expect(result.length, equals(1));
    });

    test('should return along with metadata with includeMetadata option',
        () async {
      await channels[0].createMetaData({'key1': 'value1'});
      await channels[1].createMetaData({'key2': 'value2'});

      final query = GroupChannelListQuery();
      query.customTypes = ['custom 1'];
      query.includeMetaData = true; //default is true
      query.limit = 5;
      final result = await query.loadNext();
      expect(result.length, equals(2));
      expect(result.first.getCachedMetaData(), isNotNull);
      expect(result.first.getCachedMetaData().length, equals(1));
      expect(result.last.getCachedMetaData(), isNotNull);
      expect(result.last.getCachedMetaData().length, equals(1));
    });

    test('should contain values for user properties', () async {
      await user.updateMetaData({'key1': 'value1'});

      final query = GroupChannelListQuery();
      query.setUserIdsExactFilter([user.userId, users[1].userId]);
      final res = await query.loadNext();
      expect(res, isNotNull);
      expect(res.length, equals(3));

      final u1 = res.first.members.where((e) => e.userId == user.userId).first;
      expect(u1.metaData.isNotEmpty, isTrue);
    });

    test('should hasNext to be false after fetch all channels', () async {
      final query = GroupChannelListQuery();
      query.limit = 5;
      var result = await query.loadNext();
      expect(result.length, equals(5));
      expect(query.hasNext, isTrue);
      result = await query.loadNext();
      expect(result.length, equals(1));
      expect(query.hasNext, isFalse);
    });

    test('should return all channels after fetch hiddenstatefilter all',
        () async {
      final query = GroupChannelListQuery()
        ..channelHiddenStateFilter = ChannelHiddenStateFilter.all;

      var result = await query.loadNext();
      expect(result.length, equals(6));
    });

    test('should return json from toJson', () {
      final query = GroupChannelListQuery()
        ..channelUrls = ['123', '456']
        ..includeEmptyChannel = false
        ..limit = 12
        ..includeMemberList = false
        ..channelHiddenStateFilter =
            ChannelHiddenStateFilter.hiddenAllowAutoUnhide;

      final res = query.toJson();

      expect(res['channel_urls'], equals(['123', '456']));
      expect(res['include_member_list'], isFalse);
      expect(res['include_empty_channel'], isFalse);
      expect(res['limit'], equals(12));
      expect(
        res['channel_hidden_state_filter'],
        equals('hidden_allow_auto_hide'),
      );
    });
  });

  group('group channel list query with meta data filter', () {
    final randomString = getRandomString(10);
    final randomKey = randomString + '1';
    final randomKey2 = randomString + '2';

    setUpAll(() async {
      final u1 = await testApiCreateUser(userId: 'bob1', nickname: 'bob');
      final u2 = await testApiCreateUser(userId: 'john1', nickname: 'john');
      final u3 =
          await testApiCreateUser(userId: 'danielle1', nickname: 'danielle');
      users = [u1, u2, u3];
      user = await testApiConnect(u1.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [u1.userId, u2.userId];
      final channel = await createGroupChannel(params);
      await channel.createMetaData({randomKey: 'v1'});

      final params2 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [u2.userId];
      final channel2 = await createGroupChannel(params2);
      await channel2.createMetaData({randomKey: 'v2'});

      final params3 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      final channel3 = await createGroupChannel(params3);
      await channel3.createMetaData({randomKey2: 'v3'});

      channels = [channel, channel2, channel3];
    });

    test('should return channel(s) with matching metaData key', () async {
      final query = GroupChannelListQuery()
        ..limit = 5
        ..includeMetaData = true
        ..setMetaDataFilterWithValues(randomKey, ['v1']);

      var result = await query.loadNext();
      expect(result.length, equals(1));
      expect(query.hasNext, isFalse);
    });

    test('should return channel(s) with matching metaData values', () async {
      final query = GroupChannelListQuery()
        ..limit = 5
        ..includeMetaData = true
        ..setMetaDataFilterWithValues(randomKey, ['v1', 'v2']);

      var result = await query.loadNext();
      expect(result.length, equals(2));
      expect(query.hasNext, isFalse);
    });

    test('should return channel(s) with matching metaData value startsWith',
        () async {
      final query = GroupChannelListQuery()
        ..limit = 5
        ..includeMetaData = true
        ..setMetaDataFilterWithStartWith(randomKey, 'v');

      var result = await query.loadNext();
      expect(result.length, equals(2));
      expect(query.hasNext, isFalse);
    });
  });

  group('group channel list query with super channel filter', () {
    setUpAll(() async {
      final u1 = await testApiCreateUser(userId: 'bob2', nickname: 'bob');
      final u2 = await testApiCreateUser(userId: 'john2', nickname: 'john');
      final u3 =
          await testApiCreateUser(userId: 'danielle2', nickname: 'danielle');
      users = [u1, u2, u3];
      user = await testApiConnect(u1.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isSuper = true
        ..name = 'channel 1'
        ..customType = 'custom 1'
        ..userIds = [u1.userId, u2.userId];
      final channel = await createGroupChannel(params);

      final params2 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isSuper = true
        ..isBroadcast = true
        ..name = 'channel 2'
        ..customType = 'custom 1'
        ..userIds = [u2.userId];
      final channel2 = await createGroupChannel(params2);

      final params3 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..name = 'sendbird 1'
        ..customType = 'sendbird 1'
        ..userIds = users.map((e) => e.userId).toList();
      final channel3 = await createGroupChannel(params3);

      channels = [channel, channel2, channel3];
    });

    test('should return broadcast channel only', () async {
      final query = GroupChannelListQuery();
      query.superChannelFilter = SuperChannelFilter.broadcastOnly;

      final res = await query.loadNext();
      expect(res.length, 1);
      expect(res.first.name, 'channel 2');
    });

    test('should return non super channel only', () async {
      final query = GroupChannelListQuery();
      query.superChannelFilter = SuperChannelFilter.nonsuperChannel;

      final res = await query.loadNext();
      expect(res.length, 1);
      expect(res.first.name, 'sendbird 1');
    });

    test('should return super channel only', () async {
      final query = GroupChannelListQuery();
      query.superChannelFilter = SuperChannelFilter.superChannel;

      final res = await query.loadNext();
      expect(res.length, 2);
    });
  });
}
