import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../common/common.dart';
import '../common/sdksetup.dart';
import '../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDownAll(() async {
    await sdk.disconnect();
  });

  group('groupchannel member list query', () {
    late List<User> users;
    late User op, other;
    late GroupChannel channel;

    setUpAll(() async {
      users = await testApiCreateUsers(3);
      op = await testApiCreateUser(nickname: 'testuser1');
      other = await testApiCreateUser(nickname: 'other1');
    });

    setUp(() async {
      await testApiConnect(op.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [op.userId]
        ..userIds =
            users.map((e) => e.userId).toList() + [op.userId, other.userId];
      channel = await createGroupChannel(params);
    });

    test('should return all members', () async {
      final query = GroupChannelMemberListQuery(channelUrl: channel.channelUrl);
      final res = await query.loadNext();

      expect(res, isNotNull);
      expect(res.length, 5);
    });

    test('should return only operators', () async {
      final query = GroupChannelMemberListQuery(channelUrl: channel.channelUrl)
        ..operatorFilter = OperatorFilter.operator;
      final res = await query.loadNext();
      expect(res, isNotNull);
      expect(res.length, 1);
      expect(res.first.userId, equals(op.userId));
    });

    test('should return only match members with nickname', () async {
      final query = GroupChannelMemberListQuery(channelUrl: channel.channelUrl)
        ..nicknameStartsWith = 'other';

      final res = await query.loadNext();
      expect(res, isNotNull);
      expect(res.length, 1);
      expect(res.first.userId, equals(other.userId));
    });

    test('should return json from toJson', () {
      final query = GroupChannelMemberListQuery(channelUrl: '123')
        ..memberStateFilter = MemberStateFilter.all
        ..nicknameStartsWith = 'test'
        ..order = MemberListOrder.nicknameAlphabetical
        ..limit = 15;

      final res = query.toJson();

      expect(res['channel_url'], equals('123'));
      expect(res['member_state_filter'], equals('all'));
      expect(res['nickname_starts_with'], equals('test'));
      expect(res['limit'], equals(15));
    });
  });
}
