import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../common/common.dart';
import '../common/sdksetup.dart';
import '../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late User user;
  late List<User> users;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    user = await testApiCreateUserAndConnect();
    users = await testApiCreateUsers(3);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('previous message list for group channel', () {
    late GroupChannel channel;
    late UserMessage msg1, msg2;

    setUp(() async {
      final gparams = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      channel = await createGroupChannel(gparams);

      final params = UserMessageParams(message: 'test');
      msg1 = await sendUserMessageViaAPI(
        channel: channel,
        params: params..message = 'this is test',
        senderId: user.userId,
      );

      await sendUserMessageViaAPI(
        channel: channel,
        params: params..message = 'testing more',
        senderId: users[0].userId,
      );

      await sendUserMessageViaAPI(
        channel: channel,
        params: params..message = 'hi there',
        senderId: users[2].userId,
      );

      msg2 = await sendUserMessageViaAPI(
        channel: channel,
        params: params..message = 'last message',
        senderId: users[1].userId,
      );
    });

    test('should return all previous messages', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      );

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(4));
      expect(messages.first.channelType, equals(channel.channelType));
      expect(messages[0].messageId, equals(msg1.messageId));
      expect(messages[3].messageId, equals(msg2.messageId));
    });

    test('should return only limited message', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )..limit = 2;

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(2));
      expect(messages[0].message, equals('hi there'));
      expect(messages[1].message, equals('last message'));

      final messages2 = await query.loadNext();
      expect(messages2, isNotNull);
      expect(messages2.length, equals(2));
      expect(messages2[0].message, equals('this is test'));
      expect(messages2[1].message, equals('testing more'));
    });

    test('should return only limited message in reverse', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )
        ..limit = 2
        ..reverse = true;

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(2));
      expect(messages[0].message, equals('last message'));
      expect(messages[1].message, equals('hi there'));

      final messages2 = await query.loadNext();
      expect(messages2, isNotNull);
      expect(messages2.length, equals(2));
      expect(messages2[0].message, equals('testing more'));
      expect(messages2[1].message, equals('this is test'));
    });

    test('should return messages with filtered sender ids', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )..senderIdsFilter = [user.userId, users.first.userId];

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(2));
    });

    test('should return messages with filtered message type', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )..messageTypeFilter = MessageTypeFilter.file;

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(0));
    });

    test('should return json from toJson', () {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )
        ..customTypesFilter = ['custom', 'custom2']
        ..includeParentMessageText = true
        ..senderIdsFilter = ['123', '456']
        ..messageTypeFilter = MessageTypeFilter.file;

      final res = query.toJson();

      expect(res['channel_url'], equals(channel.channelUrl));
      expect(res['channel_type'], equals('group'));
      expect(res['custom_types_filter'], equals(['custom', 'custom2']));
      expect(res['sender_ids_filter'], equals(['123', '456']));
      expect(res['message_type_filter'], equals('FILE'));
    });
  });

  group('previous message list for open channel', () {
    late OpenChannel channel;
    late UserMessage msg1, msg2;

    setUp(() async {
      final oparams = OpenChannelParams()..operatorUserIds = [user.userId];
      channel = await createOpenChannel(params: oparams, user: user);

      final params = UserMessageParams(message: 'test');
      msg1 = await sendUserMessageViaAPI(
        channel: channel,
        params: params..message = 'this is test',
        senderId: user.userId,
      );

      await sendUserMessageViaAPI(
        channel: channel,
        params: params..message = 'testing more',
        senderId: users[0].userId,
      );

      await sendUserMessageViaAPI(
        channel: channel,
        params: params..message = 'hi there',
        senderId: users[2].userId,
      );

      msg2 = await sendUserMessageViaAPI(
        channel: channel,
        params: params..message = 'last message',
        senderId: users[1].userId,
      );
    });

    test('should return all previous messages', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      );

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(4));
      expect(messages.first.channelType, equals(channel.channelType));
      expect(messages[0].messageId, equals(msg1.messageId));
      expect(messages[3].messageId, equals(msg2.messageId));
    });

    test('should return only limited message', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )..limit = 2;

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(2));

      final messages2 = await query.loadNext();
      expect(messages2, isNotNull);
      expect(messages2.length, equals(2));
    });

    test('should return messages with filtered sender ids', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )..senderIdsFilter = [user.userId, users.first.userId];

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(2));
    });

    test('should return messages with filtered message type', () async {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )..messageTypeFilter = MessageTypeFilter.file;

      final messages = await query.loadNext();
      expect(messages, isNotNull);
      expect(messages.length, equals(0));
    });

    test('should return json from toJson', () {
      final query = PreviousMessageListQuery(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      )
        ..customTypesFilter = ['custom', 'custom2']
        ..includeParentMessageText = true
        ..senderIdsFilter = ['123', '456']
        ..messageTypeFilter = MessageTypeFilter.file;

      final res = query.toJson();

      expect(res['channel_url'], equals(channel.channelUrl));
      expect(res['channel_type'], equals('open'));
      expect(res['custom_types_filter'], equals(['custom', 'custom2']));
      expect(res['sender_ids_filter'], equals(['123', '456']));
      expect(res['message_type_filter'], equals('FILE'));
    });
  });
}
