import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/utils/string_utils.dart';

import '../common/common.dart';
import '../common/sdksetup.dart';
import '../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late User user;
  late List<User> users;
  late List<GroupChannel> channels;

  setUpAll(() async {
    testInfo = await testCreateApp();
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDownAll(() async {
    await sdk.disconnect();
  });

  group('public group channel list query', () {
    setUpAll(() async {
      final u1 = await testApiCreateUser(userId: 'bob', nickname: 'bob');
      final u2 = await testApiCreateUser(userId: 'john', nickname: 'john');
      final u3 =
          await testApiCreateUser(userId: 'danielle', nickname: 'danielle');
      users = [u1, u2, u3];
      user = await testApiConnect(u1.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isPublic = true
        ..isDiscoverable = true
        ..name = 'channel 1'
        ..customType = 'custom 1'
        ..userIds = [u1.userId, u2.userId];
      final channel = await createGroupChannel(params);

      final params2 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isSuper = true
        ..name = 'channel 2'
        ..customType = 'custom 1'
        ..userIds = [u2.userId];
      final channel2 = await createGroupChannel(params2);

      final params3 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isPublic = true
        ..name = 'sendbird 1'
        ..customType = 'sendbird 1'
        ..userIds = users.map((e) => e.userId).toList();
      final channel3 = await createGroupChannel(params3);

      final params4 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isPublic = true
        ..name = 'sendbird 2'
        ..customType = 'sendbird 2'
        ..userIds = [u1.userId, u2.userId];
      final channel4 = await createGroupChannel(params4);

      final params5 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..name = 'channel sendbird'
        ..customType = 'custom 2'
        ..userIds = [u1.userId, u3.userId];
      final channel5 = await createGroupChannel(params5);

      final params6 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isPublic = true
        ..name = 'sendbird channel 1'
        ..customType = 'custom 3'
        ..userIds = users.map((e) => e.userId).toList();
      final channel6 = await createGroupChannel(params6);

      channels = [channel, channel2, channel3, channel4, channel5, channel6];
    });

    test('should return all existing public group channels', () async {
      final query = PublicGroupChannelListQuery();
      query.limit = 10;
      final result = await query.loadNext();
      expect(result.length, equals(4));
    });

    test('should return correct channels with custom type start filter',
        () async {
      final query = PublicGroupChannelListQuery();
      query.customTypeStartWithFilter = 'sendbird';
      final result = await query.loadNext();
      expect(result.length, equals(2));
    });

    test('should return correct channels with custom types filter', () async {
      final query = PublicGroupChannelListQuery();
      query.customTypesFilter = ['sendbird 1', 'custom 1'];
      final result = await query.loadNext();
      expect(result.length, equals(2));
    });

    test('should return along with metadata with includeMetadata option',
        () async {
      await channels[0].createMetaData({'key1': 'value1'});

      final query = PublicGroupChannelListQuery();
      query.customTypesFilter = ['custom 1'];
      query.includeMetaData = true; //default is true
      query.limit = 5;
      final result = await query.loadNext();
      expect(result.length, equals(1));
      expect(result.first.getCachedMetaData(), isNotNull);
      expect(result.first.getCachedMetaData().length, equals(1));
    });

    test('should return json from toJson', () {
      final query = PublicGroupChannelListQuery()
        ..channelUrls = ['123', '456']
        ..channelNameContainsFilter = 'name'
        ..customTypeStartWithFilter = 'custom'
        ..includeMemberList = true
        ..metaDataOrderKeyFilter = 'meta';

      final res = query.toJson();

      expect(res['channel_urls'], equals(['123', '456']));
      expect(res['channel_name_contains_filter'], equals('name'));
      expect(res['custom_type_start_with_filter'], equals('custom'));
      expect(res['include_member_list'], isTrue);
      expect(res['meta_data_order_key_filter'], equals('meta'));
    });
  });

  group('public group channel list query with meta data filter', () {
    final randomString = getRandomString(10);
    final randomKey = randomString + '1';
    final randomKey2 = randomString + '2';

    setUpAll(() async {
      final u1 = await testApiCreateUser(userId: 'bob1', nickname: 'bob');
      final u2 = await testApiCreateUser(userId: 'john1', nickname: 'john');
      final u3 =
          await testApiCreateUser(userId: 'danielle1', nickname: 'danielle');
      users = [u1, u2, u3];
      user = await testApiConnect(u1.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isPublic = true
        ..userIds = [u1.userId, u2.userId];
      final channel = await createGroupChannel(params);
      await channel.createMetaData({randomKey: 'v1'});

      final params2 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isPublic = true
        ..userIds = [u2.userId];
      final channel2 = await createGroupChannel(params2);
      await channel2.createMetaData({randomKey: 'v2'});

      final params3 = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..isPublic = true
        ..userIds = users.map((e) => e.userId).toList();
      final channel3 = await createGroupChannel(params3);
      await channel3.createMetaData({randomKey2: 'v3'});

      channels = [channel, channel2, channel3];
    });

    test('should return channel(s) with matching metaData key', () async {
      final query = PublicGroupChannelListQuery()
        ..limit = 5
        ..includeMetaData = true
        ..setMetaDataFilterWithValues(randomKey, ['v1']);

      var result = await query.loadNext();
      expect(result.length, equals(1));
      expect(query.hasNext, isFalse);
    });

    test('should return channel(s) with matching metaData values', () async {
      final query = PublicGroupChannelListQuery()
        ..limit = 5
        ..includeMetaData = true
        ..setMetaDataFilterWithValues(randomKey, ['v1', 'v2']);

      var result = await query.loadNext();
      expect(result.length, equals(2));
      expect(query.hasNext, isFalse);
    });

    test('should return channel(s) with matching metaData value startsWith',
        () async {
      final query = PublicGroupChannelListQuery()
        ..limit = 5
        ..includeMetaData = true
        ..setMetaDataFilterWithStartWith(randomKey, 'v');

      var result = await query.loadNext();
      expect(result.length, equals(2));
      expect(query.hasNext, isFalse);
    });
  });
}
