import 'package:flutter_test/flutter_test.dart';
import 'package:sendbird_sdk/core/models/restricted_user.dart';
import 'package:sendbird_sdk/request/user_meta_data/create_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../common/common.dart';
import '../common/helpers/private_api_client.dart';
import '../common/sdksetup.dart';
import '../common/user.dart';

void main() {
  late SendbirdSdk sdk;

  setUpAll(() async {});

  setUp(() async {
    final testInfo = await testCreateApp();
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('application user list query', () {
    test('should success with included userIds', () async {
      await testApiCreateUserAndConnect();
      final users = await testApiCreateUsers(3);
      final userIds = users.map((e) => e.userId).toList();
      final query = ApplicationUserListQuery()..userIds = userIds;
      final res = await query.loadNext();
      expect(res, isNotNull);
      expect(res.length, equals(3));
    });

    test('should return all users', () async {
      await testApiCreateUserAndConnect();
      await testApiCreateUsers(3);
      final query = ApplicationUserListQuery();
      final res = await query.loadNext();
      expect(res, isNotNull);
      expect(res.length, equals(4));
    });

    test('should return limited users with given limit', () async {
      final user = await testApiCreateUserAndConnect();
      var others = await testApiCreateUsers(5);
      others.add(user);
      final users = others.map((u) => u.userId).toList();

      final query = ApplicationUserListQuery();
      query.limit = 3;
      var result = <String>[];

      var res = await query.loadNext();
      expect(res, isNotNull);
      expect(res.length, equals(3));
      result.addAll(res.map((e) => e.userId).toList());

      res = await query.loadNext();
      expect(res.length, equals(3));
      result.addAll(res.map((e) => e.userId).toList());

      users.forEach((userId) => expect(result.contains(userId), isTrue));
    });

    test('should return users with metaKey and/or metaValues', () async {
      final user = await testApiCreateUserAndConnect();
      await user.createMetaData({'key1': 'value1'});

      final user1 = await testApiCreateUser(nickname: 'john');
      await PrivateApiClient().send(UserMetaDataCreateRequest(
        {'key1': 'value1'},
        userId: user1.userId,
      ));
      final user2 = await testApiCreateUser(nickname: 'brandon');
      await PrivateApiClient().send(UserMetaDataCreateRequest(
        {'key2': 'value1'},
        userId: user2.userId,
      ));

      final query1 = ApplicationUserListQuery()
        ..metaDataKey = 'key1'
        ..metaDataValues = ['value1'];
      var res = await query1.loadNext();
      expect(res.length, equals(2));
      expect(res.where((u) => u.nickname == user1.nickname).length, equals(1));
      expect(res.where((u) => u.nickname == user.nickname).length, equals(1));

      final query2 = ApplicationUserListQuery()
        ..metaDataKey = 'key2'
        ..metaDataValues = ['value1'];
      res = await query2.loadNext();
      expect(res.length, equals(1));
    });

    test('should return users with given nickname', () async {
      final user = await testApiCreateUser(nickname: 'tester 1');
      await testApiCreateUser(nickname: 'tester 2');
      await testApiCreateUser(nickname: 'john');
      await testApiCreateUser(nickname: 'brandon');
      await testApiCreateUser(nickname: 'jonny');

      await testApiConnect(user.userId);

      final query = ApplicationUserListQuery();
      query.nicknameStartsWith = 'tester';
      final res = await query.loadNext();
      expect(res.length, equals(2));
      expect(res.where((e) => e.nickname == 'tester 1').length, equals(1));
      expect(res.where((e) => e.nickname == 'tester 2').length, equals(1));

      final query2 = ApplicationUserListQuery();
      query2.nicknameStartsWith = 'b';
      final res2 = await query2.loadNext();
      expect(res2.length, equals(1));
      expect(res2.first.nickname, equals('brandon'));
    });

    // test('should not perform without connection', () async {});
  });

  group('filtered user list query', () {
    test('should return given users', () async {
      await testApiCreateUserAndConnect();
      final users = await testApiCreateUsers(3);
      final query = UserListQuery(queryType: UserListQueryType.filtered);
      query.userIds = users.getRange(0, 2).map((e) => e.userId).toList();
      //NOTE: query params list parsing is something wrong
      final res = await query.loadNext();
      expect(res, isNotNull);
      expect(res.length, equals(2));
    });
  });

  group('muted user list query', () {
    test('should return muted users a channel', () async {
      final users = await testApiCreateUsers(5);
      final connector = await testApiConnect(users.first.userId);
      final params = GroupChannelParams()
        ..userIds = users.map((e) => e.userId).toList()
        ..operatorUserIds = [connector.userId];

      final channel = await createGroupChannel(params);
      //shouldve waited all callback
      await channel.muteUser(
        userId: users[1].userId,
        seconds: 10000,
        description: 'muted!',
      );
      await channel.muteUser(
        userId: users[2].userId,
        seconds: 10000,
        description: 'muted!',
      );

      final query = UserListQuery(
        queryType: UserListQueryType.muted,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      );
      final mutedUsers = await query.loadNext();
      expect(mutedUsers, isNotNull);
      expect(mutedUsers.length, equals(2));
      final expected = users.getRange(1, 3).map((e) => e.userId).toList();
      final actual = mutedUsers.map((e) => e.userId).toList();
      expected.sort();
      actual.sort();
      expect(actual, equals(expected));
      mutedUsers.forEach((element) {
        expect(element is RestrictedUser, isTrue);
      });
    });
  });

  group('banned user list query', () {
    test('should return banned users from a channel', () async {
      final users = await testApiCreateUsers(5);
      final connector = await testApiConnect(users.first.userId);
      final params = GroupChannelParams()
        ..userIds = users.map((e) => e.userId).toList()
        ..operatorUserIds = [connector.userId];

      final channel = await createGroupChannel(params);
      //shouldve waited all callback
      await channel.banUser(
        userId: users[1].userId,
        seconds: 10000,
        description: 'banned!',
      );
      await channel.banUser(
        userId: users[2].userId,
        seconds: 10000,
        description: 'banned!',
      );

      final query = UserListQuery(
        queryType: UserListQueryType.banned,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      );
      final bannedUsers = await query.loadNext();
      expect(bannedUsers, isNotNull);
      expect(bannedUsers.length, equals(2));
      final expected = users.getRange(1, 3).map((e) => e.userId).toList();
      final actual = bannedUsers.map((e) => e.userId).toList();
      expected.sort();
      actual.sort();
      expect(actual, equals(expected));
      bannedUsers.forEach((element) {
        expect(element is RestrictedUser, isTrue);
      });
    });
  });

  group('participants list query', () {
    test('should return participants from an open channel', () async {
      final user = await testApiCreateUserAndConnect();
      final params = OpenChannelParams();
      final channel = await OpenChannel.createChannel(params);
      await channel.enter();

      final query = UserListQuery(
        queryType: UserListQueryType.participants,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
      );
      final participants = await query.loadNext();
      expect(participants, isNotNull);
      expect(participants.length, equals(1));
      expect(participants.first.userId, equals(user.userId));
    });
  });

  group('blocked list query', () {
    test('should return blocked users', () async {
      final users = await testApiCreateUsers(5);
      await testApiConnect(users.first.userId);

      await sdk.blockUser(users[1].userId);
      await sdk.blockUser(users[2].userId);

      final query = UserListQuery(queryType: UserListQueryType.blocked);
      final blocked = await query.loadNext();
      expect(blocked, isNotNull);
      expect(blocked.length, equals(2));

      await sdk.unblockUser(users[1].userId);

      final query2 = UserListQuery(queryType: UserListQueryType.blocked);
      final blocked2 = await query2.loadNext();
      expect(blocked2, isNotNull);
      expect(blocked2.length, equals(1));
    });
  });

  group('user list query', () {
    test('should return json from toJson', () {
      final query = PublicGroupChannelListQuery()
        ..channelUrls = ['123', '456']
        ..channelNameContainsFilter = 'name'
        ..customTypeStartWithFilter = 'custom'
        ..includeMemberList = true
        ..metaDataOrderKeyFilter = 'meta';

      final res = query.toJson();

      expect(res['channel_urls'], equals(['123', '456']));
      expect(res['channel_name_contains_filter'], equals('name'));
      expect(res['custom_type_start_with_filter'], equals('custom'));
      expect(res['include_member_list'], isTrue);
      expect(res['meta_data_order_key_filter'], equals('meta'));
    });
  });
}
