import 'package:sendbird_sdk/params/group_channel_total_unread_channel_count_params.dart';
import 'package:sendbird_sdk/request/channel/group_channel_create_request.dart';
import 'package:sendbird_sdk/request/channel/group_channel_update_request.dart';
import 'package:sendbird_sdk/request/channel_operation/group_channel_invite_request.dart';
import 'package:sendbird_sdk/request/messages/user_message_send_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../common/callback.dart';
import '../common/common.dart';
import '../common/helpers/private_api_client.dart';
import '../common/user.dart';
import '../common/sdksetup.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {});

  group('sdk mark as read', () {
    test('should return count 0 whne maskAsRead for specific channel',
        () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();

      final channel = await createGroupChannel(params);
      final event = ChannelEventListener();

      final msgParams = UserMessageParams(message: 'test message');
      PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
        params: msgParams,
      ));

      await event.waitFor([EventFuncName.onChannelChanged]);
      expect(channel.unreadMessageCount, equals(1));

      await sdk.markAsRead(channelUrls: [channel.channelUrl]);
      await event.waitFor([EventFuncName.onChannelChanged]);

      expect(channel.unreadMessageCount, equals(0));
    });

    test('should throw error for invalid params for markAsRead', () async {
      await testApiCreateUserAndConnect();
      expect(sdk.markAsRead(channelUrls: []),
          throwsA(isA<InvalidParameterError>()));
    });

    test('should return count 0 when markAsReadAll', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();

      var channel = await createGroupChannel(params);
      var channel2 = await createGroupChannel(params);

      final msgParams = UserMessageParams(message: 'test message');

      PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
        params: msgParams,
      ));

      PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        channelType: channel2.channelType,
        channelUrl: channel2.channelUrl,
        senderId: users.last.userId,
        params: msgParams,
      ));

      final callback = ChannelEventListener();
      await callback.waitFor([
        EventFuncName.onMessageReceived,
        EventFuncName.onMessageReceived,
      ]);
      expect(channel.unreadMessageCount, equals(1));
      expect(channel2.unreadMessageCount, equals(1));

      await sdk.markAsReadAll();
      await callback.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onChannelChanged,
      ]);

      expect(channel.unreadMessageCount, equals(0));
      expect(channel2.unreadMessageCount, equals(0));
    });
  });

  group('sdk group channel count with member filter', () {
    test('should return correct joined channel count', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);
      final userIds = users.map((e) => e.userId).toList();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = userIds;

      final req = GroupChannelCreateRequest(params);
      final channel = await PrivateApiClient().send<GroupChannel>(req);
      final req2 = GroupChannelCreateRequest(params);
      final channel2 = await PrivateApiClient().send<GroupChannel>(req2);
      final req3 = GroupChannelCreateRequest(params);
      final channel3 = await PrivateApiClient().send<GroupChannel>(req3);

      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: userIds,
        channelUrl: channel.channelUrl,
        inviterId: users.last.userId,
      ));
      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: userIds,
        channelUrl: channel2.channelUrl,
        inviterId: users.last.userId,
      ));
      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: userIds,
        channelUrl: channel3.channelUrl,
        inviterId: users.last.userId,
      ));

      final result = await sdk.getGroupChannelCount(MemberStateFilter.joined);

      expect(result, equals(3));
    });

    test('should return correct invited channel count', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);
      final userIds = users.map((e) => e.userId).toList();
      await sdk.setChannelInvitationPreference(false);
      final params = GroupChannelParams()..userIds = userIds.sublist(1);

      final req = GroupChannelCreateRequest(params);
      final channel = await PrivateApiClient().send<GroupChannel>(req);
      final req2 = GroupChannelCreateRequest(params);
      final channel2 = await PrivateApiClient().send<GroupChannel>(req2);
      final req3 = GroupChannelCreateRequest(params);
      final channel3 = await PrivateApiClient().send<GroupChannel>(req3);

      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: [user.userId],
        channelUrl: channel.channelUrl,
        inviterId: users.last.userId,
      ));
      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: [user.userId],
        channelUrl: channel2.channelUrl,
        inviterId: users.last.userId,
      ));
      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: [user.userId],
        channelUrl: channel3.channelUrl,
        inviterId: users.last.userId,
      ));

      final result = await sdk.getGroupChannelCount(MemberStateFilter.invited);
      final result2 = await sdk.getGroupChannelCount(MemberStateFilter.joined);
      expect(result, equals(3));
      expect(result2, equals(0));
    });

    test('should return correct all channel count', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);
      final userIds = users.map((e) => e.userId).toList();
      await sdk.setChannelInvitationPreference(false);
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = userIds;

      final req = GroupChannelCreateRequest(params);
      final channel = await PrivateApiClient().send<GroupChannel>(req);
      final req2 = GroupChannelCreateRequest(params);
      final channel2 = await PrivateApiClient().send<GroupChannel>(req2);
      final req3 = GroupChannelCreateRequest(params);
      final channel3 = await PrivateApiClient().send<GroupChannel>(req3);

      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: userIds,
        channelUrl: channel.channelUrl,
        inviterId: users.last.userId,
      ));
      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: userIds,
        channelUrl: channel2.channelUrl,
        inviterId: users.last.userId,
      ));
      await PrivateApiClient().send(GroupChannelInviteRequest(
        userIds: userIds,
        channelUrl: channel3.channelUrl,
        inviterId: users.last.userId,
      ));

      await channel2.acceptInvitation();

      final result = await sdk.getGroupChannelCount(MemberStateFilter.invited);
      final result2 = await sdk.getGroupChannelCount(MemberStateFilter.joined);
      final result3 = await sdk.getGroupChannelCount(MemberStateFilter.all);
      expect(result, equals(2));
      expect(result2, equals(1));
      expect(result3, equals(3));
    });
  });
  group('sdk total unread channel count', () {
    test('should return correct channel count', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);
      final userIds = users.map((e) => e.userId).toList();

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = userIds;

      final channel = await createGroupChannel(params);
      final channel2 = await createGroupChannel(params);
      await createGroupChannel(params);

      final msgParams = UserMessageParams(message: 'test');
      final event = ChannelEventListener();
      await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel2.channelType,
        channelUrl: channel2.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      expect(await sdk.getTotalUnreadChannelCount(), equals(2));
    });
  });

  test('should return correct channel count with including parameter',
      () async {
    final users = await testApiCreateUsers(3);
    final user = await testApiConnect(users.first.userId);
    final userIds = users.map((e) => e.userId).toList();

    final params = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = userIds
      ..customType = 'custom1.1'
      ..isSuper = true;

    final params2 = GroupChannelParams()
      ..operatorUserIds = [user.userId]
      ..userIds = userIds;

    final channel = await createGroupChannel(params);
    final channel2 = await createGroupChannel(params2);

    final msgParams = UserMessageParams(message: 'test');
    final event = ChannelEventListener();
    await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
      params: msgParams,
      channelType: channel.channelType,
      channelUrl: channel.channelUrl,
      senderId: users.last.userId,
    ));

    await event.waitFor([
      EventFuncName.onChannelChanged,
      EventFuncName.onMessageReceived,
    ]);

    await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
      params: msgParams,
      channelType: channel2.channelType,
      channelUrl: channel2.channelUrl,
      senderId: users.last.userId,
    ));

    await event.waitFor([
      EventFuncName.onChannelChanged,
      EventFuncName.onMessageReceived,
    ]);

    final countParams = GroupChannelTotalUnreadChannelCountParams()
      ..superChannelFilter = SuperChannelFilter.superChannel
      ..customTypes = ['custom1.1'];
    expect(
        await sdk.getTotalUnreadChannelCountWithParams(countParams), equals(1));
  });

  group('sdk total unread message count', () {
    test('should return correct unread message count', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);
      final userIds = users.map((e) => e.userId).toList();

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = userIds;

      final channel = await createGroupChannel(params);
      final msgParams = UserMessageParams(message: 'test');
      final event = ChannelEventListener();

      await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      expect(await sdk.getTotalUnreadMessageCount(), equals(2));
    });

    test('should return correct unread message count given custom types',
        () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);
      final userIds = users.map((e) => e.userId).toList();

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = userIds
        ..customType = 'custom1';

      final channel = await createGroupChannel(params);

      params.isSuper = true;
      params.customType = 'custom2';
      final channel2 = await createGroupChannel(params);

      final event = ChannelEventListener();
      final msgParams = UserMessageParams(message: 'test');
      await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
      ));
      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      await PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel2.channelType,
        channelUrl: channel2.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      final countParams = GroupChannelTotalUnreadMessageCountParams();
      countParams.customTypes = ['custom1'];
      expect(await sdk.getTotalUnreadMessageCount(), equals(3));
      expect(await sdk.getTotalUnreadMessageCountWithParams(countParams),
          equals(2));

      countParams.customTypes = ['custom1', 'custom2'];
      expect(
          await SendbirdSdk().getTotalUnreadMessageCountWithParams(countParams),
          equals(3));

      countParams.customTypes = ['custom2'];
      countParams.superChannelFilter = SuperChannelFilter.superChannel;
      expect(await sdk.getTotalUnreadMessageCountWithParams(countParams),
          equals(1));
    });
  });

  group('sdk unread item count', () {
    test('should return correct non super group message & invitation count',
        () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);
      final userIds = users.map((e) => e.userId).toList();

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = userIds;

      final channel = await createGroupChannel(params);
      sdk.setChannelInvitationPreference(false);

      final channel2 = await createGroupChannel(params);
      params.isSuper = true;

      final superChannel = await createGroupChannel(params);
      final event = ChannelEventListener();
      final msgParams = UserMessageParams(message: 'test');
      PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel.channelType,
        channelUrl: channel.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      msgParams
        ..mentionType = MentionType.users
        ..mentionedUserIds = [user.userId];

      PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: superChannel.channelType,
        channelUrl: superChannel.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      PrivateApiClient().send(ChannelUserMessageSendApiRequest(
        params: msgParams,
        channelType: channel2.channelType,
        channelUrl: channel2.channelUrl,
        senderId: users.last.userId,
      ));

      await event.waitFor([
        EventFuncName.onChannelChanged,
        EventFuncName.onMessageReceived,
      ]);

      final result = await sdk.getUnreadItemCount([
        // UnreadItemKey.nonSuperGroupChannelInvitationCount,
        UnreadItemKey.nonSuperGroupChannelUnreadMessageCount,
      ]);

      expect(result.nonSuperGroupChannelInvitationCount, isNull);
      expect(result.nonSuperGroupChannelUnreadMentionCount, isNull);
      expect(result.nonSuperGroupChannelUnreadMessageCount, equals(2));
    });
  });

  group('sdk group channel change logs', () {
    test('should return correct logs', () async {
      final user = await testApiCreateUserAndConnect();

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];

      final event = ChannelEventListener();
      final channel2 = await createGroupChannel(params);
      final channel3 = await createGroupChannel(params);

      final updatedParams = GroupChannelParams.withChannel(channel2)
        ..name = 'change name'
        ..customType = 'custom';
      PrivateApiClient().send(GroupChannelUpdateRequest(updatedParams));
      await event.waitFor([EventFuncName.onChannelChanged]);

      channel3.deleteChannel();
      await event.waitFor([EventFuncName.onChannelDeleted]);

      final logParams = GroupChannelChangeLogsParams();
      logParams.includeEmptyChannel = true;

      final result = await sdk.getMyGroupChannelChangeLogs(params: logParams);
      expect(result.updatedChannels.length, equals(1));
      expect(result.deletedChannelUrls.length, equals(1));
      expect(result.hasMore, isFalse);
    });

    // flacky
    // test('should return correct change logs given timestamp', () async {
    //   final users = await testApiCreateUsers(3);
    //   final user = await testApiConnect(users.first.userId);
    //   final userIds = users.map((e) => e.userId).toList();

    //   final params = GroupChannelParams()
    //     ..operatorUserIds = [user.userId]
    //     ..userIds = userIds;

    //   final channel2 = await createGroupChannel(params);
    //   final channel3 = await createGroupChannel(params);
    //   await channel2.sendUserMessage(
    //     'text',
    //     onCompleted: (message, error) {},
    //   );
    //   final event = ChannelEventTester();
    //   final updatedParams = GroupChannelParams.withChannel(channel2)
    //     ..name = 'change name'
    //     ..customType = 'custom';
    //   PrivateApiClient().updateGroupChannel(updatedParams);
    //   await event.waitFor([EventFuncName.onChannelChanged]);

    //   final ts = DateTime.now().millisecondsSinceEpoch;
    //   channel3.deleteChannel();
    //   await event.waitFor([EventFuncName.onChannelDeleted]);

    //   final logParams = GroupChannelChangeLogsParams();
    //   final result = await sdk.getMyGroupChannelChangeLogsWithTimestamp(
    //     params: logParams,
    //     timestamp: ts,
    //   );
    //   expect(result.updatedChannels.length, equals(0));
    //   expect(result.deletedChannelUrls.length, equals(1));
    //   expect(result.hasMore, isFalse);
    // });
  });
}
