import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../common/user.dart';
import '../common/sdksetup.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('sdk do not distrub', () {
    test('should success to set with given valid params', () async {
      await testApiCreateUserAndConnect();
      await sdk.setDoNotDisturb(
        enable: true,
        startHour: 12,
        startMin: 12,
        endHour: 12,
        endMin: 40,
        timezone: 'Asia/Seoul',
      );
    });

    test('should return exist setting', () async {
      await testApiCreateUserAndConnect();
      await sdk.setDoNotDisturb(
        enable: true,
        startHour: 12,
        startMin: 12,
        endHour: 12,
        endMin: 40,
        timezone: 'Asia/Seoul',
      );

      final result = await sdk.getDoNotDisturb();
      expect(result.enabled, isTrue);
      expect(result.startHour, equals(12));
      expect(result.startMin, equals(12));
      expect(result.endHour, equals(12));
      expect(result.endMin, equals(40));
    });

    test('should throw an exception with given invalid timezone', () async {
      await testApiCreateUserAndConnect();
      expect(
        sdk.setDoNotDisturb(
          enable: true,
          startHour: 12,
          startMin: 12,
          endHour: 12,
          endMin: 40,
          timezone: 'KST',
        ),
        throwsA(isA<BadRequestError>()),
      );
    });

    test('should throw an exception with given invalid time period', () async {
      await testApiCreateUserAndConnect();
      expect(
        sdk.setDoNotDisturb(
          enable: true,
          startHour: 12,
          startMin: 91,
          endHour: 12,
          endMin: 40,
          timezone: 'Asia/Seoul',
        ),
        throwsA(isA<InvalidParameterError>()),
      );
    });
  });

  group('sdk snooze', () {
    test('should success to set with given valid params', () async {
      await testApiCreateUserAndConnect();
      await sdk.setSnoozePeriod(
        enable: true,
        startDate: DateTime.now(),
        endDate: DateTime.now().add(Duration(days: 100)),
      );
    });

    test('should return exist setting', () async {
      final startDate = DateTime.now();
      final endDate = DateTime.now().add(Duration(days: 100));
      await testApiCreateUserAndConnect();
      await sdk.setSnoozePeriod(
        enable: true,
        startDate: startDate,
        endDate: endDate,
      );

      final result = await sdk.getSnoozePeriod();
      expect(result.enabled, isTrue);
      expect(result.startDate?.second, equals(startDate.second));
      expect(result.endDate?.second, equals(endDate.second));
    });

    test('should throw an exception with invalid params', () async {
      final startDate = DateTime.now();
      final endDate = DateTime.now().add(Duration(days: 100));
      await testApiCreateUserAndConnect();
      expect(
        sdk.setSnoozePeriod(
          enable: true,
          startDate: null,
          endDate: endDate,
        ),
        throwsA(isA<InvalidParameterError>()),
      );

      expect(
        sdk.setSnoozePeriod(
          enable: true,
          startDate: startDate,
          endDate: null,
        ),
        throwsA(isA<InvalidParameterError>()),
      );

      expect(
        sdk.setSnoozePeriod(
          enable: true,
          startDate: endDate,
          endDate: startDate,
        ),
        throwsA(isA<InvalidParameterError>()),
      );
    });
  });

  group('sdk push trigger option', () {
    test('should success with given valid params', () async {
      await testApiCreateUserAndConnect();
      await sdk.setPushTriggerOption(PushTriggerOption.all);
    });

    test('should return exist option', () async {
      await testApiCreateUserAndConnect();
      await sdk.setPushTriggerOption(PushTriggerOption.mentionOnly);
      expect(await sdk.getPushTriggerOption(),
          equals(PushTriggerOption.mentionOnly));
    });
  });

  group('sdk invitation preference', () {
    test('should success with given valid param', () async {
      await testApiCreateUserAndConnect();
      await sdk.setChannelInvitationPreference(false);
    });

    test('should return exist preference', () async {
      await testApiCreateUserAndConnect();
      await sdk.setChannelInvitationPreference(false);
      expect(await sdk.getChannelInvitationPreference(), isFalse);
    });
  });

  group('sdk push sound', () {
    test('should success with given valid param', () async {
      await testApiCreateUserAndConnect();
      await sdk.setPushSound('Sound');
    });

    test('should return exist push sound', () async {
      await testApiCreateUserAndConnect();
      await sdk.setPushSound('Sound');
      expect(await sdk.getPushSound(), equals('Sound'));
    });

    test('should throw an exception with invalid params', () async {
      await testApiCreateUserAndConnect();
      expect(sdk.setPushSound(''), throwsA(isA<InvalidParameterError>()));
    });
  });

  group('sdk push template', () {
    test('should success with given valid param', () async {
      await testApiCreateUserAndConnect();
      await sdk.setPushTemplate('default');
    });

    test('should return exist preference', () async {
      await testApiCreateUserAndConnect();
      await sdk.setPushTemplate('alternative');
      expect(await sdk.getPushTemplate(), equals('alternative'));
    });

    test('should throw an exception with invalid params', () async {
      await testApiCreateUserAndConnect();
      expect(sdk.setPushTemplate('random'), throwsA(isA<BadRequestError>()));
    });
  });
}
