import 'dart:async';
import 'package:sendbird_sdk/managers/connection_manager.dart';
import 'package:test/test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import '../common/callback.dart';
import '../common/common.dart';
import '../common/prep.dart';
import '../common/user.dart';
import '../common/sdksetup.dart';

class Test with ChannelEventHandler, ConnectionEventHandler, UserEventHandler {}

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    sdk.setOptions(Options(
      fileTransferTimeout: 300,
      typingIndicatorThrottle: 1,
      websocketTimeout: 100,
    ));
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('sdk initialization', () {
    test('should contain correct option values', () {
      final option = sdk.getOptions();
      expect(option.fileTransferTimeout, equals(300));
      expect(option.websocketTimeout, equals(100));
      expect(option.typingIndicatorThrottle, equals(1));
    });
  });

  group('sdk add handlers', () {
    test('should handle multiple listeners properly', () {
      final listener = Test();

      final identifier = 'handler';
      sdk.addChannelEventHandler(identifier, listener);
      sdk.addConnectionEventHandler(identifier, listener);
      sdk.addUserEventHandler(identifier, listener);

      expect(sdk.getChannelEventHandler(identifier), isNotNull);
      expect(sdk.getConnectionEventHandler(identifier), isNotNull);
      expect(sdk.getUserEventHandler(identifier), isNotNull);
    });
  });

  group('sdk connect', () {
    test('should succeed with created user', () async {
      final usera = await testApiCreateUser();
      expect(usera, isNotNull);
      var userb = await sdk.connect(
        usera.userId,
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );
      expect(userb, isNotNull);
      expect(usera.userId, equals(userb.userId));
    });

    test('should succeed connect back and forth', () async {
      final users = await testApiCreateUsers(2);
      expect(users, isNotNull);
      expect(users.length, equals(2));
      var usera = await sdk.connect(
        users[0].userId,
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );
      expect(usera, isNotNull);
      expect(usera.userId, equals(users[0].userId));
      var userb = await sdk.connect(
        users[1].userId,
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );
      expect(userb, isNotNull);
      expect(userb.userId, equals(users[1].userId));
    });

    test('should return same user when get current user', () async {
      final usera = await testApiCreateUser();
      expect(usera, isNotNull);
      var userb = await sdk.connect(
        usera.userId,
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );
      expect(userb, isNotNull);
      expect(usera.userId, equals(userb.userId));
    });

    test('should throw an exception with invalid wshost', () async {
      final user = await testApiCreateUser();
      late Error error;
      try {
        await sdk.connect(user.userId, wsHost: 'invalid_url');
      } catch (e) {
        error = e as Error;
      }

      expect(error, isNotNull);
      expect(error, isA<WebSocketError>());
    });

    test('should throw an exception timeout', () async {
      final user = await testApiCreateUser();
      sdk.setOptions(Options()..connectionTimeout = 0);

      try {
        await sdk.connect(
          user.userId,
          apiHost: env.getApiUrl(),
          wsHost: env.getWsUrl(),
        );
      } catch (e) {
        expect(e, isNotNull);
        expect(e, isA<LoginTimeoutError>());
      }
    });
  });

  group('sdk connect with session', () {
    test('should succeed with pre session token', () async {
      final expiresAt =
          DateTime.now().add(Duration(seconds: 10)).millisecondsSinceEpoch;

      final user = await testApiCreateUser(
        issueSessionToken: true,
        sessionTokenExpiresAt: expiresAt,
      );

      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: user.sessionToken,
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(user.userId, equals(connectedUser.userId));
    });

    // test('should refresh session when requesting with expired session',
    //     () async {
    //   await configureAppForSessionExpiration(
    //     appId: testInfo.appId,
    //     sessionDuration: 20,
    //     expiring: true,
    //   );

    //   final expiresAt =
    //       DateTime.now().add(Duration(seconds: 60)).millisecondsSinceEpoch;

    //   final user = await testApiCreateUser(
    //     issueSessionToken: true,
    //     sessionTokenExpiresAt: expiresAt,
    //   );

    //   final event = SessionEventListener();
    //   final connection = ConnectionEventListener();

    //   await sdk.connect(
    //     user.userId,
    //     accessToken: user.sessionToken,
    //     apiHost: env.getApiUrl(),
    //     wsHost: env.getWsUrl(),
    //   );

    //   await Future.delayed(Duration(seconds: 15));

    //   final params = GroupChannelParams();
    //   final channel = await GroupChannel.createChannel(params);

    //   await event.waitFor([
    //     EventFuncName.onSessionExpired,
    //     EventFuncName.onSessionRefreshed,
    //   ], timeout: 20);
    //   await connection.waitFor([EventFuncName.onReconnectionSucceeded]);
    //   await configureAppForSessionExpiration(
    //       appId: testInfo.appId, expiring: false);
    // });

    test('should refresh expired session and suceeded to create channel',
        () async {
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 25,
        expiring: true,
      );

      final expiresAt =
          DateTime.now().add(Duration(seconds: 300)).millisecondsSinceEpoch;

      final user = await testApiCreateUser(
        issueSessionToken: true,
        sessionTokenExpiresAt: expiresAt,
      );

      final event = SessionEventListener();
      final connection = ConnectionEventListener();

      await sdk.connect(
        user.userId,
        accessToken: user.sessionToken,
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      await event.waitFor([EventFuncName.onSessionRefreshed], timeout: 30);
      await connection.waitFor([EventFuncName.onReconnectionSucceeded]);

      try {
        final params = GroupChannelParams()
          ..userIds = [user.userId]
          ..operatorUserIds = [user.userId];
        final channel = await createGroupChannel(params);
        expect(channel, isNotNull);
      } catch (e) {
        expect(e, isNull);
        //should not throw
      }

      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout(Duration(seconds: 180)));

    test('should throw an exception with expired session token', () async {
      final expiresAt =
          DateTime.now().add(Duration(seconds: 2)).millisecondsSinceEpoch;

      final user = await testApiCreateUser(
        issueSessionToken: true,
        sessionTokenExpiresAt: expiresAt,
      );

      await Future.delayed(Duration(seconds: 2));

      expect(
          () => sdk.connect(
                user.userId,
                accessToken: user.sessionToken,
                apiHost: env.getApiUrl(),
                wsHost: env.getWsUrl(),
              ),
          throwsA(isA<SBError>()));
    });

    // TODO: need to figure out how to test multiple awaits
    test('should throw an exception while updating session with expired token',
        () async {
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 25,
        expiring: true,
      );

      final expiresAt =
          DateTime.now().add(Duration(seconds: 5)).millisecondsSinceEpoch;

      final user = await testApiCreateUser(
        issueSessionToken: true,
        sessionTokenExpiresAt: expiresAt,
      );

      final event = SessionEventListener();

      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: user.sessionToken,
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(user.userId, equals(connectedUser.userId));

      // EXPR already arrives here

      final result = await event
          .waitFor([EventFuncName.onSessionTokenRequired], timeout: 35);
      // session require token called
      try {
        result.first.data.first('invalid_token');
      } catch (e) {
        print('Refresh error $e');
      }

      await event.waitFor([EventFuncName.onSessionError]);
      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout(Duration(seconds: 180)));
  });

  group('sdk reconnect', () {
    test('should return true and reconnect to the server', () async {
      await testApiCreateUserAndConnect();
      expect(sdk.reconnect(), isTrue);
      //delay to finish connect
      //should reconnect return future?
      await Future.delayed(Duration(seconds: 2));
    });

    test('should return false if no previous connection was made', () async {
      expect(sdk.reconnect(), isFalse);
    });

    test(
        'should queue all interface request while reconnecting and execute once the connectio is made',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);

      // sdk.reconnect();
      // demonstrate reconnecting situation by manually flaggin reconnecting
      sdk.getInternal().state.reconnecting = true;
      final wait = Completer(), wait2 = Completer();
      Error? err1, err2;
      UserMessage? msg1, msg2;

      channel.sendUserMessageWithText('test', onCompleted: (message, error) {
        msg1 = message;
        err1 = error;
        wait.complete();
      });

      channel.sendUserMessageWithText('test', onCompleted: (message, error) {
        msg2 = message;
        err2 = error;
        wait2.complete();
      });

      //arbitary waiting
      await Future.delayed(Duration(seconds: 3));

      //simulate logi complete
      sdk.getInternal().state.reconnecting = false;
      ConnectionManager.flushCompleters();

      await wait.future;
      await wait2.future;

      expect(msg1, isNotNull);
      expect(msg1?.sendingStatus, MessageSendingStatus.succeeded);
      expect(err1, isNull);
      expect(msg2, isNotNull);
      expect(msg2?.sendingStatus, MessageSendingStatus.succeeded);
      expect(err2, isNull);
    });

    test('should return all queued requests with error when error happens',
        () async {
      final user = await testApiCreateUserAndConnect();
      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = [user.userId];
      final channel = await createGroupChannel(params);

      // sdk.reconnect();
      // demonstrate reconnecting situation by manually flaggin reconnecting
      sdk.getInternal().state.reconnecting = true;

      final wait = Completer(), wait2 = Completer();
      Error? err1, err2;
      UserMessage? msg1, msg2;

      channel.sendUserMessageWithText('test', onCompleted: (message, error) {
        msg1 = message;
        err1 = error;
        wait.complete();
      });

      channel.sendUserMessageWithText('test', onCompleted: (message, error) {
        msg2 = message;
        err2 = error;
        wait2.complete();
      });

      //arbitary waiting
      await Future.delayed(Duration(seconds: 3));
      //simulate error case manually
      sdk.getInternal().state.reconnecting = false;
      ConnectionManager.flushCompleters(error: ConnectionClosedError());

      await wait.future;
      await wait2.future;

      expect(msg1, isNotNull);
      expect(msg1?.sendingStatus, MessageSendingStatus.failed);
      expect(err1, isA<ConnectionClosedError>());
      expect(msg2, isNotNull);
      expect(msg2?.sendingStatus, MessageSendingStatus.failed);
      expect(err2, isA<ConnectionClosedError>());
    });
  });

  group('sdk disconnect', () {
    test('should clean all existing information', () async {
      await testApiCreateUserAndConnect();
      await sdk.disconnect();

      final state = sdk.getInternal().state;
      expect(state.apiHost, isNull);
      expect(state.wsHost, isNull);
      expect(state.connected, isFalse);
      expect(state.connecting, isFalse);
      expect(state.currentUser, isNull);
    });
  });

  group('sdk get current user', () {
    test('should return null before connect', () async {
      var user = sdk.currentUser;
      expect(user, isNull);
    });

    test('should return current connected user', () async {
      final user = await testApiCreateUserAndConnect();
      final currUser = sdk.currentUser;

      expect(user.userId, equals(currUser?.userId));
      expect(user.nickname, equals(currUser?.nickname));
      expect(user.lastSeenAt, equals(currUser?.lastSeenAt));
    });

    test('should return null after disconnect', () async {
      await testApiCreateUserAndConnect();
      expect(sdk.currentUser, isNotNull);
      await sdk.disconnect();
      expect(sdk.currentUser, isNull);
    });
  });

  group('sdk get last connected at', () {
    test('should return valid value after a user connects', () async {
      await testApiCreateUserAndConnect();
      expect(sdk.getLastConnectedAt(), isNotNull);
      expect(sdk.getLastConnectedAt(), greaterThan(0));
    });

    test('should return null if no one connects yet', () async {
      expect(sdk.getLastConnectedAt(), isNull);
    });
  });

  group('sdk callback handlers', () {
    test('should succeed to add channel handler', () {
      final handler = ChannelEventHandler();
      sdk.addChannelEventHandler('handler', handler);
      expect(sdk.getChannelEventHandler('handler'), equals(handler));
    });

    test('should succeed to remove channel handler', () {
      final handler = ChannelEventHandler();
      sdk.addChannelEventHandler('handler', handler);
      expect(sdk.getChannelEventHandler('handler'), equals(handler));
      sdk.removeChannelEventHandler('handler');
      expect(sdk.getChannelEventHandler('handler'), isNull);
    });

    test('should succeed to add connection handler', () {
      final handler = ConnectionEventHandler();
      sdk.addConnectionEventHandler('handler', handler);
      expect(sdk.getConnectionEventHandler('handler'), equals(handler));
    });

    test('should succeed to remove connection handler', () {
      final handler = ConnectionEventHandler();
      sdk.addConnectionEventHandler('handler', handler);
      expect(sdk.getConnectionEventHandler('handler'), equals(handler));
      sdk.removeConnectionEventHandler('handler');
      expect(sdk.getConnectionEventHandler('handler'), isNull);
    });

    test('should succeed to add session handler', () {
      final handler = SessionEventHandler();
      sdk.addSessionEventHandler(handler);
      expect(sdk.getSessionEventHandler(), equals(handler));
    });

    test('should succeed to remove session handler', () {
      final handler = SessionEventHandler();
      sdk.addSessionEventHandler(handler);
      expect(sdk.getSessionEventHandler(), equals(handler));
      sdk.removeSessionEventHandler();
      expect(sdk.getSessionEventHandler(), isNull);
    });

    test('should succeed to add user event handler', () {
      final handler = UserEventHandler();
      sdk.addUserEventHandler('handler', handler);
      expect(sdk.getUserEventHandler('handler'), equals(handler));
    });

    test('should succeed to remove user event handler', () {
      final handler = UserEventHandler();
      sdk.addUserEventHandler('handler', handler);
      expect(sdk.getUserEventHandler('handler'), equals(handler));
      sdk.removeUserEventHandler('handler');
      expect(sdk.getUserEventHandler('handler'), isNull);
    });
  });

  group('sdk get app info', () {
    test('should return valid application info after connected', () async {
      await testApiCreateUserAndConnect();
      expect(sdk.getAppInfo(), isNotNull);
    });

    test('should return null if no one connects yet', () async {
      expect(sdk.getAppInfo(), isNull);
      await testApiCreateUserAndConnect();
      expect(sdk.getAppInfo(), isNotNull);
      await sdk.disconnect();
      expect(sdk.getAppInfo(), isNull);
    });
  });

  group('sdk get sdk version', () {
    test('should return proper sdk version', () async {
      expect(sdk.getSdkVersion(), isNotNull);
      expect(sdk.getSdkVersion(), isNot(equals('')));
    });
  });

  group('sdk get application id', () {
    test('should return current initalized application id', () {
      expect(sdk.getApplicationId(), isNotNull);
      expect(sdk.getApplicationId(), equals(testInfo.appId));
    });
  });

  group('sdk get connection state', () {
    test('should return close on intial state', () async {
      expect(sdk.getConnectionState(), ConnectionState.closed);
    });

    test('should return open after connect was invoked', () async {
      await testApiCreateUserAndConnect();
      expect(sdk.getConnectionState(), ConnectionState.open);
    });
  });

  group('sdk push test', () {
    setUp(() async {
      await testApiCreateUserAndConnect();
    });

    test('should sc registering push token', () async {
      final result = await sdk.registerPushToken(
        type: PushTokenType.apns,
        token: 'ace8efc363725b65201c07220199a57df9f2cb2f',
      );
      expect(result, PushTokenRegistrationStatus.success);
    });

    test('should succeed unregistering push token', () async {
      final result = await sdk.registerPushToken(
        type: PushTokenType.apns,
        token: 'ace8efc363725b65201c07220199a57df9f2cb2f',
      );
      expect(result, PushTokenRegistrationStatus.success);

      await sdk.unregisterPushToken(
        type: PushTokenType.apns,
        token: 'ace8efc363725b65201c07220199a57df9f2cb2f',
      );
    });

    test('should succeed unregistering all push token', () async {
      final result = await sdk.registerPushToken(
        type: PushTokenType.apns,
        token: 'ace8efc363725b65201c07220199a57df9f2cb2f',
      );
      expect(result, PushTokenRegistrationStatus.success);

      await sdk.unregisterAllPushToken();
    });

    test('should succeed registering hms push token', () async {
      final result = await sdk.registerPushToken(
        type: PushTokenType.hms,
        token: 'ace8efc363725b65201c07220199a57df9f2cb2f',
      );
      expect(result, PushTokenRegistrationStatus.success);
    });

    test('should succeed unregistering hms push token', () async {
      final result = await sdk.registerPushToken(
        type: PushTokenType.apns,
        token: 'ace8efc363725b65201c07220199a57df9f2cb2f',
      );
      expect(result, PushTokenRegistrationStatus.success);

      await sdk.unregisterPushToken(
        type: PushTokenType.apns,
        token: 'ace8efc363725b65201c07220199a57df9f2cb2f',
      );
    });
  });

  group('sdk push trigger option', () {
    setUp(() async {
      await testApiCreateUserAndConnect();
    });

    test('should properly set option', () async {
      await sdk.setPushTriggerOption(PushTriggerOption.all);
      var result = await sdk.getPushTriggerOption();
      expect(result, equals(PushTriggerOption.all));

      await sdk.setPushTriggerOption(PushTriggerOption.mentionOnly);
      result = await sdk.getPushTriggerOption();
      expect(result, equals(PushTriggerOption.mentionOnly));

      await sdk.setPushTriggerOption(PushTriggerOption.off);
      result = await sdk.getPushTriggerOption();
      expect(result, equals(PushTriggerOption.off));
    });
  });

  // group('sdk mark as delivery with payload', () {
  //   test('should succeed with valid payload', () async {});
  //   test('should fail with invalid payalod', () async {});
  // });

  group('sdk subscribed count', () {
    late List<User> users;
    late User user;
    late GroupChannel channel;

    setUp(() async {
      users = await testApiCreateUsers(3);
      user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      channel = await createGroupChannel(params);
    });

    test('should return total unread message count', () async {
      final event = ChannelEventListener();

      final mParams = UserMessageParams(message: 'text');
      sendUserMessageViaAPI(
        channel: channel,
        params: mParams,
        senderId: users.last.userId,
      );

      await event.waitFor([EventFuncName.onChannelChanged]);
      expect(sdk.subscribedTotalUnreadMessageCount, equals(1));
    });

    test('should return custom type total unread message count', () async {
      final channelParams = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..customType = 'CUSTOM_TYPE_A'
        ..userIds = users.map((e) => e.userId).toList();
      channel = await createGroupChannel(channelParams);
      final event = ChannelEventListener();

      final params = UserMessageParams(message: 'text');
      sendUserMessageViaAPI(
        channel: channel,
        params: params,
        senderId: users.last.userId,
      );

      await event.waitFor([EventFuncName.onChannelChanged]);

      channelParams.customType = 'CUSTOM_TYPE_B';
      final channel2 = await createGroupChannel(channelParams);
      sendUserMessageViaAPI(
        channel: channel2,
        params: params,
        senderId: users.last.userId,
      );

      await event.waitFor([EventFuncName.onChannelChanged]);

      //not register custom type
      channelParams.customType = 'CUSTOM_TYPE_T';
      final channel3 = await createGroupChannel(channelParams);
      sendUserMessageViaAPI(
        channel: channel3,
        params: params,
        senderId: users.last.userId,
      );

      await event.waitFor([EventFuncName.onChannelChanged]);

      expect(sdk.subscribedCustomTypeTotalUnreadMessageCount, equals(2));
    });

    test('should return custom type unread message count with custom type',
        () async {
      final channelParams = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..customType = 'CUSTOM_TYPE_A'
        ..userIds = users.map((e) => e.userId).toList();
      channel = await createGroupChannel(channelParams);
      final event = ChannelEventListener();

      final params = UserMessageParams(message: 'text');
      sendUserMessageViaAPI(
        channel: channel,
        params: params,
        senderId: users.last.userId,
      );

      await event.waitFor([EventFuncName.onChannelChanged]);

      channelParams.customType = 'CUSTOM_TYPE_B';
      final channel2 = await createGroupChannel(channelParams);
      sendUserMessageViaAPI(
        channel: channel2,
        params: params,
        senderId: users.last.userId,
      );

      await event.waitFor([EventFuncName.onChannelChanged]);

      //not register custom type
      channelParams.customType = 'CUSTOM_TYPE_T';
      final channel3 = await createGroupChannel(channelParams);
      sendUserMessageViaAPI(
        channel: channel3,
        params: params,
        senderId: users.last.userId,
      );

      await event.waitFor([EventFuncName.onChannelChanged]);

      expect(
        sdk.subscribedCustomTypeUnreadMessageCount('CUSTOM_TYPE_A'),
        equals(1),
      );
      expect(
        sdk.subscribedCustomTypeUnreadMessageCount('CUSTOM_TYPE_B'),
        equals(1),
      );
      expect(
        sdk.subscribedCustomTypeUnreadMessageCount('CUSTOM_TYPE_T'),
        isNull,
      );
    });
  });
}
