import 'package:flutter/scheduler.dart';
import 'package:sendbird_sdk/request/user/session/session_token_issue_request.dart';
import 'package:sendbird_sdk/request/user/session/session_token_revoke_request.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';
import '../common/callback.dart';
import '../common/common.dart';
import '../common/helpers/private_api_client.dart';
import '../common/prep.dart';
import '../common/sdksetup.dart';
import '../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late SessionEventHandler sessionHandler;

  setUpAll(() async {
    sessionHandler = SessionEventHandler();
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
    sdk.addSessionEventHandler(sessionHandler);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  //* Authentication Connecting
  group('session authentication while connecting', () {
    test('should success when session token is valid', () async {
      //valid session token
      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient()
          .send(UserSessionTokenIssueRequest(userId: user.userId));

      //* Connect SB server with user
      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(connectedUser, isA<User>());
      expect(sdk.getInternal().state.connected, isTrue);
    });

    test('should throw exception when session token is invalid', () async {
      final event = SessionEventListener();
      //create user and issueSessionToken
      final user = await testApiCreateUser(issueSessionToken: true);

      //invalid session token
      const sessionToken = '';

      sdk
          .connect(
            user.userId,
            accessToken: sessionToken,
            apiHost: env.getApiUrl(),
            wsHost: env.getWsUrl(),
          )
          .then(
            (value) => {
              print('user recieved'),
            },
          )
          .catchError((e) => {isA<LoginTimeoutError>()});

      await event.waitFor([EventFuncName.onSessionError]);
      await Future.delayed(
          Duration(seconds: sdk.getOptions().connectionTimeout));
    }, timeout: Timeout.none);

    test('should throw exception when session token is expired', () async {
      final event = SessionEventListener();
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 200,
        expiring: true,
      );
      final sessionTokenExpiresAt =
          DateTime.now().add(Duration(seconds: 25)).millisecondsSinceEpoch;

      //create user
      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient().send(UserSessionTokenIssueRequest(
          userId: user.userId, expiresAt: sessionTokenExpiresAt));

      await Future.delayed(Duration(seconds: 30));

      //* Connect SB server with user
      sdk
          .connect(
            user.userId,
            accessToken: result['token'],
            apiHost: env.getApiUrl(),
            wsHost: env.getWsUrl(),
          )
          .then(
            (value) => {
              print('user recieved'),
            },
          )
          .catchError((e) => {isA<LoginTimeoutError>()});
      ;
      await event.waitFor([EventFuncName.onSessionError]);

      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout.none);

    test('should throw exception when session token is revoked', () async {
      final event = SessionEventListener();
      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient()
          .send(UserSessionTokenIssueRequest(userId: user.userId));

      //revoke session token
      await PrivateApiClient()
          .send(UserSessionTokenRevokeRequest(userId: user.userId));

      //* Connect SB server with user
      sdk
          .connect(
            user.userId,
            accessToken: result['token'],
            apiHost: env.getApiUrl(),
            wsHost: env.getWsUrl(),
          )
          .then(
            (value) => {
              print('user recieved'),
            },
          )
          .catchError((e) => {isA<LoginTimeoutError>()});

      await event.waitFor([EventFuncName.onSessionClosed]);
    });
  }, timeout: Timeout.none);

  //* Authentication Reconnecting
  group('session authentication while reconnecting', () {
    test('should success when session token and session key are valid',
        () async {
      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient()
          .send(UserSessionTokenIssueRequest(userId: user.userId));

      // retrieve session key
      //* Connect SB server with user
      await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      // background
      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.paused);
      await Future.delayed(Duration(seconds: 10));
      // foreground
      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.resumed);
      await Future.delayed(Duration(seconds: 10));
    }, timeout: Timeout.none);

    test(
        'should success when session token is valid and session key is expired',
        () async {
      final event = SessionEventListener();
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 30,
        expiring: true,
      );
      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient()
          .send(UserSessionTokenIssueRequest(userId: user.userId));

      // retrieve session key
      //* Connect SB server with user
      await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.paused);
      await Future.delayed(Duration(seconds: 35));

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.resumed);

      await event.waitFor([EventFuncName.onSessionRefreshed]);

      await Future.delayed(Duration(seconds: 35));

      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout.none);

    test(
        'should return throw error when session token is expired and session key is valid',
        () async {
      final event = SessionEventListener();

      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 180,
        expiring: true,
      );

      final expiresAt =
          DateTime.now().add(Duration(seconds: 30)).millisecondsSinceEpoch;

      final user = await testApiCreateUser();

      //* get session token that expires
      final result = await PrivateApiClient().send(UserSessionTokenIssueRequest(
        userId: user.userId,
        expiresAt: expiresAt,
      ));

      //* Connect user with the expiring session token
      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(user.userId, equals(connectedUser.userId));

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.paused);
      await Future.delayed(Duration(seconds: 35));

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.resumed);
      final eventResult =
          await event.waitFor([EventFuncName.onSessionTokenRequired]);
      // session require token called
      try {
        eventResult.first.data.first('invalid_token');
      } catch (e) {
        print('Refresh error $e');
      }

      await event.waitFor([EventFuncName.onSessionError]);
      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout.none);

    test(
        'should throw exception when session token and session key are expired',
        () async {
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 40,
        expiring: true,
      );

      final expiresAt =
          DateTime.now().add(Duration(seconds: 35)).millisecondsSinceEpoch;

      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient().send(UserSessionTokenIssueRequest(
          userId: user.userId, expiresAt: expiresAt));

      final event = SessionEventListener();

      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(user.userId, equals(connectedUser.userId));

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.paused);
      await Future.delayed(Duration(seconds: 5));
      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.resumed);
      await Future.delayed(Duration(seconds: 5));

      // EXPR already arrives here
      final eventResult = await event
          .waitFor([EventFuncName.onSessionTokenRequired], timeout: 50);
      // session require token called
      try {
        eventResult.first.data.first('invalid_token');
      } catch (e) {
        print('Refresh error $e');
      }

      await event.waitFor([EventFuncName.onSessionError]);
      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout.none);

    test(
        'should throw exception when session token is revoked and session key is valid',
        () async {
      final event = SessionEventListener();
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 180,
        expiring: true,
      );

      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient()
          .send(UserSessionTokenIssueRequest(userId: user.userId));

      //* Connect SB server with user
      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(user.userId, equals(connectedUser.userId));

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.paused);
      await Future.delayed(Duration(seconds: 5));
      //* Revoke User Session Token
      await PrivateApiClient()
          .send(UserSessionTokenRevokeRequest(userId: user.userId));

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.resumed);
      // EXPR already arrives here
      await event.waitFor([EventFuncName.onSessionClosed]);

      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout.none);
    test(
        'should throw exception when session token is revoked and session key is expired',
        () async {
      final event = SessionEventListener();
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 30,
        expiring: true,
      );

      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient()
          .send(UserSessionTokenIssueRequest(userId: user.userId));

      //* Connect SB server with user
      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(user.userId, equals(connectedUser.userId));

      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.paused);
      await Future.delayed(Duration(seconds: 35));

      //* Revoke User Session Token
      await PrivateApiClient()
          .send(UserSessionTokenRevokeRequest(userId: user.userId));
      sdk.getInternal().didChangeAppLifecycleState(AppLifecycleState.resumed);

      await event.waitFor([EventFuncName.onSessionError]);

      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout.none);
  });

  //* Authentication Connected
  group('session authentication while connected', () {
    test('should success when session token and session key are valid',
        () async {
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 5,
        expiring: true,
      );

      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient()
          .send(UserSessionTokenIssueRequest(userId: user.userId));

      //* Connect SB server with user
      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(connectedUser, isA<User>());
      expect(sdk.getInternal().state.connected, isTrue);
    }, timeout: Timeout.none);

    test(
        'should throw exception when session token is expired and session key is valid',
        () async {
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 30,
        expiring: true,
      );

      final expiresAt =
          DateTime.now().add(Duration(seconds: 5)).millisecondsSinceEpoch;

      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient().send(UserSessionTokenIssueRequest(
          userId: user.userId, expiresAt: expiresAt));

      final event = SessionEventListener();

      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(user.userId, equals(connectedUser.userId));

      // EXPR already arrives here

      final eventResult = await event
          .waitFor([EventFuncName.onSessionTokenRequired], timeout: 30);
      // session require token called
      try {
        eventResult.first.data.first('invalid_token');
      } catch (e) {
        print('Refresh error $e');
      }

      await event.waitFor([EventFuncName.onSessionError]);
      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout.none);

    test('should throw exception when session token is revoked', () async {
      final event = SessionEventListener();
      await configureAppForSessionExpiration(
        appId: testInfo.appId,
        sessionDuration: 25,
        expiring: true,
      );

      final user = await testApiCreateUser();

      //get session token
      final result = await PrivateApiClient()
          .send(UserSessionTokenIssueRequest(userId: user.userId));

      //* Connect SB server with user
      final connectedUser = await sdk.connect(
        user.userId,
        accessToken: result['token'],
        apiHost: env.getApiUrl(),
        wsHost: env.getWsUrl(),
      );

      expect(connectedUser, isNotNull);
      expect(user.userId, equals(connectedUser.userId));

      //* Revoke User Session Token
      PrivateApiClient()
          .send(UserSessionTokenRevokeRequest(userId: user.userId));

      await event.waitFor([EventFuncName.onSessionClosed]);

      await configureAppForSessionExpiration(
          appId: testInfo.appId, expiring: false);
    }, timeout: Timeout.none);
  });
}
