import 'dart:async';

import 'package:sendbird_sdk/request/feature/group_channel_read_request.dart';
import 'package:test/test.dart';
import 'package:sendbird_sdk/sendbird_sdk.dart';

import '../common/common.dart';
import '../common/helpers/private_api_client.dart';
import '../common/sdksetup.dart';
import '../common/user.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('sdk total unread message stream', () {
    test('should notify when unread message count has been changed', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      final channel = await createGroupChannel(params);

      final wait = Completer();
      sdk.totalUnreadMessageCountStream?.listen((event) {
        expect(event, equals(1));
        wait.complete();
      });

      final mParams = UserMessageParams(message: 'text');
      await sendUserMessageViaAPI(
        channel: channel,
        params: mParams,
        senderId: users.last.userId,
      );
      await wait.future.timeout(Duration(seconds: 5));
    });

    test('should not get notified when count remains same', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      final channel = await createGroupChannel(params);

      final wait = Completer();
      sdk.totalUnreadMessageCountStream?.listen((event) {
        wait.complete();
      });

      final mParams = UserMessageParams(message: 'text');
      await sendUserMessageViaAPI(
        channel: channel,
        params: mParams,
        senderId: users.first.userId,
      );
      expect(() => wait.future.timeout(Duration(seconds: 5)),
          throwsA(isA<TimeoutException>()));
    });
  });

  group('sdk message receive stream', () {
    test('should notify when new message has been received', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      final channel = await createGroupChannel(params);

      final wait = Completer();
      sdk.messageReceiveStream(channelUrl: channel.channelUrl).listen((event) {
        expect(event, isNotNull);
        expect(event?.message, equals('text'));
        wait.complete();
      });

      final mParams = UserMessageParams(message: 'text');
      await sendUserMessageViaAPI(
        channel: channel,
        params: mParams,
        senderId: users.last.userId,
      );

      await wait.future.timeout(Duration(seconds: 5));
    });
  });

  group('sdk message update stream', () {
    test('should notify when message has been updated', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      final channel = await createGroupChannel(params);

      final wait = Completer();
      sdk.messageUpdateStream(channelUrl: channel.channelUrl).listen((event) {
        expect(event, isNotNull);
        expect(event?.message, equals('updated'));
        wait.complete();
      });

      final mParams = UserMessageParams(message: 'text');
      final msg = await sendUserMessageViaAPI(
        channel: channel,
        params: mParams,
        senderId: user.userId,
      );

      mParams.message = 'updated';
      await channel.updateUserMessage(msg.messageId, mParams);
      await wait.future.timeout(Duration(seconds: 5));
    });
  });

  group('sdk channel read stream', () {
    test('should notify when channel has been updated with read', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      final channel = await createGroupChannel(params);

      final wait = Completer();
      sdk.messageReceiveStream(channelUrl: channel.channelUrl).listen((event) {
        expect(event, isNotNull);
        expect(event?.message, equals('text'));

        PrivateApiClient().send(GroupChannelMarkAsReadApiRequest(
          channelUrls: [channel.channelUrl],
          userId: users[1].userId,
        ));
      });

      sdk.readStream(channelUrl: channel.channelUrl).listen((event) {
        wait.complete();
      });

      final mParams = UserMessageParams(message: 'text');
      await sendUserMessageViaAPI(
        channel: channel,
        params: mParams,
        senderId: users.last.userId,
      );

      await wait.future.timeout(Duration(seconds: 5));
    });
  });

  group('sdk channel change', () {
    test('should notify when channel has been changed', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      final channel = await createGroupChannel(params);

      final wait = Completer();
      sdk.channelChangedStream(channelUrl: channel.channelUrl).listen((event) {
        expect(event, isNotNull);
        wait.complete();
      });

      final mParams = UserMessageParams(message: 'text');
      await sendUserMessageViaAPI(
        channel: channel,
        params: mParams,
        senderId: users.last.userId,
      );

      await wait.future;
    });
  });

  group('sdk channel delivery stream', () {
    test('should notify when channel has been updated with delivery', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();
      final channel = await createGroupChannel(params);

      final wait = Completer();
      sdk.messageReceiveStream(channelUrl: channel.channelUrl).listen((event) {
        expect(event, isNotNull);
        expect(event?.message, equals('text'));

        PrivateApiClient().send(GroupChannelMarkAsReadApiRequest(
          channelUrls: [channel.channelUrl],
          userId: users[1].userId,
        ));
      });

      sdk.deliveryStream(channelUrl: channel.channelUrl).listen((event) {
        wait.complete();
      });

      final mParams = UserMessageParams(message: 'text');
      await sendUserMessageViaAPI(
        channel: channel,
        params: mParams,
        senderId: users.last.userId,
      );

      await wait.future.timeout(Duration(seconds: 5));
    });
  });
}
