import 'package:sendbird_sdk/sendbird_sdk.dart';
import 'package:test/test.dart';

import '../common/common.dart';
import '../common/user.dart';
import '../common/sdksetup.dart';

void main() {
  late TestAppInfo testInfo;
  late SendbirdSdk sdk;
  late User user;

  setUpAll(() async {
    testInfo = await testCreateApp();
  });

  setUp(() async {
    sdk = await testApiSdkSetup(testInfo: testInfo);
  });

  tearDown(() async {
    await sdk.disconnect();
  });

  tearDownAll(() async {
    //delete app?
  });

  group('sdk update user information', () {
    test('should success with given valid params', () async {
      final user = await testApiCreateUserAndConnect();
      await sdk.updateCurrentUserInfo(
        nickname: 'change my name',
        fileInfo: FileInfo.fromUrl(
          url: 'http://www.naver.com',
          mimeType: 'image/jpeg',
        ),
        preferredLanguages: ['en', 'ko'],
      );
      expect(user.nickname, equals('change my name'));
      expect(user.profileUrl, equals('http://www.naver.com'));
      expect(user.preferredLanguages, equals(['en', 'ko']));
    });

    test('should throw an exception with given invalid params', () async {
      await testApiCreateUserAndConnect();
      expect(
          sdk.updateCurrentUserInfo(), throwsA(isA<InvalidParameterError>()));
    });
  });

  group('sdk block user', () {
    test('should success to block given valid userId', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();

      var channel = await createGroupChannel(params);

      await sdk.blockUser(users.last.userId);
      await Future.delayed(Duration(seconds: 3));

      channel.members.forEach((member) {
        if (member.userId == users.last.userId) {
          expect(member.isBlockedByMe, isTrue);
        }
      });
    });

    test('should throw an exception with given invalid userId', () async {
      final users = await testApiCreateUsers(3);
      await testApiConnect(users.first.userId);

      expect(sdk.blockUser('1234'), throwsA(isA<BadRequestError>()));
    });
  });

  group('sdk unblock user', () {
    test('should success to unblock given valid userId', () async {
      final users = await testApiCreateUsers(3);
      final user = await testApiConnect(users.first.userId);

      final params = GroupChannelParams()
        ..operatorUserIds = [user.userId]
        ..userIds = users.map((e) => e.userId).toList();

      var channel = await createGroupChannel(params);

      await sdk.blockUser(users.last.userId);
      await Future.delayed(Duration(seconds: 3));

      channel.members.forEach((member) {
        if (member.userId == users.last.userId) {
          expect(member.isBlockedByMe, isTrue);
        }
      });

      await sdk.unblockUser(users.last.userId);
      await Future.delayed(Duration(seconds: 3));

      channel.members.forEach((member) {
        if (member.userId == users.last.userId) {
          expect(member.isBlockedByMe, isFalse);
        }
      });
    });

    test('should do nothing with given unblocked userId', () async {
      final users = await testApiCreateUsers(3);
      await testApiConnect(users.first.userId);
      await sdk.unblockUser(users.last.userId);
    });
  });

  group('user create metadata', () {
    test('should success with given valid params', () async {
      final user = await testApiCreateUserAndConnect();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await user.createMetaData(metaData);
      expect(user.metaData, isNotNull);
      expect(user.metaData['key1'], equals('value1'));
      expect(user.metaData['key2'], equals('value2'));
    });

    test('should throw an exception with given invalid params', () async {
      final user = await testApiCreateUserAndConnect();
      expect(() => user.createMetaData({}), throwsA(isA<SBError>()));
    });
  });

  group('user update metadata', () {
    test('should success with given valid params', () async {
      final user = await testApiCreateUserAndConnect();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await user.createMetaData(metaData);
      await user.updateMetaData({'key1': 'value3'});
      expect(user.metaData.length, equals(2));
      expect(user.metaData['key1'], equals('value3'));
      expect(user.metaData['key2'], equals('value2'));
    });
    test('should throw an exception with given invalid params', () async {
      final user = await testApiCreateUserAndConnect();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await user.createMetaData(metaData);
      expect(() => user.updateMetaData({}), throwsA(isA<SBError>()));
    });
  });

  group('user delete metadata', () {
    setUp(() async {
      user = await testApiCreateUserAndConnect();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await user.createMetaData(metaData);
    });

    test('should success with given valid params', () async {
      await user.deleteMetaData('key1');
      expect(user.metaData.length, equals(1));
      expect(user.metaData['key1'], isNull);
      expect(user.metaData['key2'], equals('value2'));
    });

    test('should throw an exception with given not exist key', () async {
      expect(() => user.deleteMetaData('key3'), throwsA(isA<SBError>()));
    });

    test('should throw an exception with invalid params', () async {
      expect(() => user.deleteMetaData(''), throwsA(isA<SBError>()));
    });
  });

  group('user delete metadata all', () {
    test('should success', () async {
      user = await testApiCreateUserAndConnect();
      final metaData = {'key1': 'value1', 'key2': 'value2'};
      await user.createMetaData(metaData);
      expect(user.metaData.length, equals(2));
      await user.deleteAllMetaData();
      expect(user.metaData.isEmpty, isTrue);
    });
  });
}
